#include "AdcCalibrationDialog.h"
#include "ui_AdcCalibrationDialog.h"

#include <QTableWidget>
#include <QtMath>

#include "base/DeviceIndexEnabled.h"
#include "cal/CalResult.h"

namespace {
const int RMS_LIMIT = 100;
constexpr const char *k_index = "index";

enum COLS {COL_MEAN, COL_RMS, COL_TOTAL};
}

AdcCalibrationDialog::AdcCalibrationDialog(QWidget *parent)
    : QDialog(parent)
    , ui(new Ui::AdcCalibrationDialog)
{
    ui->setupUi(this);
    qRegisterMetaType<CalResult>();
}

AdcCalibrationDialog::~AdcCalibrationDialog()
{
    delete ui;
}

void AdcCalibrationDialog::onFsmStateChange(const FsmState &state)
{
    fsmState = state;
    ui->pushButtonStartStop->setEnabled(state == FsmState::Ready);
    ui->pushButtonClear->setEnabled(state == FsmState::Ready || state == FsmState::Error);
}

void AdcCalibrationDialog::setCalResult(const CalResult &newCalResult)
{
    auto index = newCalResult.index;
    auto it = config.knownConfigs.find(index);
    if(it == config.knownConfigs.cend())
        return;
    if(it->calResult == newCalResult)
        return;

    it->calResult = newCalResult;
    updateTable(index);
    emit configChanged(config);
}

void AdcCalibrationDialog::setConfig(const ConfigKey &key, const BaseDeviceAppConfig &newConfig)
{
    if (newConfig == config)
        return;

    auto newDevSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    auto oldDevSet = config.getDeviceIndexEnabledMap().getEnabled();

    for(auto i : oldDevSet-newDevSet)
        ui->tabWidget->removeTab(getTabIndex(i));

    auto modifiedDevices = newDevSet - oldDevSet;
    auto devList = newDevSet.values();
    for(auto i : modifiedDevices)
        addTab(devList.indexOf(i), i);

    for(auto i : newDevSet & oldDevSet) {
        if(config.knownConfigs.value(i).calResult
            != newConfig.knownConfigs.value(i).calResult)
            modifiedDevices << i;
    }

    config = newConfig;

    for(auto i : modifiedDevices)
        updateTable(i);

    bool hasEnabledDevices = !newDevSet.isEmpty();
    ui->labelNoDevices->setVisible(!hasEnabledDevices);
    ui->tabWidget->setVisible(hasEnabledDevices);
}

void AdcCalibrationDialog::on_pushButtonClear_clicked()
{
    auto cfg = config;
    for(auto i=cfg.knownConfigs.begin(); i!=cfg.knownConfigs.cend(); ++i)
        i->calResult = {};

    emit configChanged(cfg);
}

void AdcCalibrationDialog::on_pushButtonStartStop_clicked()
{
    emit performCalibrate();
}

int AdcCalibrationDialog::getTabIndex(const DeviceIndex &i) const
{
    auto tab = ui->tabWidget;
    for(int j=0; j<tab->count(); ++j) {
        auto widget = tab->widget(j);
        auto prop = widget->property(k_index);
        auto devIndex = prop.value<DeviceIndex>();
        if(devIndex == i)
            return j;
    }
    return -1;
}

QTableWidget *AdcCalibrationDialog::getTableWidget(const DeviceIndex &i) const
{
    return qobject_cast<QTableWidget *>(ui->tabWidget->widget(getTabIndex(i)));
}

void AdcCalibrationDialog::addTab(int tabIndex, const DeviceIndex &i) const
{
    auto table = new QTableWidget(ui->tabWidget);
    table->setProperty(k_index, QVariant::fromValue(i));
    ui->tabWidget->insertTab(tabIndex, table, i.getSerialStr());
    static const auto headers = QStringList{"Mean", "RMS"};
    table->setColumnCount(headers.size());
    table->setHorizontalHeaderLabels(headers);
}

void AdcCalibrationDialog::updateTable(const DeviceIndex &index) const
{
    auto table = getTableWidget(index);
    if(!table) {
        Q_ASSERT(!devSet.contains(index));
        return;
    }

    auto calRes = config.knownConfigs[index].calResult;
    auto nch = calRes.nch;
    auto devBaseline = calRes.baseline_x1;

    const auto oldNCh = table->rowCount();
    table->setRowCount(nch);

    QColor tabColor;
    for(int ch=0; ch<nch; ++ch) {
        if(ch >= oldNCh) {
            table->setVerticalHeaderItem(ch, new QTableWidgetItem(QString::number(ch+1)));
            table->setItem(ch, COL_MEAN, new QTableWidgetItem());
            table->setItem(ch, COL_RMS, new QTableWidgetItem());
        }

        auto rmsItem = table->item(ch, COL_RMS);
        auto chIt = devBaseline.constFind(ch);

        QString meanTxt("-");
        QString rmsTxt("-");
        QColor itemColor;
        if(chIt != devBaseline.cend()) {
            meanTxt = QString::number(chIt->mean, 'f', 2);
            rmsTxt = QString::number(chIt->stddev, 'f', 2);
            if(chIt->stddev > RMS_LIMIT) {
                itemColor = tabColor = Qt::red;
            }
        }

        table->item(ch, COL_MEAN)->setText(meanTxt);
        rmsItem->setText(rmsTxt);
        auto brush = rmsItem->foreground();
        if(brush.color() != itemColor)
            rmsItem->setForeground(itemColor);
    }
    ui->tabWidget->tabBar()->setTabTextColor(getTabIndex(index), tabColor);
}
