/*
**    Copyright 2014-2020 Ivan Filippov
**    Copyright 2010-2020 Alexey Shutov
**    Copyright 2010-2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "Adc64MainWindow.h"

#include <stdexcept>

#include <QDebug>
#include <QDesktopServices>
#include <QDialogButtonBox>
#include <QFile>
#include <QHostAddress>
#include <QLabel>
#include <QMap>
#include <QMessageBox>
#include <QInputDialog>
#include <QSettings>
#include <QSplitter>
#include <QStatusBar>
#include <QStringList>
#include <QStyleOptionGroupBox>
#include <QTextCodec>
#include <QTextEdit>
#include <QThread>
#include <QTimer>
#include <QUrl>
#include <QVBoxLayout>

#include "Adc64SignalRouter.h"
#include "AdcSerDesCalDialog/AdcSerDesCalDialog.h"
#include "AdcStatusTableModel.h"
#include "AdcStatusWidget.h"
#include "AdcCalibrationDialog.h"
#include "ChannelsSetupForm.h"
#include "DominoDeviceManager.h"
#include "PreamplifierDialog.h"
#include "LemoMuxDialog.h"
#include "MonitorEventCombiner.h"
#include "SparseDialog/SparseDialog.h"
#include "base/AbstractConfigConverter.h"
#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"
#include "base/ThreadController.h"
#include "configuration-manager/ConfigurationManager.h"
#include "DeserializersControl/DeserializersControlBuildParams.h"
#include "AdcSystem/DominoSettings.h"
#include "mldiscover/LldpDiscovery.h"
#include "qwt_plot_marker.h"
#include "qxw/WaveformPlot.h"
#include "sdb/AfiSdbId.h"
#include "ui_Adc64MainWindow.h"

#include "DeviceSettingsDialog.h"

#include "fsm/BaseFsm.h"
#include "histogramm/DeviceData.h"

namespace  {
constexpr static const char* k_show_config_widget = "showConfigWidget";
constexpr static const char* k_show_waveforms = "showWaveforms";

const int chMapV2[7][9] = {
    {5,2,1,4,3,29,26,31,30},
    {25,28,27,24,23,22,21,18,17},
    {14,13,10,9,12,11,8,7,6},
    {20,19,64,63,62,61,58,57,60},
    {39,38,37,34,33,36,35,16,15},
    {59,56,55,54,53,49,50,51,52},
    {48,47,46,45,42,41,44,43,40}
};
// <X, <guiCh, pcbCh> >
const QMap<int, QMap<int,int> > chMapV4X
( { {1, QMap<int, int>({{14, 28},{07, 29},{15, 30},{ 8, 31},{16, 32},{26, 33},{19, 34},{25, 35},{18, 36} })},
    {2, QMap<int, int>({{27, 59},{20, 58},{28, 57},{21, 56},{29, 55},{22, 54},{30, 53},{23, 52},{31, 51} })},
    {3, QMap<int, int>({{02, 19},{10, 20},{03, 21},{11, 22},{04, 23},{12, 24},{05, 25},{13, 26},{06, 27} })},
    {4, QMap<int, int>({{24, 50},{32, 49},{49, 48},{57, 47},{50, 46},{58, 45},{51, 44},{59, 43},{52, 42} })},
    {5, QMap<int, int>({{45, 10},{38, 11},{46, 12},{39, 13},{47, 14},{40, 15},{48, 16},{ 1, 17},{ 9, 18} })},
    {6, QMap<int, int>({{60, 41},{53, 40},{61, 39},{54, 38},{62, 37},{63, 63},{55, 62},{64, 61},{56, 60} })},
    {7, QMap<int, int>({{33, 01},{41, 02},{34, 03},{42, 04},{35, 05},{43, 06},{36, 07},{44,  8},{37,  9} })},
    {0, QMap<int, int>({{17, 64}})}
  } );
// <gui, pcbCh>
QMap<int,int> chMapV4;

const QString CH_PREFIX_OCTAL = "Octal";
const QString CH_PREFIX_AMP_X = "Amp X";
constexpr const bool ADD_AMP_X_CH_SELECT=false;

const int CH_AMP_X_GROUP_SIZE = 9;
struct DisplayGroupIndex {
    DeviceIndex dev;
    int group;
    bool octet; // true - octet display; flase - amp display
};
}

Q_DECLARE_METATYPE(DisplayGroupIndex)

Adc64MainWindow::Adc64MainWindow(const Adc64StartupOptions &startup_options, QWidget *parent)
    : ReadoutAppMainWindow(parent)
    , ui(new Ui::Adc64MainWindow)
    , evLabel(new QLabel(this))
    , lastEventData( new CompleteEvent)
    , channelsSetupForm(new ChannelsSetupForm(&config, this))
    , deviceSettingsDialog(new DeviceSettingsDialog(this))
    , lemoMuxDialog(new LemoMuxDialog(this))
    , preamplifier_dialog(new PreamplifierDialog(this))
    , calibDialog(new AdcCalibrationDialog(this))
    , pcbSwitchingEnabled(startup_options.pcb_ch)
    , monEvCombiner(new MonitorEventCombiner())
{
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);
    AbstractConfigConverter<BaseDeviceAppConfig>::registerTypes();
    qRegisterMetaType<DominoSettings>();
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<DeviceIndexEnabledMap>();
    qRegisterMetaType<LldpInfoCache>();

    auto &signalRouter = Adc64SignalRouter::instance();
    signalRouter.setMainWindow(this);
    signalRouter.setCalibrationDialog(calibDialog);
    signalRouter.setMonitorEventCombiner(monEvCombiner);

    monEvCombiner->setObjectName("MonEvCombiner");
    new ThreadController(monEvCombiner, this);

    {
        std::map<int,int> tmpChMap2;
        for(const auto& m : chMapV4X.values()){
            const auto tmpChMap = m.toStdMap();
            tmpChMap2.insert(tmpChMap.begin(), tmpChMap.end());
        }
        chMapV4 = QMap<int,int>(tmpChMap2);
    }

    if(Globals::getDebugMode()){
        QMap<int, bool> allCh;
        for(int x=0;x<64/CH_AMP_X_GROUP_SIZE;++x){
            for(int y=0; y<CH_AMP_X_GROUP_SIZE; ++y){
                int ch = chMapV2[x][y];
                Q_ASSERT(!allCh.contains(ch));
                allCh[ch] = true;
            }
        }
        Q_ASSERT(allCh.size() == 64/CH_AMP_X_GROUP_SIZE*CH_AMP_X_GROUP_SIZE);
        allCh.clear();
    }

    statusBar()->addPermanentWidget(evLabel, 2);
    ui->actionShow_PCB_ch->setVisible(pcbSwitchingEnabled);

    auto adcModel = ui->adcStatusWidget->getModel();
    connect(this, &Adc64MainWindow::setAdcStatus,
            ui->adcStatusWidget, &AdcStatusWidget::setAdcStatus);

    connect(ui->adcStatusWidget, &AdcStatusWidget::currentDeviceIndexChanged,
            adcSerDesCalDialog, &AdcSerDesCalDialog::onDeviceChange);
    connect(channelsSetupForm, &ChannelsSetupForm::channelsSetupChanged,
            this, &Adc64MainWindow::channelsSetupFormChanged);
    connect(sparseDialog, &SparseDialog::showSparseMarkersUpdated,
            this, &Adc64MainWindow::setShowSparseMarkers);

    connect(getAdcConfigWidget(), &AdcConfigWidget::updateSparseMarkers,
            this, &Adc64MainWindow::updateSparseMarkers);

    connect(ui->actionPrint_FSM_inputs, &QAction::triggered,
            this, &Adc64MainWindow::printFsmInput);

    connect(ui->pushButtonStart, &QAbstractButton::clicked, this, &Adc64MainWindow::performLocalStart);
    connect(ui->pushButtonStop, &QAbstractButton::clicked, this, &Adc64MainWindow::performLocalStop);

//    QSplitter *splitter_vertical = new QSplitter();
//    splitter_vertical->setOrientation(Qt::Vertical);
//    splitter_vertical->addWidget(ui->widget_settings);
//    splitter_vertical->addWidget(ui->deviceTableWidget);
//    splitter_vertical->addWidget(ui->widget_plot);
//    splitter_vertical->setCollapsible(0, false);
//    splitter_vertical->setCollapsible(1, true);
//    splitter_vertical->setCollapsible(2, false);
//    splitter_vertical->setStretchFactor(0,1);
//    splitter_vertical->setStretchFactor(1,1);
//    splitter_vertical->setStretchFactor(2,10);

//    QGridLayout *layout_main = new QGridLayout();
//    layout_main->addWidget(splitter_vertical);
//    ui->centralWidget->setLayout(layout_main);


    QString styleLargeFont = "font-size: 18pt";
    ui->pushButtonStart->setStyleSheet(styleLargeFont);
    ui->pushButtonStop->setStyleSheet(styleLargeFont);
    ui->pushButtonStart->setIcon(style()->standardIcon(QStyle::SP_MediaPlay));
    ui->pushButtonStop->setIcon(style()->standardIcon(QStyle::SP_MediaStop));

    configuration_manager->setDefaultConfig(BaseDeviceAppConfig());

    ui->actionChannel_map->setVisible(false);

    emit configChanged(config);
}

Adc64MainWindow::~Adc64MainWindow()
{
    emit destroyAll();
    delete ui;
}

DeviceStatusTableViewFancy *Adc64MainWindow::getDeviceStatusTableView()
{
    return ui->adcStatusWidget;
}

AdcConfigWidget *Adc64MainWindow::getAdcConfigWidget()
{
    return ui->adcConfigWidget;
}

LemoMuxDialog *Adc64MainWindow::getLemoMuxDialog()
{
    return lemoMuxDialog;
}

PreamplifierDialog *Adc64MainWindow::getPreamplifierDialog()
{
    return preamplifier_dialog;
}

DeviceChModel *Adc64MainWindow::getDeviceTreeModel() const
{
    return channelsSetupForm->getDeviceTreeModel();
}

DeviceChTreeView *Adc64MainWindow::getDeviceChTreeView() const
{
    return channelsSetupForm->getDeviceChTreeView();
}

void Adc64MainWindow::setConfig(const ConfigKey &key, const BaseDeviceAppConfig &_config)
{
    // no need to check for modifications here, see ConfigDispatcher
    if(config == _config)
        return;
    ReadoutAppMainWindow::setConfig(key, _config);

    updateSparseMarkers();
}

void Adc64MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void Adc64MainWindow::addMenuToolsConfigureActions()
{
    ReadoutAppMainWindow::addMenuToolsConfigureActions();

    QAction *action;
    action = menuTools->addAction("LEMO MUX config...", lemoMuxDialog, &QDialog::show);
    action->setProperty(k_sdb_list, QVariantList{AFI_SDB_ADC_LEMO_MUX_CTRL});
    action->setVisible(false);

    action = menuTools->addAction("AD5622 preamplifier...", preamplifier_dialog, &QDialog::show);
    action->setProperty(k_sdb_list, QVariantList{AFI_SDB_AD5622});
    action->setVisible(false);
}

void Adc64MainWindow::saveAppGuiSettings(QSettings &s) const
{
    ReadoutAppMainWindow::saveAppGuiSettings(s);
    s.beginGroup(SETTING_GROUP_MAINWINDOW);
    s.setValue(k_show_config_widget, guiConfig.showConfigWidget);
    s.setValue(k_show_waveforms, guiConfig.showWaveforms);
    s.endGroup();
}

void Adc64MainWindow::loadAppGuiSettings(const QSettings &s)
{
    ReadoutAppMainWindow::loadAppGuiSettings(s);

    guiConfig.showConfigWidget = s.value(SETTING_GROUP_MAINWINDOW + "/" + k_show_config_widget, true).toBool();
    ui->actionShow_configuration_widget->setChecked(guiConfig.showConfigWidget);

    guiConfig.showWaveforms = s.value(SETTING_GROUP_MAINWINDOW + "/" + k_show_waveforms, true).toBool();
    ui->actionShow_waveforms->setChecked(guiConfig.showWaveforms);
}

void Adc64MainWindow::updateControlState()
{
    ui->widgetStartStop->setEnabled(!underRemoteCtrl);
}

void Adc64MainWindow::channelsSetupFormChanged()
{
    setupChanged(); // TODO: move to setConfig
}

void Adc64MainWindow::setDeserializersControlBuildParams(DeviceIndex index, const DeserializersControlBuildParams &bp)
{
    if(!config.knownConfigs.contains(index)){
        qWarning() << "There is no device in settings with index:" << index;
//        return;
    }

    bool devBpUpdated = true;
    auto nch = bp.getNCh();
    auto it = devicesMetaInfo.find(index);
    if(it == devicesMetaInfo.end()) {
        devicesMetaInfo[index].nch = nch;
    } else {
        if(it->nch == nch) {
            devBpUpdated = false;
        } else {
            it->nch = nch;
        }
    }

    if(devBpUpdated) {
        updateAdcChGroups();
    }
}

void Adc64MainWindow::onMonitorEvent(const CompleteEvent &eventData)
{
    DrawScope(eventData);
}

void Adc64MainWindow::DrawScope(const CompleteEvent &eventData, bool force)
{
    if (!force && scopeDisplayTime.isValid() && scopeDisplayTime.elapsed() < 20) // update too fast
        return;

    *lastEventData = eventData;

    WaveformDataAll mvv;
    QString singleTypeName;
//    QVector<double> trig_tai_ts;
    std::optional<quint32> evNumMin, evNumMax;
    for(auto tqdcEvent : eventData.tqdcEv) {
        if(evNumMax) {
            evNumMin = std::min(*evNumMin, tqdcEvent.evNum);
            evNumMax = std::max(*evNumMax, tqdcEvent.evNum);
        } else {
            evNumMin = evNumMax = tqdcEvent.evNum;
        }
        auto chHits = tqdcEvent.hits;

        auto dsIt = config.knownConfigs.constBegin();
        while(dsIt != config.knownConfigs.cend()) {
            auto i = dsIt.key();
            if(DeviceIndex{i.getDevId(), quint32(i.getSerial())} == tqdcEvent.index)
                break;
            ++dsIt;
        }
        if(dsIt == config.knownConfigs.cend())
            continue;
        auto fullDevIndex = dsIt.key();
        auto typeName = fullDevIndex.getDeviceIdName();
        if(singleTypeName.isNull()) {
            singleTypeName = typeName;
        } else if(!singleTypeName.isEmpty()){
            if(singleTypeName != typeName)
                singleTypeName = "";
        }
//        trig_tai_ts.push_back(dominoData.trigStatus.trig_tai_ts.bit.toSeconds());

//        std::cout << QString("%1").arg(dominoData.trigStatus.trig_in_delay_ns).toStdString() << std::endl;

        int channelOffset = 0;
        auto devStIt = devicesMetaInfo.constFind(fullDevIndex);
        if(devStIt != devicesMetaInfo.constEnd())
            channelOffset = devStIt->chRange.first;
        for(auto chHits : tqdcEvent.hits) {
            for(auto hit : chHits.adcHits) {
                if(displayEn && !(displayEn->second.contains(hit.ch) && displayEn->first == tqdcEvent.index))
                    continue;
                if (hit.wf.empty())
                    continue;
                WaveformData wd;
                wd.c = channelOffset+hit.ch;
                if(pcbSwitchingEnabled && guiConfig.showPcbCh)
                    wd.name = QString("pcb%1").arg(chMapV4.value(hit.ch+1, 0));
                else
                    wd.name = QString("%1 %2 #%3")
                            .arg(typeName)
                            .arg(fullDevIndex.getSerialStr())
                            .arg(hit.ch);

                wd.p.resize(hit.wf.size());
                for(int j=0; j < hit.wf.size(); j++)
                {
                    wd.p[j].y = hit.wf[j];
                    wd.p[j].x = j;
    //                double x_ns = j*dominoData.stepNs - dominoData.trigStatus.trig_in_delay_ns;
    //                wd.p[j].x = x_ns / 1e3;
                }
                mvv.push_back(wd);
            }
        }
        channelOffset += chHits.size();
//        qDebug() << "channelOffset=" << channelOffset;
    }
//    // print timestamp delta
//    if (trig_tai_ts.size() > 1) {
//        for (int i=1; i<trig_tai_ts.size(); i++) {
//            double delta = trig_tai_ts[i] - trig_tai_ts[0];
//            std::cout << QString("%1").arg(delta, 0, 'f', 15).toStdString() << std::endl;
//        }
//    }
    if(!singleTypeName.isEmpty()) {
        for(auto &wd : mvv)
            wd.name = wd.name.remove(singleTypeName).trimmed();
    }
    ui->xPlot->setAllData(mvv);
    QString evTxt = "Ev ";
    if(evNumMin) {
        evTxt += QString("%L1").arg(*evNumMin);
        if(evNumMin != evNumMax)
            evTxt += QString("-%L1").arg(*evNumMax);
    } else
        evTxt += "-";
    evLabel->setText(evTxt);
    //    dominoData.PrintRMS();
    scopeDisplayTime.start();
}

void Adc64MainWindow::onFsmStateChange(const FsmState &state)
{
    ReadoutAppMainWindow::onFsmStateChange(state);

    ui->pushButtonStart->setEnabled(state == FsmState::Ready);
    ui->pushButtonStop->setEnabled(state == FsmState::Run);
}

void Adc64MainWindow::deviceUpdated(const DeviceIndex &index, const LldpInfo &dd)
{
    if(!config.knownConfigs.contains(index)) return;

    updateDisplayedChannels();
}

void Adc64MainWindow::updateSparseMarkers()
{
    bool wasAutoReplot = ui->xPlot->autoReplot();
    ui->xPlot->setAutoReplot(false);

    for(QwtPlotMarker *m : sparseMarkers) {
        delete m;
    }
    sparseMarkers.clear();

    int period=0;
    int sparseNum = 0;
    if(guiConfig.showSparseMarkers) {
        for (const auto &setting : config.knownConfigs) {
            const auto &sparseParams = setting.adcErc.sparse;
            if (setting.enabled && !sparseParams.en) {
                period = sparseParams.period;
                sparseNum = sparseParams.readCellNum;
                break;
            }
        }
    }

    if(sparseNum == 0) {
        ui->xPlot->replot();
    }

    for(int i=0; i<sparseNum; ++i) {
        auto * marker = new QwtPlotMarker();
        marker->setLineStyle(QwtPlotMarker::VLine);
        marker->setXValue(i*period);
        marker->label().font().italic();
        QwtText label(QString("sparse: %1").arg(i*period));
        QFont font = label.font();
        font.setItalic(true);
        label.setFont(font);
        marker->setLabel(label);
        marker->setLabelOrientation(Qt::Vertical);
        marker->setLabelAlignment(Qt::AlignRight | Qt::AlignBottom);
        marker->attach(ui->xPlot);
        sparseMarkers.push_back(marker);
    }
    ui->xPlot->setAutoReplot(wasAutoReplot);
}

void Adc64MainWindow::setupChanged()
{
    emit configChanged(config);
}

void Adc64MainWindow::updateDisplayedChannels()
{
    displayEn.reset();
    auto data = ui->comboBoxOctalADC->currentData();
    if(data.isNull())
        return;

    QSet<int> channels;
    DisplayGroupIndex dgi= data.value<DisplayGroupIndex>();
    if(dgi.octet) {
        for(auto i=0; i<8; ++i)
            channels << dgi.group*8+i;
    } else {
        auto keys = chMapV4X[dgi.group].keys();
        channels = {keys.cbegin(), keys.cend()};
    }

    displayEn = {dgi.dev, channels};
}

void Adc64MainWindow::updateAdcChGroups()
{
    QSignalBlocker blocker(ui->comboBoxOctalADC);
    ui->comboBoxOctalADC->clear();
    int id = 0;
    ui->comboBoxOctalADC->insertItem(id++, "All channels");

    // findout maxDevNCh
    {
        std::optional<int> maxDevNCh;
        for(auto miIt=devicesMetaInfo.cbegin(); miIt!=devicesMetaInfo.cend(); ++miIt)
            maxDevNCh = std::max(maxDevNCh.value_or(0), miIt->nch);
        int devChOffset = 0;
        for(auto miIt=devicesMetaInfo.begin(); miIt!=devicesMetaInfo.end(); ++miIt) {
            miIt->chRange.first = devChOffset;
            miIt->chRange.second = devChOffset+miIt->nch-1;
            devChOffset += *maxDevNCh;
        }
    }

    const bool multiDev = devicesMetaInfo.size() > 1;
    for(auto miIt=devicesMetaInfo.cbegin(); miIt!=devicesMetaInfo.cend(); ++miIt) {
        int octIndex = 0;
        for(int i=0; i<miIt->nch/8; i++) {
            ++octIndex;
            auto txt = QString("%1 %2 (%3-%4)")
                    .arg(CH_PREFIX_OCTAL).arg(octIndex)
                    .arg(miIt->chRange.first+i*8+1).arg(miIt->chRange.first+(i+1)*8);
            auto data = QVariant::fromValue(DisplayGroupIndex{miIt.key(), i, true});
            if(multiDev)
                txt.prepend(miIt.key().getSerialStr() + " ");
            ui->comboBoxOctalADC->insertItem(id++, txt, data);
        }
    }

    if(ADD_AMP_X_CH_SELECT) {
        for(auto miIt=devicesMetaInfo.cbegin(); miIt!=devicesMetaInfo.cend(); ++miIt) {
            for(int xGroupIndex=1; xGroupIndex<=miIt->nch/CH_AMP_X_GROUP_SIZE; ++xGroupIndex){
                auto txt = QString("%1%2 %3")
                        .arg(CH_PREFIX_AMP_X).arg(xGroupIndex)
                        .arg(miIt.key().getSerialStr());
                auto data = QVariant::fromValue(DisplayGroupIndex{miIt.key(), xGroupIndex, false});
                ui->comboBoxOctalADC->insertItem(id++, txt, data);
            }
        }
    }
    updateDisplayedChannels();
}

//--- GUI slots
//--- options

void Adc64MainWindow::on_actionBaseline_RMS_check_triggered()
{
    calibDialog->show();
}

void Adc64MainWindow::on_actionShow_PCB_ch_toggled(bool val)
{
    guiConfig.showPcbCh = val;
    DrawScope(*lastEventData, true);
}

void Adc64MainWindow::on_actionShow_configuration_widget_toggled(bool val)
{
    guiConfig.showConfigWidget = val;
    ui->adcConfigWidget->setVisible(val);
}

void Adc64MainWindow::on_actionShow_waveforms_toggled(bool val)
{
    guiConfig.showWaveforms = val;
    ui->xPlot->setVisible(val);
    emit monitorDecodeEnabledChanged(val);
}

//--about
void Adc64MainWindow::on_actionDocs_triggered()
{
    QDesktopServices::openUrl(QUrl("https://afi-project.jinr.ru/projects/adc64-sw/wiki/"));
}

void Adc64MainWindow::on_actionChannel_map_triggered()
{
    if(!channelMapDialog){
        channelMapDialog = new QDialog(this);
        channelMapDialog->setWindowTitle("Channel map help");
        auto *view = new QTextEdit(channelMapDialog);
        view->setReadOnly(true);
        QFile htmlFile(":/channelMap.html");
        QString html;
        if(htmlFile.open(QIODevice::ReadOnly)){
            html = QString(htmlFile.readAll());
            htmlFile.close();
            if(html.isEmpty())
                html = "<b>Channel map file is empty.</b>";
        } else
            html = "<b>There is no channel map file.</b>";
        view->setText(html);
        channelMapDialog->setLayout(new QVBoxLayout(channelMapDialog));
        channelMapDialog->layout()->addWidget(view);
        channelMapDialog->resize(800, 600);
    }
    channelMapDialog->show();
}

void Adc64MainWindow::on_pushButtonChannelsSetup_clicked()
{
    channelsSetupForm->show();
}

void Adc64MainWindow::on_comboBoxOctalADC_currentIndexChanged(int)
{
    updateDisplayedChannels();

    DrawScope(*lastEventData, true);
}

void Adc64MainWindow::setShowSparseMarkers(bool val)
{
    guiConfig.showSparseMarkers = val;
    updateSparseMarkers();
}
