/*
**    Copyright 2017-2022 Ivan Filippov, Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**
*/

#include "DfpMainWindow.h"
#include "ui_DfpMainWindow.h"

#include <cmath>
#include <QCheckBox>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QHostAddress>
#include <QLabel>
#include <QMessageBox>
#include <QProgressBar>
#include <QSettings>
#include <QStatusBar>
#include <QThread>
#include <QTimer>

#include "DfpFsmMain.h"
#include "DfpFsmStates.h"
#include "DfpTask.h"
#include "EthFlashDev.h"
#include "EthFlashDevGlobal.h"
#include "base/DeviceId.h"
#include "base/fw_version.h"
#include "device-discover/DiscoverDialog.h"
#include "flash-prog/FpgaDataBase.h"
#include "fsm/FsmState.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "lib-common/ReadoutAppConfigDispatcher.h"
#include "lib-common/ReadoutAppSignalRouter.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "widgets/DeviceStatusDialog.h"
#include "widgets/FileDialog.h"

namespace {
enum {
    COL_EN,
    COL_TYPE,
    COL_SERIAL,
    COL_FIRMWARE,
    COL_TEMP,
    COL_FILE_NAME,
    COL_BROWSE_PATH,
    COL_PROGRESS,
    COL_NUM
};
constexpr fw_version_t REV_ETHERNET_WRITE = fw_version_t(1, 0, 24200);
static const constexpr char *PROP_INDEX = "index";
constexpr const auto defaultSearchPath = "/net/afi-nas/vol/firmware/fpga/auto";
QString getSearchPath()
{
    const QSettings settings;
    return settings.value("searchPath", defaultSearchPath).toString();
}

void setSearchPath(const QString &path)
{
    QSettings settings;
    settings.setValue("searchPath", path);
}

bool getSingleCheck()
{
    const QSettings settings;
    return settings.value("singleCheck", false).toBool();
}

} // namespace

Q_DECLARE_METATYPE(fw_version_t);
DfpMainWindow::DfpMainWindow(QWidget *parent)
    : ReadoutAppMainWindow(parent)
    , ui(new Ui::DfpMainWindow)
    , signalRouter(new ReadoutAppSignalRouter())
    , configDb(new ReadoutAppConfigDb(this))
    , fsm(new DfpFsmMain(this))
    , discoverDialog(new DiscoverDialog(this))
    , deviceDispatcher(new DeviceDispatcher(this))
    , lastStatusLabel(new QLabel(this))
    , dataBase(new FpgaDataBase(this))
    , singleCheck(getSingleCheck())
    , applyChaceTimer(new QTimer(this))
{
    qRegisterMetaType<DFP_TASK>();
    qRegisterMetaType<FlashImageType>();

    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);

    ui->mainToolBar->hide();

    auto configDispatcher = new ReadoutAppConfigDispatcher(this);
    configDispatcher->subscribeConfig(configDb);
    configDispatcher->subscribeCS(configDb);

    signalRouter->setReadoutAppMainWindow(this);
    signalRouter->setConfigDispatcher(configDispatcher);
    signalRouter->setDeviceDispatcher(deviceDispatcher);
    signalRouter->setDeviceMgrFsm(fsm);
    signalRouter->connectAll();

    connect(this, &ReadoutAppMainWindow::setStatus,
            this, &DfpMainWindow::setDeviceStatus);

    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.subscribe(discoverDialog);
    lldpDiscovery.subscribe(deviceDispatcher);
    lldpDiscovery.subscribe(this);
    deviceDispatcher->subscribeDevices(configDispatcher);
    deviceDispatcher->subscribeDevices(discoverDialog);
    deviceDispatcher->subscribeDevices(fsm);

    connect(fsm, &BaseFsm::stateChanged,
            this, &DfpMainWindow::onFsmStateChange);
    connect(this, &DfpMainWindow::printFsmInput,
            fsm, &BaseFsm::printFsmInput);

    initTable();
    ui->actionSingle_check_mode->setChecked(singleCheck);

    statusBar()->addPermanentWidget(lastStatusLabel);

    auto deviceStatusTableView = ui->deviceStatusTableView;
    // BaseSignalRouter::connectAll
    if (deviceStatusTableView && deviceStatusDialog) {
        connect(deviceStatusTableView, &DeviceStatusTableView::currentDeviceIndexChanged,
                deviceStatusDialog, &DeviceStatusDialog::selectDevice);
        connect(deviceStatusTableView, &DeviceStatusTableViewFancy::showDeviceStatus,
                deviceStatusDialog, &DeviceStatusDialog::showDevice);
    }
    // BaseSignalRouter::connect_deviceDispatcher
    if (deviceStatusTableView)
        deviceDispatcher->subscribeDevices(deviceStatusTableView);

    ui->verticalLayoutCentral->setAlignment(ui->horizontalLayoutButtons, Qt::AlignTop);

    connect(applyChaceTimer, &QTimer::timeout, this, &DfpMainWindow::applyStatusCache);
    applyChaceTimer->setSingleShot(true);
    applyChaceTimer->setInterval(500);
    applyChaceTimer->start();
}

DfpMainWindow::~DfpMainWindow()
{
    delete signalRouter;
    delete ui;
}

DeviceStatusTableViewFancy *DfpMainWindow::getDeviceStatusTableView()
{
    return ui->deviceStatusTableView;
}

void DfpMainWindow::on_actionWrite_Golden_Image_triggered()
{
    QMessageBox msgBox;
    msgBox.setText("Do you want to write \"Golden Image\"?");
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setDefaultButton(QMessageBox::No);
    int ret = msgBox.exec();
    if (ret != QMessageBox::Yes)
        return;
    if (!alienStateCheck())
        return;

    curTask = {DFP_TASK_WRITE, FlashImageType::Golden};
    fsm->setRequestState(DfpFsmStateWriting);
}

void DfpMainWindow::on_actionGet_record_logs_triggered()
{
    QList<DeviceIndex> devIndexes;

    const auto &ranges = ui->tableWidget->selectedRanges();
    for (const auto &range : ranges) {
        for (int r = range.bottomRow(); r <= range.bottomRow(); ++r)
            devIndexes.append(getIndexByRow(r));
    }
    if (devIndexes.isEmpty()) {
        for (int r = 0; r < ui->tableWidget->rowCount(); r++)
            devIndexes.append(getIndexByRow(r));
    }

    for (const auto &i : devIndexes) {
        qInfo().noquote() << QString("\t=== Log for %1 ===").arg(i.getSerialStr());
        dataBase->readInfo(i.getSerial());
    }
}

void DfpMainWindow::on_actionVerify_Golden_image_triggered()
{
    curTask = {DFP_TASK_VERIFY, FlashImageType::Golden};
    fsm->setRequestState(DfpFsmStateReading);
}

void DfpMainWindow::on_actionVerify_regular_image_triggered()
{
    curTask = {DFP_TASK_VERIFY, FlashImageType::Regular};
    fsm->setRequestState(DfpFsmStateReading);
}

void DfpMainWindow::on_actionErase_Golden_part_triggered()
{
    startErase(FlashImageType::Golden);
}

void DfpMainWindow::on_actionErase_regular_part_triggered()
{
    startErase(FlashImageType::Regular);
}

void DfpMainWindow::on_pushButtonSelectDevices_clicked()
{
    discoverDialog->show();
}

void DfpMainWindow::on_pushButtonWrite_clicked()
{
    if (!alienStateCheck())
        return;

    curTask = {DFP_TASK_WRITE, FlashImageType::Regular};
    fsm->setRequestState(DfpFsmStateWriting);
}

void DfpMainWindow::deviceConnected(const DeviceIndex &index, QString fwStr, const fw_version_t &fwVersion)
{
    const int row = getRowByIndex(index);
    if (row != -1) {
        const bool supported = fwVersion >= REV_ETHERNET_WRITE;
        ui->tableWidget->item(row, COL_FIRMWARE)->setText(fwStr);
        ui->tableWidget->item(row, COL_FIRMWARE)->setTextColor(supported ? Qt::black : Qt::red);
        updateTableSize();
    }
}

void DfpMainWindow::setDeviceStatus(const ModularDeviceStatus &r)
{
    devStCache[r.getDeviceIndex()] = r;
}

void DfpMainWindow::imageWritten(const DeviceIndex &index, FlashImageType imageType, const QString &path)
{
    if (dataBase)
        dataBase->writeInfo(index.getSerial(), imageType == FlashImageType::Golden, path);
}

void DfpMainWindow::liveMagicFailed(const DeviceIndex &index)
{
    int row = getRowByIndex(index);
    if (row == -1)
        return;
    //    config.devices[index].enabled = false;

    QWidget *w = ui->tableWidget->cellWidget(row, COL_EN);
    QCheckBox *cb = qobject_cast<QCheckBox *>(w);
    Q_ASSERT(cb);

    qWarning() << QString("Disable device %1 due to liveMagic error").arg(index.getIdent());
    cb->setChecked(false);
}

void DfpMainWindow::deviceEnabled(bool st)
{
    QObject *cb = sender();
    auto index = cb->property(PROP_INDEX).value<DeviceIndex>();

    Q_ASSERT(!index.isNull());
    auto cfg = config;
    cfg.knownConfigs[index].enabled = st;
    emit configChanged(cfg);
    if (!st) {
        auto row = getRowByIndex(index);
        if (row != -1) {
            for (auto col : {COL_FIRMWARE, COL_TEMP})
                ui->tableWidget->item(row, col)->setText({});
            setTableRowColor(row, {});
        }
    }
}

void DfpMainWindow::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    ReadoutAppMainWindow::onFsmStateChange(state);

    switch (state) {
    case DfpFsmStateErasing:
    case DfpFsmStateReading:
    case DfpFsmStateWriting: {
        taskTimer.start();
        Q_ASSERT(curTask.has_value());
        const auto task = curTask->type;
        const auto imageType = curTask->isGold;
        setLabelStatus("in progress...", "gray");

        QString taskStr = dfpTaskToString(task);
        qInfo().noquote() << QString("\n=== Start %1 ===").arg(taskStr);
        qInfo().noquote() << QString("%1 %2 image %3")
                                 .arg(QDateTime::currentDateTime().toString())
                                 .arg(flashImageTypeStr(imageType))
                                 .arg(taskStr);

        emit startDevTask(task, imageType);
        break;
    }
    case DfpFsmStateDone:
        emit allowFsmTransit();
        break;
    default:;
    }

    switch (oldState) {
    case DfpFsmStateErasing:
    case DfpFsmStateReading:
    case DfpFsmStateWriting: {
        curTask.reset();

        bool allOk = state == DfpFsmStateDone;

        qInfo().noquote() << "===" << (allOk ? "end" : "fail") << "===";
        setLabelStatus(QString("%1 (%2)")
                           .arg(allOk ? QString("ok in %1 sec").arg(taskTimer.elapsed() / 1000) : "fail")
                           .arg(QTime::currentTime().toString()),
                       allOk ? "green" : "red");
        break;
    }
    default:;
    }

    updateTableSize();
    setPushButtonEnabled();
}

void DfpMainWindow::onDevFsmStateChange(const DeviceIndex &index, const FsmState &state)
{
    auto row = getRowByIndex(index);
    if (row == -1)
        return;
    QColor col;
    switch (state) {
    case DfpFsmStateError:
        col = Qt::red;
        break;
    case DfpFsmStateMine:
        col = Qt::darkGreen;
        break;
    case DfpFsmStateAlien:
        col = Qt::darkYellow;
        break;
    default:
        return;
    }
    setTableRowColor(row, col);
}

void DfpMainWindow::changeDevPath()
{
    auto pb = qobject_cast<QPushButton *>(sender());
    if (!pb)
        return;
    auto index = pb->property(PROP_INDEX).value<DeviceIndex>();
    if (index.isNull())
        return;
    auto cfgIt = config.knownConfigs.find(index);
    Q_ASSERT(cfgIt != config.knownConfigs.end());

    QString fileName = FileDialog::getOpenFileName(this,
                                                    tr("Open Bit file"),
                                                    cfgIt->flashProg.imageFilePath,
                                                    tr("Bit Files (*.bit)"));
    if (!fileName.isEmpty()) {
        cfgIt->flashProg.imageFilePath = fileName;
        emit devFileChanged(index, fileName);
        saveConfig();
    }
}

void DfpMainWindow::devFileUpdated(const DeviceIndex &index, const QString &filePath)
{
    auto cfgIt = config.knownConfigs.find(index);
    if (cfgIt != config.knownConfigs.end() && cfgIt->flashProg.imageFilePath != filePath) {
        cfgIt->flashProg.imageFilePath = filePath;
        saveConfig();
        emit devFileChanged(index, filePath);
    }
    int row = getRowByIndex(index);
    if (row < 0)
        return;

    auto fn = filePath.split("/").last();
    auto item = ui->tableWidget->item(row, COL_FILE_NAME);
    item->setText(fn.isEmpty() ? "--- missing ---" : fn);
    item->setToolTip(filePath);
    updateTableSize();
}

void DfpMainWindow::lldpCacheUpdated(const LldpInfoCache &lldp)
{
    for (auto it = lldp.cbegin(); it != lldp.cend(); ++it) {
        auto row = getRowByIndex(it.key());
        if (row == -1)
            continue;
        const auto &mn = it.value().med.modelname_str.trimmed();
        if (mn.isEmpty())
            continue;
        auto item = ui->tableWidget->item(row, COL_TYPE);
        if (item->text() != mn)
            item->setText(mn);
    }
}

void DfpMainWindow::setConfig(const ConfigKey &configKey, const BaseDeviceAppConfig &newConfig)
{
    if (config == newConfig)
        return;
    const auto &oldDie = config.getDeviceIndexEnabledMap();
    const auto &newDie = newConfig.getDeviceIndexEnabledMap();
    DeviceIndexSet oldAllSet(oldDie);
    DeviceIndexSet newAllSet(newDie);
    const auto &oldEnSet = oldDie.getEnabled();
    const auto &newEnSet = newDie.getEnabled();
    ReadoutAppMainWindow::setConfig(configKey, newConfig);

    const auto toRemoveDevice = oldEnSet - newEnSet;
    for (const auto &i : toRemoveDevice)
        deleteDevice(i);

    const auto toRemoveRow = oldAllSet - newAllSet;
    for (const auto &i : toRemoveRow) {
        auto row = getRowByIndex(i);
        if (row != -1) {
            ui->tableWidget->removeRow(row);
        }
    }

    const auto toAddRow = newAllSet - oldAllSet;
    auto list = toAddRow.values();
    std::sort(list.begin(), list.end());
    for (const auto &i : list) {
        addDeviceTableRow(i);
    }

    const auto toAddDevice = newEnSet - oldEnSet;
    for (const auto &i : toAddDevice) {
        addDevice(i);
    }

    for (int r = 0; r < ui->tableWidget->rowCount(); ++r) {
        auto cb = qobject_cast<QCheckBox *>(ui->tableWidget->cellWidget(r, COL_EN));
        Q_ASSERT(cb);
        if (!cb)
            continue;
        QSignalBlocker{cb};
        const auto &deviceIndex = getIndexByRow(r);
        cb->setChecked(config.knownConfigs[deviceIndex].enabled);
    }
}

void DfpMainWindow::startErase(FlashImageType imageType)
{
    QMessageBox msgBox(this);
    msgBox.setText(
        QString("Do you want to erase \"%1 Image\"?").arg(flashImageTypeStr(imageType)));
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setDefaultButton(QMessageBox::No);
    int ret = msgBox.exec();
    if (ret == QMessageBox::Yes) {
        curTask = {DFP_TASK_ERASE, imageType};
        fsm->setRequestState(DfpFsmStateErasing);
    }
}

void DfpMainWindow::initTable()
{
    QTableWidget *t = ui->tableWidget;
    t->verticalHeader()->hide();
    t->setColumnCount(COL_NUM);
    t->setHorizontalHeaderItem(COL_TYPE, new QTableWidgetItem("Type"));
    t->setHorizontalHeaderItem(COL_SERIAL, new QTableWidgetItem("Serial"));
    t->setHorizontalHeaderItem(COL_EN, new QTableWidgetItem("En"));
    t->setHorizontalHeaderItem(COL_FIRMWARE, new QTableWidgetItem("f/w"));
    t->setHorizontalHeaderItem(COL_TEMP, new QTableWidgetItem("Temp"));
    t->setHorizontalHeaderItem(COL_FILE_NAME, new QTableWidgetItem("File"));
    t->setHorizontalHeaderItem(COL_BROWSE_PATH, new QTableWidgetItem("Browse path"));
    t->setHorizontalHeaderItem(COL_PROGRESS, new QTableWidgetItem("Progress"));

    t->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    t->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    updateTableSize();
}

void DfpMainWindow::updateTableSize()
{
    QTableWidget *table = ui->tableWidget;
    table->resizeRowsToContents();
    table->resizeColumnsToContents();

    int width = table->verticalHeader()->width() + 5;
    for (int col = 0; col < table->columnCount(); ++col) {
        if (table->isColumnHidden(col))
            continue;
        width += table->columnWidth(col);
    }
    if (width < 20)
        width = 20;

    int height = table->horizontalHeader()->height() + 5;
    for (int row = 0; row < table->rowCount(); ++row) {
        if (table->isRowHidden(row))
            continue;
        height += table->rowHeight(row);
    }
    if (height < 20)
        height = 20;
    table->setFixedSize(width, height);
    table->updateGeometry();
}

DeviceIndex DfpMainWindow::getIndexByRow(int row) const
{
    QTableWidget *t = ui->tableWidget;
    auto item = t->item(row, COL_SERIAL);
    Q_ASSERT(item);
    const auto &index = item->data(Qt::UserRole).value<DeviceIndex>();
    return index;
}

int DfpMainWindow::getRowByIndex(const DeviceIndex &index) const
{
    QTableWidget *t = ui->tableWidget;

    for (int row = 0; row < t->rowCount(); ++row) {
        if (getIndexByRow(row) == index)
            return row;
    }
    return -1;
}

void DfpMainWindow::setPushButtonEnabled() const
{
    auto st = fsm->getState();
    bool isReady = st == DfpFsmStateMine || st == DfpFsmStateAlien;
    bool isError = st == DfpFsmStateError;

    auto table = ui->tableWidget;
    for (auto r = 0; r < table->rowCount(); ++r) {
        table->cellWidget(r, COL_BROWSE_PATH)->setEnabled(isReady || isError);
        table->cellWidget(r, COL_EN)->setEnabled(isReady || isError);
    }

    ui->pushButtonSelectDevices->setEnabled(isReady || isError);
    ui->pushButtonRefreshPaths->setEnabled(isReady || isError);
    for(auto a : {ui->actionSingle_check_mode,
        getActionDiscoverDialog(),
        getActionConfigurationManger()}) {
        a->setEnabled(isReady || isError);
    }

    for(auto a : {ui->actionVerify_Golden_image,
        ui->actionVerify_regular_image,
        ui->actionErase_Golden_part,
        ui->actionErase_regular_part}) {
        a->setEnabled(isReady);
    }
    if (isReady) {
        updateWriteButtonState();
    } else {
        ui->pushButtonWrite->setEnabled(isReady);
        ui->actionWrite_Golden_Image->setEnabled(isReady);
    }
}

void DfpMainWindow::setTableRowColor(int row, const QColor &color)
{
    //    for(auto col = 0; col < ui->tableWidget->columnCount(); ++col) {
    for (auto col : {COL_TYPE, COL_SERIAL, COL_FILE_NAME}) {
        auto i = ui->tableWidget->item(row, col);
        if (i)
            i->setForeground(color);
    }
}

void DfpMainWindow::setLabelStatus(QString msg, QString color) const
{
    lastStatusLabel->setText(msg);
    lastStatusLabel->setStyleSheet(QString("QLabel { color : %1; }").arg(color));
}

bool DfpMainWindow::alienStateCheck()
{
    if (fsm->getState() == DfpFsmStateAlien) {
        QMessageBox msgBox(this);
        msgBox.setText("Some devices will wrote image with non-native signature\n"
                       "Proceed?");
        msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        msgBox.setDefaultButton(QMessageBox::No);
        int ret = msgBox.exec();
        return ret == QMessageBox::Yes;
    }
    return true;
}

void DfpMainWindow::addDeviceTableRow(const DeviceIndex &index)
{
    QTableWidget *table = ui->tableWidget;
    int row = getRowByIndex(index);
    if (row != -1)
        return;

    row = table->rowCount();
    table->setRowCount(row + 1);
    table->setItem(row, COL_TYPE, new QTableWidgetItem(index.getDeviceIdName()));
    table->setItem(row, COL_SERIAL, new QTableWidgetItem(index.getSerialStr()));
    table->item(row, COL_SERIAL)->setData(Qt::UserRole, QVariant::fromValue(index));
    QCheckBox *cb = new QCheckBox(table);
    const auto &devCfg = config.knownConfigs[index];
    cb->setChecked(devCfg.enabled);
    cb->setProperty(PROP_INDEX, QVariant::fromValue(index));
    table->setCellWidget(row, COL_EN, cb);
    connect(cb, &QCheckBox::toggled, this, &DfpMainWindow::deviceEnabled);
    table->setItem(row, COL_FIRMWARE, new QTableWidgetItem());
    table->setItem(row, COL_TEMP, new QTableWidgetItem());
    table->setItem(row, COL_FILE_NAME, new QTableWidgetItem());
    auto pb = new QPushButton("Browse...", table);
    connect(pb, &QAbstractButton::clicked, this, &DfpMainWindow::changeDevPath);
    pb->setProperty(PROP_INDEX, QVariant::fromValue(index));
    table->setCellWidget(row, COL_BROWSE_PATH, pb);
    QProgressBar *bar = new QProgressBar(table);
    bar->setMinimumWidth(111);
    table->setCellWidget(row, COL_PROGRESS, bar);
    bar->setValue(0);

    devFileUpdated(index, devCfg.flashProg.imageFilePath);
}

void DfpMainWindow::addDevice(const DeviceIndex &index)
{
    Q_ASSERT(!devices.contains(index));
    auto devGlob = new EthFlashDevGlobal(index, this);
    devices.insert(index, QSharedPointer<EthFlashDevGlobal>(devGlob));
    const auto &dev = devGlob->getDevice<EthFlashDev>();

    qRegisterMetaType<fw_version_t>();
    connect(dev, &EthFlashDev::fsmStateChanged,
            fsm, &DfpFsmMain::setDevState);
    connect(dev, &EthFlashDev::fsmStateChanged,
            this, &DfpMainWindow::onDevFsmStateChange);

    connect(dev, &EthFlashDev::devConnected,
            this, &DfpMainWindow::deviceConnected);
    connect(dev, &EthFlashDev::liveMagicFailed,
            this, &DfpMainWindow::liveMagicFailed);
    connect(dev, &EthFlashDev::imageWritten,
            this, &DfpMainWindow::imageWritten);

    connect(this, &DfpMainWindow::printFsmInput,
            dev, &EthFlashDev::printFsmInput);
    connect(this, &DfpMainWindow::allowFsmTransit,
            dev, &EthFlashDev::allowFsmTransit);

    connect(this, &DfpMainWindow::setSingleCheckMode,
            dev, &EthFlashDev::setSingleCheckMode);
    connect(this, &DfpMainWindow::autoDetectFile,
            dev, &EthFlashDev::autoDetectFile);
    connect(this, &DfpMainWindow::startDevTask,
            dev, &EthFlashDev::startTask);
    connect(this, &DfpMainWindow::devFileChanged,
            this, &DfpMainWindow::devFileUpdated);
    connect(this, &DfpMainWindow::devFileChanged,
            dev, &EthFlashDev::devFileUpdated);
    connect(dev, &EthFlashDev::devFileChanged,
            this, &DfpMainWindow::devFileUpdated);

    connect(dev, &ModularMregDevice::statusUpdated,
            this, &ReadoutAppMainWindow::setStatus);

    signalRouter->connect_device(devGlob, dev);
    deviceDispatcher->subscribeHostAddressUpdates(index, dev);
    emit devFileChanged(index, config.knownConfigs.value(index).flashProg.imageFilePath);
    emit setSingleCheckMode(singleCheck);

    qInfo() << QString("Added device %1").arg(index.getIdent());
    updateTableSize();
}

void DfpMainWindow::deleteDevice(const DeviceIndex &i)
{
    if (i.isNull()) {
        devices.clear();
    } else {
        devices.remove(i);
    }
}

void DfpMainWindow::saveConfig()
{
    emit configChanged(config);
}

void DfpMainWindow::updateWriteButtonState() const
{
    bool enable = fsm->getState() == DfpFsmStateMine || fsm->getState() == DfpFsmStateAlien;
    //    enable &= hasEnabledDevice;

    ui->pushButtonWrite->setEnabled(enable);
    ui->actionWrite_Golden_Image->setEnabled(enable);
}

void DfpMainWindow::on_pushButtonRefreshPaths_clicked()
{
    emit autoDetectFile(DeviceIndex(), getSearchPath());
}

void DfpMainWindow::on_actionPrint_FSM_input_triggered() const
{
    emit printFsmInput();
}

void DfpMainWindow::on_actionSet_common_bit_file_triggered()
{
    const constexpr static char* KEY = "commonFile";
    QSettings settings;
    auto fPath = settings.value(KEY).toString();

    fPath = FileDialog::getOpenFileName(this,
                                                    tr("Open Bit file"),
                                                    fPath,
                                                    tr("Bit Files (*.bit)"));
    if(fPath.isEmpty())
        return;

    settings.setValue(KEY, fPath);
    for(auto i : config.knownConfigs.keys()) {
        devFileUpdated(i, fPath);
    }

    saveConfig();
}

void DfpMainWindow::on_actionSet_auto_detect_folder_triggered()
{
    QString dir = FileDialog::getExistingDirectory(this, tr("Open FPGA directory"), getSearchPath());
    if (dir.isEmpty())
        return;
    setSearchPath(dir);
    saveConfig();
}

void DfpMainWindow::on_actionSingle_check_mode_toggled(bool checked)
{
    if(singleCheck == checked)
        return;
    singleCheck = checked;
    QSettings settings;
    settings.setValue("singleCheck", singleCheck);

    emit setSingleCheckMode(singleCheck);
}

void DfpMainWindow::applyStatusCache()
{
    bool doResize = false;
    for(auto devSt : devStCache) {
        const int row = getRowByIndex(devSt.getDeviceIndex());
        if (row != -1) {
            QProgressBar *bar = qobject_cast<QProgressBar *>(
                ui->tableWidget->cellWidget(row, COL_PROGRESS));
            if (bar) {
                const auto &st = devSt.flashProgStatus;
                if(st) {
                    auto msg = st->toString();
                    if(bar->format() != msg) {
                        bar->setFormat(msg);
                        doResize = true;
                    }
                    if(!st->taskName.isNull()) {
                        int v = 100.0*st->progress;
                        if(bar->value() != v) {
                            bar->setValue(v);
                            doResize = true;
                        }
                    }
                    doResize = true;
                }
            }
        }
    }
    devStCache.clear();
    if(doResize)
        updateTableSize();
    applyChaceTimer->start();
}
