#include "HrbAnalyzer.h"

#include <QFile>
#include <QVector>
#include <QDebug>
#include <QString>
#include <QDataStream>
#include <QDateTime>
#include <QSettings>
#include <QSemaphore>
#include <qmath.h>
#ifndef _WIN32
#include <unistd.h>
#endif

namespace {
const int CHANNEL_SIZE = 128;
const int HIST_REFRESH_PERIOD = 1000;
}

template <class T>
void reset(QVector<T> &vec)
{
//    vec.clear();
//    vec.resize(CHANNEL_SIZE);
    memset(&vec[0], 0, sizeof(T)*CHANNEL_SIZE);
}

template <class T>
void reset(QVector< QVector<T> > &vec)
{
    for(int samId=0;samId<vec.size();++samId){
        reset(vec[samId]);
    }
}


HrbAnalyzer::HrbAnalyzer(QSemaphore *sem, QObject *p):
    QObject(p),
    queueFreeSizeSem(sem),
    prevEventNum(0),
    curFrameStartPos(0),
    absPos(0),
    sendEndingTimer(this),
    checkParentTimer(this),
    autoClear(false),
    enabled(true),
    lastSampleNumber(0),
    spillNumber(0),
    readyForAutoClear(false)
{
    sendEndingTimer.setSingleShot(true);
    sendEndingTimer.setInterval(HIST_REFRESH_PERIOD);
    connect(&sendEndingTimer, SIGNAL(timeout()), SLOT(sendResults()), Qt::QueuedConnection);
    previousProceed=1;

    checkParentTimer.setSingleShot(false);
    connect(&checkParentTimer, SIGNAL(timeout()), SLOT(checkParent()));
    checkParentTimer.start(1000);
}


void HrbAnalyzer::dataAcived(QByteArray newData)
{
    if(rawData.size())
        rawData.append(newData);
    else
        rawData = newData;
    if(!skipData()){
        previousProceed=true;
//        qDebug()<<"\t newData.size="<<newData.size();
        if(enabled){
            testDummy();
            //    qDebug()<<"Analyzer got "<<newData.size()<<" bytes of data";
//            Q_ASSERT_X(index==i, "dataAcived", "wrong analyzer connection");
//            if(rawData.size())
//                rawData.append(newData);
//            else
//                rawData = newData;
//            int s = rawData.size();
//            if(s > 1e7){
//                absPos+=s-1e7;
//                rawData = rawData.right(1e7);
//                s = rawData.size();
//            }

            read();
//            flushCounters();
            sendResults();
        }
    } else {
        previousProceed = false;
        rawData.clear();
        absPos += newData.size();
    }
    queueFreeSizeSem->release();
    emit eventNumUpdated(prevEventNum);
}

bool HrbAnalyzer::skipData()
{
    return false;
//    progressive porosity
    int queueNumber = queueFreeSizeSem->available();
    if(queueNumber<10)
        return false;
    int s;
    if(queueNumber>60)
        s = 99; // 1%
    else // 10-50
        s = (queueNumber-10)*2; // 100..0 %

    const int rand = qrand()%100; // 0-99
    bool res = rand < s;
    if(previousProceed)
        qDebug()<<"QueueSize="<<queueNumber<<" RandNumer="<<rand<<" threshold="<<s<<(res?" - skipp":" - proceed");

    return res;
}

double HrbAnalyzer::testDummy()
{
    return 0;
    double res=2;
    for(int i=0;i<10000000;++i)
        res = qSqrt(res);
    return res;
}


void HrbAnalyzer::read()
{
    handledSize = 0;
    curPos = 0;
    curFrameStartPos=0;
    quint32 ret;
    while(true){
        try{
            ret = readEvent();
            if(ret) {
                handledSize += ret;
//                qInfo() << "Read event:"<<mpdEventHdr->evNum;
            } else
                break; // event without ending
        } catch (const char* message) {
            if(findNextEventHeader()){
                qWarning().noquote()
                    << message;
            } else {
                prevEventNum = lastEventNumAtPrevFrame;
                handledSize = rawData.size();
                break;
            }
        } catch (const QString message) {
            if(findNextEventHeader()){
                qWarning().noquote()
                    << message;
            } else {
                prevEventNum = lastEventNumAtPrevFrame;
                handledSize = rawData.size();
                break;
            }
        }
    }

    //Erase all up to curFrameStartPos
    if(handledSize){
        if(handledSize==rawData.size())
            rawData.clear();
        else
            rawData = rawData.right(rawData.size()-handledSize);
    }
//    qDebug()<<"data.size="<<rawData.size() << "\thandledSize:"<<handledSize;
//    qDebug()<<"frameCount="<<frameCount;
//    if(data.size()==0)
//        qDebug()<<"\n";
}


quint32 HrbAnalyzer::readEvent()
{
    curFrameStartPos = curPos;
    if(rawData.size() < curPos+sizeof(MpdEventHdr))
        return 0;

//    mpdEventHdr = (MpdEventHdr*)(rawData.data()+curPos);
    mpdHdr = reinterpret_cast<MpdHdr*>(rawData.data()+curPos);

    if(mpdHdr->sync != MPD_SYNC_NORM && mpdHdr->sync != MPD_SYNC_NORM_TLV) {
        if(mpdHdr->sync == MPD_SYNC_EOB || mpdHdr->sync == MPD_SYNC_EOB_TLV) {
            const quint32 fullSize = mpdHdr->length +
                    (mpdHdr->sync == MPD_SYNC_EOB_TLV ? sizeof (MpdHdr)
                                                      : sizeof (MpdEventHdr));

            if(rawData.size()<curPos+fullSize)
                return 0;
            curPos += fullSize;
            if(autoClear)
                readyForAutoClear = true;
            ++spillNumber;
            emit endOfSpill(spillNumber);
            return fullSize;
        } else if(mpdHdr->sync == MPD_SYNC_RUN_START || mpdHdr->sync == MPD_SYNC_RUN_STOP) {
            bool isStart = mpdHdr->sync == MPD_SYNC_RUN_START;
            if(rawData.size() < curPos+mpdHdr->length+sizeof(MpdHdr))
                return 0;
            const quint32 length = mpdHdr->length;

            int shift =0;
            curPos += sizeof(MpdHdr); // Start/stop header
            MpdRunNumberHdr *mpdRunNumHdr = nullptr;
            QString runIndex;
            while(length>shift){
                mpdHdr = (MpdHdr*)(rawData.data()+curPos+shift);
                if(rawData.size()< curPos + shift+sizeof(MpdHdr))
                    throw(QString("Error in Start/Stop format"));
                shift += sizeof(MpdHdr);
                switch (mpdHdr->sync) {
                case MPD_SYNC_RUN_NUMBER: //...7552
                    mpdRunNumHdr = (MpdRunNumberHdr*) mpdHdr;
                    break;
                case MPD_SYNC_RUN_INDEX: //...6E49
                    runIndex = QString::fromLatin1(rawData.mid(curPos+shift,mpdHdr->length));
                    break;
                default:
                    break;
                }
                shift += mpdHdr->length;

            }
            qInfo() << "Got Run" << (isStart?"Start":"Stop") << "record with length:"<<length
                    << "runNumber:" << mpdRunNumHdr->runNumber << "runIndex:" << runIndex;

            curPos += length;
            return length + sizeof(MpdHdr);
        }
        QString mess = QString("Format error: wrong MPD_SYNC_WORD: %1; at file pos:%2")
                        .arg(mpdHdr->sync, 0, 16).arg(curPos);
        throw(mess);
//        throw(mess.toStdString().c_str());
    } else if(autoClear && readyForAutoClear) {
        clear();
        readyForAutoClear = false;
    }
    mpdEventHdr = reinterpret_cast<MpdEventHdr*>(mpdHdr);
    const quint32 fullSize = mpdHdr->length+
            (mpdHdr->sync == MPD_SYNC_NORM_TLV ? sizeof (MpdHdr)
                                               : sizeof (MpdEventHdr));
    if(rawData.size()<curPos+fullSize)
        return 0;

    curPos += sizeof(MpdEventHdr);

    // ====== Event Number Check ======
    if(mpdEventHdr->evNum != prevEventNum+1){
        if(prevEventNum==0xFFFFFF && mpdEventHdr->evNum==0){
            //All ok, event number incremented by 1
        } else if(mpdEventHdr->evNum==0){
            //All ok, new spill started
            qDebug().noquote()
                << QString("#%1: New spill. Prev event number: %2").arg(index).arg(prevEventNum);
            if(autoClear)
                clear();
        } else if(mpdEventHdr->evNum == 1){
            //Seem spill started with event number equaled 1
            qDebug().noquote()
                << QString("#%1: New spill. Prev event number: %2. Spill started with evNum=1.")
                    .arg(index).arg(prevEventNum);
            if(autoClear)
                clear();
        } else {
//            QString mess = QString("Prev event num:%1; Cur event num:%2")
//                    .arg(prevEventNum).arg(evNumber);
//            addLog(mess);
//            qWarning()<<"Prev event num:"<<prevEventNum<<"; Cur event num:"<<evNumber;
        }
    } else {
        //All ok, event number incremented by 1
        qDebug()<<"Ev="<<mpdEventHdr->evNum;
    }

    quint32 byteRead;
    quint32 leftSize = fullSize - sizeof (MpdEventHdr);
    while(leftSize){
        byteRead = readDevice();
        if(leftSize < byteRead)
            throw("Mlink frame size is lesser than sum size of internal frames");
        leftSize -= byteRead;
    }
    prevEventNum = mpdEventHdr->evNum;
    lastEventNumAtPrevFrame = prevEventNum;

    return mpdEventHdr->length + sizeof(MpdEventHdr);
}

void HrbAnalyzer::flushCounters()
{
    foreach (quint64 ser, hrbData.keys()) {
    HrbData &data = hrbData[ser];
    // Add channels counters
    if(data.curCnt.size() < data.tmpCnt.size()){
        const int prevSampleSize = data.curCnt.size();
        // Reserve for additional samples
        data.curCnt.resize(data.tmpCnt.size());
        // Reserve space for channels in additional samples
        for(int i=prevSampleSize;i<data.tmpCnt.size();++i)
            data.curCnt[i].resize(CHANNEL_SIZE);
    }

    for(int samId=0;samId<data.tmpCnt.size();++samId) {
        QVector<quint64> &curSam = data.curCnt[samId];
        QVector<quint64> &tmpSam = data.tmpCnt[samId];
        Q_ASSERT(curSam.size()==tmpSam.size());
        Q_ASSERT(curSam.size()==CHANNEL_SIZE);
        QVector<quint64>::Iterator curIt = curSam.begin();
        QVector<quint64>::Iterator tmpIt = tmpSam.begin();
        QVector<quint64>::Iterator end = curSam.end();
//        double buf;
        for(;curIt!=end;){
//            buf = *tmpIt;
//            *curIt += buf;
            *curIt += *tmpIt;
            ++curIt;
            ++tmpIt;
        }
//            for(int chId=0;chId<CHANNEL_SIZE;++chId){
//                curSam[chId] += tmpSam[chId];
//            }
    }

    reset(data.tmpCnt);
    }
}


quint32 HrbAnalyzer::readDevice()
{
    qint64 eventStartPos = curPos;
    quint32 byteReaded = 0;
    int eventLen;
    int sampleLen;

//    if(curPos+sizeof(MpdDeviceHdr) > rawData.size())
//        throw("Can't read MpdDeviceHdr, end of data");
    mpdDeviceHdr = (MpdDeviceHdr *) (rawData.data()+curPos);
    curPos += sizeof(MpdDeviceHdr);
    byteReaded += sizeof(MpdDeviceHdr);

    if(mpdDeviceHdr->deviceId != 0xc2){
        // Trigger info frame
        curPos += mpdDeviceHdr->length;
        byteReaded += mpdDeviceHdr->length;
//        qInfo()<< hex<<mpdDeviceHdr->deviceSerial;
//        qDebug()<<QString("Event frame is not hrb (type is %1 but must be 0xc2, Skip it")
//                  .arg(mStreamHeader->deviceId, 2, 16, QChar('0'));
//        emit addLog("Trigger info frame, Skip it");
        return byteReaded;
    }

//    if(curPos+sizeof(MpdMStreamHdr) > rawData.size())
//        throw("Can't read MpdMStreamHdr, end of data");
    mpdMStreamHdr = (MpdMStreamHdr *) (rawData.data()+curPos);
    curPos += sizeof(MpdMStreamHdr);
    byteReaded += sizeof(MpdMStreamHdr);

    eventLen = mpdMStreamHdr->words32b*4; // length in bytes

    if(curPos+sizeof(HrbTS) > rawData.size())
        throw("Can't read HrbTS, end of data");
//    HrbTS *hrbTs = (HrbTS *) (rawData.data()+curPos);
    curPos += sizeof(HrbTS);
    byteReaded += sizeof(HrbTS);

    sampleLen = mpdMStreamHdr->words32b*4-sizeof(HrbTS);
//    sampleLen = eventLen-sizeof(HrbTS);

//    int sampleNumber = eventLen; // size of hrb frame in bytes
//    sampleNumber -= sizeof(HrbHeader); // 16 bytes

// 4 bytes per 32-bit words
// 4 32-bit words per sample
    int sampleNumber = sampleLen/16;
    if(eventLen != sampleNumber*16+sizeof(HrbTS)){
        QString mess = QString("Wrong event fragmentLenth:%1(not k*16+%3); eventStartPos:0x%2")
                .arg(eventLen).arg(eventStartPos, 0, 16).arg(sizeof(HrbTS));
        throw(mess.toStdString().c_str());
    }

    // ====== Time Stamp Check ======
//    quint64 system_ts = (((quint64)(hrbHeader->timeStampHi) << 32) | (quint64)hrbHeader->timeStampLo);
//    const quint64 TS_FREQ = 125000000;
////    double system_ts_s = system_ts - (system_ts / TS_FREQ) * TS_FREQ;
////    double system_ts_ns = system_ts - system_ts_s;
//    double system_ts_real = (double)system_ts / (double)TS_FREQ;
//    quint64 system_ts_s = (quint64)system_ts_real;
////        quint64 system_ts_us_part = (quint64)(1000000 * (system_ts_real - system_ts_s));
//    QDateTime date = QDateTime::fromTime_t(system_ts_s);
//    QString dateStr = date.toString();
//    qDebug()<<dateStr;



    // ===== Device Id Check =====
    if(!mpdDeviceHdr->deviceSerial)
        throw("Hrb header has empty device id. Skiping this frame.");
    HrbData &data = hrbData[mpdDeviceHdr->deviceSerial];

    if(data.tmpCnt.size() != sampleNumber){
        int prevSize = data.tmpCnt.size();
        data.tmpCnt.resize(sampleNumber);
        for(int samId=prevSize;samId<sampleNumber;++samId){
            data.tmpCnt[samId].resize(CHANNEL_SIZE);
        }
    }

    if(curPos+sizeof(HrbSample)*sampleNumber > rawData.size())
        throw("There are missing samples, end of data");

    // ====== Samples processing ======
    for(int sampleId=0;sampleId<sampleNumber;++sampleId){
        //One sample per cicle
//        if(curPos+sizeof(HrbSample) > rawData.size())
//            throw("Can't read Sample, end of data");
        sample = (HrbSample *) (rawData.data()+curPos);
        curPos += sizeof(HrbSample);
        byteReaded += sizeof(HrbSample);
        QVector<quint64> &chCounters = data.tmpCnt[sampleId];
        Q_ASSERT(chCounters.size()==CHANNEL_SIZE);
#ifdef OLD_PROCESSING
        for(int ch=0;ch<CHANNEL_SIZE/4; ++ch){
            if(sample->samples[0] & 1){
                ++chCounters[ch];
            }
            sample->samples[0] >>= 1;

            if(sample->samples[1] & 1){
                ++chCounters[ch+32];
            }
            sample->samples[1] >>= 1;

            if(sample->samples[2] & 1){
                ++chCounters[ch+64];
            }
            sample->samples[2] >>= 1;

            if(sample->samples[3] & 1){
                ++chCounters[ch+96];
            }
            sample->samples[3] >>= 1;
        }
#else
        for(int word=0; word<4; ++word){
            if(sample->samples[word]){
                for(int ch=0;ch<CHANNEL_SIZE/4; ++ch){
                    if(sample->samples[word] & 1){
                        ++chCounters[ch+word*32];
                    }
                    sample->samples[word] >>= 1;
                }
            }
        }
#endif
    }
    data.modified = true;
    ++data.sumTrigPassed;
    return byteReaded;
}


bool HrbAnalyzer::findNextEventHeader()
{
    qint64 shift = 0;
    const int dataSize = rawData.size();
    const char * dataStart = rawData.data();
    int pos;
    while(true){
        shift += 2;
        pos = curFrameStartPos+shift;
        if(pos+sizeof(MpdEventHdr) > dataSize){
            break;
        }
        mpdEventHdr = (MpdEventHdr *) (dataStart+pos);
        if(mpdEventHdr->sync == MPD_SYNC_NORM
                || mpdEventHdr->sync == MPD_SYNC_EOB
                || mpdEventHdr->sync == MPD_SYNC_RUN_START
                || mpdEventHdr->sync == MPD_SYNC_RUN_STOP){
            curPos = curFrameStartPos + shift;
            qInfo().noquote()
                << QString("Found MPD_SYNC word at 0x%1; Shif from previous:0x%2")
                    .arg(curFrameStartPos,0,16).arg(shift,0,16).toStdString().c_str();
            return true;
        }
    }
    return false;
}


void HrbAnalyzer::clear(quint32 devId)
{
    sendResults(true);
    if(devId){
        hrbData[devId].clear();
    } else {
        foreach (quint32 s, hrbData.keys()) {
            hrbData[s].clear();
        }
    }
}


void HrbAnalyzer::sendResults(bool force)
{
    const QTime curTime = QTime::currentTime();
    foreach (quint64 ser, hrbData.keys()) {
        HrbData &data = hrbData[ser];
        if(!data.modified)
            continue;
        bool send=true;
        if(data.lastSendTime.isValid()) {
            int msec = data.lastSendTime.msecsTo(curTime);
            if(msec<0)
                msec += 86400000;
            if(msec < HIST_REFRESH_PERIOD)
                send = false;
        }

        if(force || send){
            data.lastSendTime = curTime;
//            emit countersUpdated(ser, data.curCnt, sumTrigs);
            emit countersUpdated(ser, data.tmpCnt, data.sumTrigPassed);
        } else
            sendEndingTimer.start();
    }
}

void HrbAnalyzer::checkParent()
{
#ifndef _WIN32
        __pid_t pId = getppid();
        if(pId==1){
            qDebug()<<"Parent was closed. Exiting...";
            exit(1);
        }
#endif
}

void HrbAnalyzer::autoClearUpdated(bool autoClr)
{
    autoClear = autoClr;
}

//void HrbAnalyzer::updateEnabled()
//{
//    QSettings settings;
//    settings.beginGroup("FileWriter");
//    enabled = !(settings.value("enabled").toBool());
//    QString mess = QString("Analyzer %1").arg(enabled?"enabled":"disabled");
//    addLog(mess);
//}


HrbData::HrbData()  : lastSendTime(QTime::currentTime())
{
}
