#include "HrbSignalRouter.h"

#include <QObject>

#include "base/DeviceId.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "fsm/MStreamDeviceMgrFsm.h"
#include "m-stream-output/MStreamStatisticDialog.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "pnp-server/PnpDiscovery.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"

#include "config/HrbConfigDb.h"
#include "config/HrbConfigDispatcher.h"
#include "hardware/HrbDeviceGlobal.h"
#include "hardware/HrbDevice.h"
#include "hardware/TtbDeviceGlobal.h"
#include "hardware/TtbDevice.h"

#include "HrbCore.h"
#include "HrbModel.h"
#include "HrbMainWindow.h"
#include "StatisticDialog.h"
#include "TtbModel.h"

namespace {
template <typename Func1, typename Func2>
inline void connect(const typename QtPrivate::FunctionPointer<Func1>::Object *sender, Func1 signal,
                    const typename QtPrivate::FunctionPointer<Func2>::Object *receiver, Func2 slot)
{
    QObject::connect(sender, signal, receiver, slot);
}
}

void HrbSignalRouter::setMainWindow(HrbMainWindow *p)
{
    Q_ASSERT(core == nullptr);
    mainWindow = p;
}

void HrbSignalRouter::connectAll()
{
    configDispatcher = &HrbConfigDispatcher::instance();
    Q_ASSERT(configDispatcher);

    connect_ConfigurationDispatcher();
    connect_deviceDispatcher();
    connect_HrbMainWindow();
    connect_PnpServer();
    connect_RemoteControlServer();
    connect_FSM();
    connect_EvNumChecker();
}

void HrbSignalRouter::connect_DeviceGlobal(const HrbDeviceGlobal *devGlob)
{
    if(!devGlob)
        return;
    Q_ASSERT(core);

    const auto &dev = devGlob->getDevice<HrbDevice>();

    connect(dev, &HrbDevice::connectedStateChanged,
            core, &HrbCore::devConnectedStateUpdated);

    connect(dev, &HrbDevice::statusUpdated,
            devGlob, &HrbDeviceGlobal::deviceStatusUpdated);

    connect(dev, &HrbDevice::deviceDisconnected,
            core, &HrbCore::deviceDisconnected);

    connect(core, &HrbCore::mstream_reset_all,
            devGlob, &HrbDeviceGlobal::resetMStream);

    if(mainWindow) {
        connect(mainWindow, &HrbMainWindow::softReset,
                dev, &HrbDevice::softReset);
        connect(mainWindow, &HrbMainWindow::mstream_reset_all,
                devGlob, &HrbDeviceGlobal::resetMStream);
    }

    if(mStreamStatisticDialog) {
        connect(devGlob, &HrbDeviceGlobal::stat_updated,
                mStreamStatisticDialog, &MStreamStatisticDialog::stat_updated);
        connect(devGlob, &HrbDeviceGlobal::stat_from_receiver_updated,
                mStreamStatisticDialog, &MStreamStatisticDialog::stat_from_receiver_updated);
    }

    if(statisticDialog)
        connect(dev, &HrbDevice::statusUpdated,
                statisticDialog, &StatisticDialog::hrbStatusUpdated);

    if(hrbModel)
        connect(dev, &HrbDevice::statusUpdated,
                hrbModel, &HrbModel::deviceStatusUpdated);

    connect(core, &HrbCore::closeProgram,
            devGlob, &HrbDeviceGlobal::closeProgram);
    PnpDiscovery::subscribeMyProgram(devGlob);

    if(evNumChecker) {
        connect(evNumChecker, &EvNumChecker::evNumRequest, dev, &HrbDevice::evNumRequest);
        connect(dev, &HrbDevice::evNumResponse, evNumChecker, &EvNumChecker::gotDevEvNum);
    }

    if(fsm) {
        connect(devGlob, &HrbDeviceGlobal::msdumpReseted,
                fsm, &MStreamDeviceMgrFsm::setDevMstreamSeqNum);
    }

    deviceDispatcher->subscribeHostAddressUpdates(devGlob->getDeviceIndex(), dev);
    configDispatcher->subscribeHrbDevice(dev);
}

void HrbSignalRouter::connect_DeviceGlobal(const TtbDeviceGlobal *devGlob)
{
    if(!devGlob)
        return;

    const auto &dev = devGlob->getDevice<TtbDevice>();

    connect(dev, &TtbDevice::connectedStateChanged,
            core, &HrbCore::devConnectedStateUpdated);

    connect(dev, &TtbDevice::statusUpdated,
            devGlob, &TtbDeviceGlobal::deviceStatusUpdated);

    connect(dev, &TtbDevice::deviceDisconnected,
            core, &HrbCore::deviceDisconnected);

    if(mainWindow){
        connect(mainWindow, &HrbMainWindow::softReset,
                dev, &TtbDevice::softReset);
    }

    if(ttbModel)
        connect(dev, &TtbDevice::statusUpdated,
                ttbModel, &TtbModel::deviceStatusUpdated);

    if(statisticDialog)
        connect(dev, &TtbDevice::statusUpdated,
                statisticDialog, &StatisticDialog::ttbStatusUpdated);

    deviceDispatcher->subscribeHostAddressUpdates(devGlob->getDeviceIndex(), dev);
    configDispatcher->subscribeTtbDevice(dev);
}

void HrbSignalRouter::connect_ConfigurationDispatcher()
{
    if(!configDispatcher)
        return;
    if(core) {
        configDispatcher->subscribeConfig(core);
        configDispatcher->subscribeCS(core);
    }
    if(mainWindow) {
        configDispatcher->subscribeConfig(mainWindow);
        configDispatcher->subscribeCS(mainWindow);
    }
    if(configDb) {
        configDispatcher->subscribeConfig(configDb);
        configDispatcher->subscribeCS(configDb);
    }
    if(configurationManager)
        configDispatcher->subscribeCS(configurationManager);
    if(statisticDialog) {
        configDispatcher->subscribeCS(statisticDialog);
        deviceDispatcher->subscribeDevices(statisticDialog);
    }
}

void HrbSignalRouter::connect_deviceDispatcher()
{
    if(!deviceDispatcher)
        return;
    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.setDeviceIdFilter({DEVICE_ID_TTB9V, DEVICE_ID_HRB6ASD});
    lldpDiscovery.subscribe(deviceDispatcher);

    if(discoverDialog) {
        lldpDiscovery.subscribe(discoverDialog);
        deviceDispatcher->subscribeDevices(discoverDialog);
    }
    if(configDispatcher)
        deviceDispatcher->subscribeDevices(configDispatcher);
    if(fsm){
        deviceDispatcher->subscribeDevices(fsm);
     }
    if(hrbModel)
        lldpDiscovery.subscribe(hrbModel);
    if(ttbModel)
        lldpDiscovery.subscribe(ttbModel);
}

void HrbSignalRouter::connect_HrbMainWindow()
{
    if(!mainWindow || !core)
        return;
    connect(mainWindow, &HrbMainWindow::performLocalStart,
            core, &HrbCore::performLocalStart);
    connect(mainWindow, &HrbMainWindow::performLocalStop,
            core, &HrbCore::performLocalStop);
    connect(core, &HrbCore::preparingRun,
            mainWindow, &HrbMainWindow::preparingRun);
}

void HrbSignalRouter::connect_PnpServer()
{
    if(!core)
        return;

    PnpDiscovery::subscribeMyProgram(core);
}

void HrbSignalRouter::connect_RemoteControlServer()
{
    if(!remoteControlServer)
        return;

    if (core) {
        connect(remoteControlServer, &RemoteControlServer::startRequest,
                core, &HrbCore::remoteStartRequest);
        connect(remoteControlServer, &RemoteControlServer::stopRequest,
                core, &HrbCore::remoteStopRequest);
        connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                core, &HrbCore::onRemoteStatusUpdated);

        connect(core, &HrbCore::closeAll,
                remoteControlServer, &RemoteControlServer::deleteLater);
        connect(core, &HrbCore::sendRunError,
                remoteControlServer,
                QOverload<const RcRunId &, QString>::of(&RemoteControlServer::onRunError));

        connect(core, &HrbCore::sendRunWarning,
                remoteControlServer, &RemoteControlServer::onRunWarning);
        connect(core, &HrbCore::allowRemoteControl,
                remoteControlServer, &RemoteControlServer::startListen);
    }
    if(mainWindow){
        connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                mainWindow, &HrbMainWindow::onRemoteStatusUpdated);
    }
}

void HrbSignalRouter::connect_FSM()
{
    if(!fsm)
        return;

    if(remoteControlServer){
        connect(fsm, &BaseFsm::stateChanged,
                remoteControlServer, &RemoteControlServer::onFsmStateChange);
    }
    if(mainWindow){
        connect(fsm, &BaseFsm::stateChanged,
                mainWindow, &HrbMainWindow::onFsmStateChange);
        connect(mainWindow, &HrbMainWindow::printFsmInputs,
                fsm, &BaseFsm::printFsmInput);
    }
    if(core){
        connect(fsm, &BaseFsm::stateChanged,
                core, &HrbCore::onFsmStateChange);
    }
    if(configDispatcher){
        connect(configDispatcher, &HrbConfigDispatcher::configKeyChanged,
                fsm, &DeviceMgrFsm::setMainSetupKey);
    }
}

void HrbSignalRouter::connect_EvNumChecker()
{
    if(!evNumChecker)
        return;
    if(remoteControlServer) {
        connect(evNumChecker, &EvNumChecker::completedEvNumCheck,
                remoteControlServer, &RemoteControlServer::evNumCheckResponse);
        connect(remoteControlServer, &RemoteControlServer::evNumCheckRequest,
                evNumChecker, &EvNumChecker::onRemoteEvNumCheckRequest);
    }
    if(core)
        connect(evNumChecker, &EvNumChecker::checkFail,
                core, &HrbCore::onRunWarning);
}
