#include "StatisticDialog.h"

#include <QDateTime>
#include <QFile>
#include <QSettings>
#include <QTextEdit>
#include <QTimer>
#include <QVBoxLayout>

#include "base/ConfigIndexName.h"
#include "base/DeviceIndexEnabled.h"

#include "config/params.h"

namespace {
const QString FAIL_RED = "<font color=\"red\">fail</font>";
const int REFRESH_PERIOD_MS = 500;
}

StatisticDialog::StatisticDialog(QWidget *parent)
    : QDialog(parent)
    , refreshTimer(new QTimer(this))
{
    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    editor = new QTextEdit(this);
    editor->setReadOnly(true);
    mainLayout->addWidget(editor);
    setWindowTitle("HRB System statistic dialog");

    setLayout(mainLayout);
    refreshTable();
//    setGeometry(0,0,maximumWidth(),500);
    connect(refreshTimer, SIGNAL(timeout()), SLOT(refreshTable()));
    refreshTimer->setInterval(REFRESH_PERIOD_MS);
    refreshTimer->setSingleShot(true);
}

void StatisticDialog::hrbStatusUpdated(const HrbStatus &newSt)
{
    hrbStatusCurrent[newSt.getDeviceIndex()] = newSt;
    if(!refreshTimer->isActive())
        refreshTimer->start();
}

void StatisticDialog::ttbStatusUpdated(const TtbStatus &newSt)
{
    ttbStatusCurrent[newSt.getDeviceIndex()] = newSt;
    if(!refreshTimer->isActive())
        refreshTimer->start();
}

QDateTime StatisticDialog::getDateTimeFromClock(const quint64 ts, quint64 *us)
{
    const quint64 TS_FREQ = 125000000;
    double system_ts_real = (double)ts / (double)TS_FREQ;
    quint64 system_ts_s = (quint64)system_ts_real;
    QDateTime date = QDateTime::fromTime_t(system_ts_s);
    if(us != NULL){
        *us = (quint64)(1000000 * (system_ts_real - system_ts_s));
    }
    return date;
}

void StatisticDialog::clearStatistic()
{
    hrbStatusCurrent.clear();
    hrbStatusDif.clear();
    hrbStatusLast.clear();
    ttbStatusCurrent.clear();
    ttbStatusDif.clear();
    ttbStatusLast.clear();
}

void StatisticDialog::saveSettings(QSettings &s) const
{
    s.beginGroup("StatisticDialog");
    s.setValue("geometry", saveGeometry());
    s.setValue("visible", isVisible());
    s.endGroup();
}

void StatisticDialog::loadSettings(const QSettings &s)
{
    restoreGeometry(s.value("StatisticDialog/geometry").toByteArray());
    setVisible(s.value("StatisticDialog/visible").toBool());
}

void StatisticDialog::setDeviceList(const DeviceIndexEnabledMap &newList)
{
    const auto &newSet = newList.getEnabled();
    if(devSet == newSet)
        return;

    devSet = newSet;
    clearStatistic();
}

void StatisticDialog::setVisible(bool visible)
{
    QDialog::setVisible(visible);
    if(visible)
        refreshTable();
}

void StatisticDialog::configSelectorChanged(const ConfigIndexName &conf)
{
    QString title="Device Statistic";
    if(!conf.getIndex().isDefault())
        title.append(" #").append(conf.getIndex().toString());
    setWindowTitle(title);
}

void StatisticDialog::reject()
{
    QDialog::reject();
}

void StatisticDialog::refreshTable()
{
    for(auto i : hrbStatusCurrent.keys()) {
        HrbStatus &diffSt = hrbStatusDif[i];
        diffSt = hrbStatusCurrent[i];
        diffSt -= hrbStatusLast[i];
        hrbStatusLast[i] = hrbStatusCurrent[i];
    }
    for(auto i : ttbStatusCurrent.keys()) {
        TtbStatus &diffSt = ttbStatusDif[i];
        diffSt = ttbStatusCurrent[i];
        diffSt -= ttbStatusLast[i];
        ttbStatusLast[i] = ttbStatusCurrent[i];
    }

    html.clear();
    addHrbInfo();
    html += "<br/><br/>\n";
    addTtbInfo();

//    QString full = html.join(" ");
//    QFile f("info.html");
//    bool res = f.open(QIODevice::WriteOnly);
//    qint64 s = f.write(full.toStdString().c_str());
//    f.close();

    if(isVisible())
        editor->setText(html.join(" "));
}
#include <QResizeEvent>
void StatisticDialog::resizeEvent(QResizeEvent *e)
{
    QDialog::resizeEvent(e);
    e->size();
}

QString getHrbClockStatusStr(quint16 val, bool isSelected){
    QStringList s;
    if(val&1<<0) s += isSelected ? "osc" : "mux";
    if(val&1<<1) s += "lemo";
    if(val&1<<2) s += "ttc";
    if(val&1<<3) s += "fmc";
    return s.join(",");
}

void StatisticDialog::addHrbInfo()
{
    if(hrbStatusCurrent.empty()){
        html += "<br/><b style=\"color:red;\">No HRB devices</b>";
        return;
    }

    html += "<b>HRB Devices</b>\n";

    html += "<table border=1 style=\"border-collapse:collapse;\">\n";

    QString header = "<tr>";
//    QString header = "<tr style=\"writing-mode: lr-tb\">";
    // RSTATUS
    header += "<td>Serial Id</style></td><td>FW</td>";
    //    // IN_STATE
    //    header += "<td>in1</td><td>in2</td><td>in3</td><td>in4</td><td>in5</td><td>in6</td><td>in7</td><td>in8</td>";
    // RSTATUS
//    header += "<td>clk_sel</td><td>clk_ok</td>";
    // RTTCSTATE
//    header += "<td>Ttl<br/>bin</td><td>Spill</td><td>Trig</td><td>Busy</td>";
    header += "<td>TTC clk</td><td>TrigSync</td><td>ResetSync</td>";
    // TTC Link Error Counters
    header += "<td>TS zero<br/>mismatch</td><td>TTC TRIG<br/>sync loss</td><td>TTC TRIG<br/>code error</td><td>TTC RESET<br/>sync loss</td><td>TTC RESET<br/>code error</td>";
    // Busy timeout
    header += "<td>Busy timeout</td>";
    // Trigs
    header += "<td>Trig pass</td><td>Trig block</td>";

    header += "<td>Spill num</td>";
    header += "<td>Event num</td>";
    header += "<td>Glob Ev. Num</td>";
    header += "<td>TS</td>";
    header += "<td>FIFO size</td>";

    header += "</tr>\n";
    html += header;

    for(auto i : hrbStatusCurrent.keys()) {
        const HrbStatus& status = hrbStatusCurrent[i];
        const HrbStatistic& statistic = status.statistic;
        const HrbStatus& dif = hrbStatusDif[i];
        const HrbStatistic& statisticDif = dif.statistic;

        QString row;
        row += "<tr>";
        row += QString("<td>%1</td>").arg(status.getDeviceIdent());
        row += QString("<td>%1</td>").arg(status.fwStr);

        for(int i=0;i<8;++i){
            //            row += QString("<td>Input State %1<td>").arg(i);
            //            row += QString("<td>%1</td>").arg(status.inState[i], 4, 16, QChar('0'));
        }
//        // RSTATUS: clock status/selected clock
//        row += QString("<td>%1</td><td>%2</td>")
//                .arg(getHrbClockStatusStr(status.rStatus & 0xF, false))
//                .arg(getHrbClockStatusStr((status.rStatus >> 4) & 0xF, true));

//        // RTTCSTATE: Ttl,LVDS Spill,LVDS Trig,LVDS Busy
//        row += QString("<td>%1</td><td>%2</td><td>%3</td><td>%4</td>")
//                .arg( status.rttcStatus & HrbStatus::RTTC_MASK_TTL_INPUT_ST, 4, 2, QChar('0'))
//                .arg((status.rttcStatus & HrbStatus::RTTC_MASK_SPILL) ?"true":"false")
//                .arg((status.rttcStatus & HrbStatus::RTTC_MASK_TRIG)  ?"true":"false")
//                .arg((status.rttcStatus & HrbStatus::RTTC_MASK_BUSY)  ?"true":"false");

        // RTTCSTATE: TTC clk,Trig Sync,TS Sync
        row += QString("<td>%1</td><td>%2</td><td>%3</td>")
                .arg((status.rttcStatus & HrbStatus::RTTC_MASK_TTC_CLK)   ?"ok":FAIL_RED)
                .arg((status.rttcStatus & HrbStatus::RTTC_MASK_TRIG_SYNC) ?"ok":FAIL_RED)
                .arg((status.rttcStatus & HrbStatus::RTTC_MASK_RESET_SYNC)?"ok":FAIL_RED);

        row += QString("<td>%1</td><td>%2</td><td>%3</td><td>%4</td><td>%5</td>")
                .arg(statistic.tsZeroMismatch)
                .arg(statistic.ttcTrigSyncLoss)
                .arg(statistic.ttcTrigCodeError)
                .arg(statistic.ttcResetSyncLoss)
                .arg(statistic.ttcResetCodeError);

        row += QString("<td>%1(%3%4)</td>")
                .arg(statistic.busyTimeoutCnt)
                .arg((qint32)statisticDif.busyTimeoutCnt<0?"-":"+").arg(qAbs((qint32)statisticDif.busyTimeoutCnt))
                ;

        row += QString("<td>%1(%3%4)</td><td>%2(%5%6)</td>")
                .arg(statistic.trigPass)
                .arg(statistic.trigBlock)
                .arg((qint32)statisticDif.trigPass<0?"-":"+").arg(qAbs((qint32)statisticDif.trigPass))
                .arg((qint32)statisticDif.trigBlock<0?"-":"+").arg(qAbs((qint32)statisticDif.trigBlock))
                //                .arg(100.0*status.trigPass/(status.trigPass+status.trigBlock), 0, 'f', 2)
                ;


        quint64 us;
        row += QString("<td>%1(%2%3)</td><td>%4(%5%6)</td><td>%7(%8%9)</td><td>%10</td>")
                .arg(status.spillNum, 10)
                .arg((qint32)dif.spillNum<0?"-":"+").arg(qAbs((qint32)dif.spillNum))
                .arg(status.evNum, 10)
                .arg((qint32)dif.evNum<0?"-":"+").arg(qAbs((qint32)dif.evNum))
                .arg(status.globEvNum, 10)
                .arg(dif.globEvNum<0?"-":"+").arg(qAbs(dif.globEvNum))
                .arg(getDateTimeFromClock(status.systemTs, &us).toString("yyyy-MM-dd hh:mm:ss"))
//                .arg(us,6, 10, QChar('0'))
                ;

        row += QString("<td>%1(%3%4)</td>")
                .arg(status.fifoSize)
                .arg(dif.fifoSize<0?"-":"+").arg(qAbs(dif.fifoSize))
                ;

        row += "</td></tr>\n";

        html += row;
    }
    html += "</table>\n";

}

void StatisticDialog::addTtbInfo()
{
    if(ttbStatusCurrent.empty()){
        html += "<br/><b style=\"color:red;\">No TTB devices</b>";
        return;
    }

    html += "<b>TTB Devices</b>\n";

    html += "<table border=1 style=\"border-collapse:collapse\">";

    QString header = "<tr>";
    header += "<td>Serial Id</td><td>FW</td>";
    // TTB_CSR_STATUS
    header += "<td>ettc_clock</td><td>ettc</td><td>rjttc clock</td><td>xoff</td><td>ettc_spill</td>";

    // TTB_COUNTERS
    header += "<td>ettcError</td><td>ttcTrig</td><td>ethRxError</td><td>ethTxError</td><td>ethRegioError</td><td>ethRegioTimeout</td>";

    header += "<td>Temp</td><td>evNR</td><td>Time</td>";

    header += "</tr>\n";
    html += header;





    for(auto i : ttbStatusCurrent.keys()) {
        const TtbStatus& status = ttbStatusCurrent[i];
        const TtbStatistic& statistic = status.statistic;
//        const TtbStatus& dif = ttbStatusDif[key];
//        const TtbStatistic& statisticDif = dif.statistic;

        QString row;
        row += "<tr>";
        //    Serial Id
        row += QString("<td>%1</td>").arg(status.getSerialStr());
        row += QString("<td>%1</td>").arg(status.fwStr);

        // REG_TTB_CSR_STATUS
        row += QString("<td>%1</td><td>%2</td><td>%3</td><td>%4</td><td>%5</td>")
                .arg(status.ettc_clock_ok?"ok":FAIL_RED)
                .arg(status.ettc_ok?"ok":FAIL_RED)
                .arg(status.rjttc_clock_ok?"ok":FAIL_RED)
                .arg(status.xoff?"ok":FAIL_RED)
                .arg(status.ettc_spill?"ok":FAIL_RED);

        // TTB_COUNTERS
        row += QString("<td>%1</td><td>%2</td><td>%3</td><td>%4</td><td>%5</td><td>%6</td>")
                .arg(statistic.ettcError)
                .arg(statistic.ttcTrig)
                .arg(statistic.ethRxError)
                .arg(statistic.ethTxError)
                .arg(statistic.ethRegioError)
                .arg(statistic.ethRegioTimeout)
                ;

        // Temp evNR Time
        quint64 us;
        row += QString("<td>%1</td><td>%2</td><td>%3</td>")
                   .arg(TempsInfo(status.getTemps()).brief())
                .arg(status.eventNR)
                .arg(getDateTimeFromClock(status.systemTs, &us).toString("yyyy-MM-dd hh:mm:ss"))
//                .arg(us,6,10,QChar('0'))
                ;


        row += "</tr>\n";

        html += row;
    }
    html += "</table>\n";


    html += "<table border=1 style=\"border-collapse:collapse\">\n";
    QString row="<tr><td rowspan=\"2\"/>";
    QMap<QString,QVector<quint32> > counters;
    counters["linkUp"        ].resize(9*ttbStatusCurrent.size());
    counters["linkError"     ].resize(9*ttbStatusCurrent.size());
    counters["linkTimeout"   ].resize(9*ttbStatusCurrent.size());
    counters["linkErrDisable"].resize(9*ttbStatusCurrent.size());
    counters["linkErrDisableBusySet"].resize(9*ttbStatusCurrent.size());
    counters["linkErrDisableBusyUnset"].resize(9*ttbStatusCurrent.size());
    counters["linkErrDisableBusyMStream"].resize(9*ttbStatusCurrent.size());
    for(int i=0;i<ttbStatusCurrent.size();++i){
        const TtbStatus& status = ttbStatusCurrent[ttbStatusCurrent.keys()[i]];
        row +=QString("<td colspan=\"9\"><center>%1</center></td>").arg(status.getSerialStr());
        for(int p=0;p<9;++p){
            counters["linkUp"        ][p+9*i] = status.statistic.linkUp[p];
            counters["linkError"     ][p+9*i] = status.statistic.linkError[p];
            counters["linkTimeout"   ][p+9*i] = status.statistic.linkTimeout[p];
            counters["linkErrDisable"][p+9*i] = status.statistic.linkErrDisable[p];

            counters["linkErrDisableBusySet"][p+9*i] = status.statistic.linkErrDisableBusySet[p];
            counters["linkErrDisableBusyUnset"][p+9*i] = status.statistic.linkErrDisableBusyUnset[p];
            counters["linkErrDisableBusyMStream"][p+9*i] = status.statistic.linkErrDisableBusyMStream[p];
        }
    }
    row += "</tr>\n";
    html += row;
    row="<tr>";
    for(int i=0;i<ttbStatusCurrent.size();++i){
        for(int p=0;p<TTB_CH_COUNT;++p)
            row += QString("<td>port%1</td>").arg(p);
    }
    row += "</tr>\n";
    html += row;
    foreach (const QString& key, counters.keys()) {
        row=QString("<tr><td>%1</td>").arg(key);
        foreach (quint32 val, counters[key]) {
            row+=QString("<td>%1</td>").arg(val);
        }
        row+="</tr>";
        html+=row;
    }


    html += "</table>\n";
}
