#include "TtbModel.h"

#include <QColor>
#include <QHostAddress>
#include <QPalette>
#include <QDateTime>


#include "base/DeviceIndexEnabled.h"
#include "config/HrbAppConfig.h"
#include "config/HrbConfigDispatcher.h"
#include "device-modules/ModularDeviceConfig.h"
#include "hardware/ClockUtil.h"
#include "hardware/TtbStatus.h"
#include "mldiscover/LldpDiscovery.h"

#include "HrbSignalRouter.h"
#include "TtbModelColumns.h"

struct TtbTableEntry
{
    ModularDeviceConfig config;
    TtbStatus status;
    LldpInfo lldp;
};

class TtbMetaTable : public QMap<DeviceIndex, TtbTableEntry>
{
public:
    int rowOf(const DeviceIndex &deviceIndex) const;
    DeviceIndex indexOf(int row) const;
};

int TtbMetaTable::rowOf(const DeviceIndex &deviceIndex) const
{
    return std::distance(begin(), find(deviceIndex));
}

DeviceIndex TtbMetaTable::indexOf(int row) const
{
    int r = 0;
    for (auto i : keys())
        if (r++ == row)
            return i;
    return DeviceIndex();
}

TtbModel::TtbModel(QObject *parent)
    : QAbstractTableModel(parent)
    , devices(new TtbMetaTable())
    , appCfg(new HrbAppConfig())
{
    HrbSignalRouter::instance().setTtbModel(this);
    HrbConfigDispatcher::instance().subscribeConfig(this);
    HrbConfigDispatcher::instance().subscribeTtbModel(this);
}

TtbModel::~TtbModel()
{
    delete devices;
    delete appCfg;
}

void TtbModel::setEnabled(bool en)
{
    enabled = en;
    sendDataChanged();
}

void TtbModel::invalidateRow(const DeviceIndex &index)
{
    int row = devices->rowOf(index);
    if (row == -1)
        return;

    auto indexFrom = createIndex(row, 0);
    auto indexTo = createIndex(row, columnCount()-1);
    emit dataChanged(indexFrom, indexTo);
}

void TtbModel::sendDataChanged(QVector<int> roles)
{
    emit dataChanged(index(0,0), index(rowCount()-1, columnCount()-1), roles);
}

void TtbModel::lldpCacheUpdated(const LldpInfoCache &lldp)
{
    for(auto it=devices->begin(); it!=devices->end(); ++it) {
        const auto &devLldp = lldp[it.key()];
        if(it->lldp == devLldp)
            continue;
        it->lldp = devLldp;
        invalidateRow(it.key());
    }
}

void TtbModel::setDevConfig(const ModularDeviceConfig &config)
{
    if (config == devices->value(config.index).config)
        return;
    (*devices)[config.index].config = config;
    invalidateRow(config.index);
}

void TtbModel::deviceStatusUpdated(const TtbStatus &status)
{
    const auto &deviceIndex = status.getDeviceIndex();
    if (status == devices->value(deviceIndex).status)
        return;
    (*devices)[deviceIndex].status = status;
    invalidateRow(deviceIndex);
}

void TtbModel::addDevice(const ModularDeviceConfig &cfg)
{
    const auto &index = cfg.index;
    auto tmp = *devices;
    tmp[index];
    int r = tmp.rowOf(index);
    beginInsertRows(QModelIndex(), r, r);
    auto &tte = (*devices)[index];
    tte.config = cfg;
    endInsertRows();
}

void TtbModel::removeDevice(const DeviceIndex &index)
{
    int row = devices->rowOf(index);
    Q_ASSERT(row!=-1);
    if(row==-1)
        return;;
    beginRemoveRows(QModelIndex(), row, row);
    devices->remove(index);
    endRemoveRows();
}

void TtbModel::setConfig(const ConfigKey &key, const HrbAppConfig &cfg)
{
    Q_UNUSED(key);
    if(*appCfg == cfg)
        return;
    const auto &newSet = cfg.getDeviceIndexEnabledMap(DeviceFilter::TTB).getEnabled();
    const auto &oldSet = appCfg->getDeviceIndexEnabledMap(DeviceFilter::TTB).getEnabled();
    *appCfg = cfg;

    for(auto i : oldSet-newSet)
        removeDevice(i);

    for(auto i : newSet-oldSet)
        addDevice(cfg.ttb[i]);

    for(auto i : oldSet & newSet) {
        Q_ASSERT(devices->contains(i));
        auto &oldCfg = (*devices)[i].config;
        auto &newCfg = cfg.ttb[i];
        if(oldCfg == newCfg)
            continue;
        oldCfg = newCfg;
        invalidateRow(i);
    }
}

bool TtbModel::isChCol(int col)
{
    return col >= COL_TTB_CH_FIRST && col <= COL_TTB_CH_LAST;
}

int TtbModel::rowCount(const QModelIndex &) const
{
    return  devices->size();
}

int TtbModel::columnCount(const QModelIndex &) const
{
    return COL_TTB_TOTAL;
}

Qt::ItemFlags TtbModel::flags(const QModelIndex & index) const
{
    auto f = QAbstractTableModel::flags(index);

    if(isChCol(index.column()))
        f |= Qt::ItemIsUserCheckable;

    if(!enabled)
        f &= ~Qt::ItemIsEnabled;

    return f;
}

QVariant TtbModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    auto row = index.row();
    auto deviceIndex = devices->indexOf(row);
    if (deviceIndex.isNull())
        return QVariant();

    switch (role) {
    // case Qt::EditRole: return column == COL_ENABLE || column == COL_VME_CHECK_WR;
    case Qt::BackgroundRole: return data_BackgroundRole(deviceIndex, column);
    case Qt::CheckStateRole: return data_CheckStateRole(deviceIndex, column);
    case Qt::DisplayRole: return data_DisplayRole(deviceIndex, column);
    }
    return QVariant();
}

QVariant TtbModel::data_BackgroundRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];
    // const auto &config = device.config;
    const auto &status = device.status;

    if(!status.connected)
        return QColor(Qt::lightGray);
    if(isChCol(column)) {
        const auto &ch = column - COL_TTB_CH_FIRST;
        const int chMask = 1<<ch;
        if((status.linkEn & chMask) == 0)
            return QVariant();
        if(status.linkErrdisable & chMask)
            return QColor(Qt::red);
        if(status.linkUp & chMask)
            return QColor(0x98fb98);
        else
            return QColor(Qt::yellow);
    }
    return QVariant();
}

QVariant TtbModel::data_CheckStateRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];
    const auto &config = device.config;
    // const auto &status = device.status;

    if(isChCol(column)) {
        const auto &ch = column - COL_TTB_CH_FIRST;
        return config.ttb.chEn[ch] ? Qt::Checked : Qt::Unchecked;
    }
    return QVariant();
}

//void MainWindow::refreshTtbTable()
//{
//    ui->tableTtb->setUpdatesEnabled(false);
//    for(int row=0; row<ui->tableTtb->rowCount(); ++row) {
//        const auto devIndex = rowIndexMapTtb[row];

//        if(!ttbLastStatus.contains(devIndex))
//            continue;
//        const TtbStatus status = ttbLastStatus[devIndex];
//        if(!status.connected)
//            continue;
//        ui->tableTtb->item(row, TTB_COLUMN_IP)->setText(status.hostStr);
//        ui->tableTtb->item(row, TTB_COLUMN_TIME)->setText(
//                    StatisticDialog::getDateTimeFromClock(status.systemTs).toString("hh:mm:ss") );
//        ui->tableTtb->item(row, TTB_COLUMN_TEMP)->setText(QString("%1%2").arg(status.temp,4,'f',1).arg("C"));
//        ui->tableTtb->item(row, TTB_COLUMN_SPILL)->setText(QString::number(status.spillNum));
//        ui->tableTtb->item(row, TTB_COLUMN_EVENT_SPILL)->setText(QString::number(status.eventNR));
//        ui->tableTtb->item(row, TTB_COLUMN_EVENT_GLOB)->setText(QString::number(status.globEvNum));
//        ui->tableTtb->item(row, TTB_COLUMN_SPILL_TIME)->setText(getTimeStr(status.statistic.spillCnt*TTB_CLK_SEC*1e3));
//        const qint64 &globEvNum = status.globEvNum;
//        QTableWidgetItem *busyItem = ui->tableTtb->item(row, TTB_COLUMN_BUSY_TIME);
//        if(config.showAveragedBusy)
//            busyItem->setText(globEvNum?QString("%1 us").arg(status.statistic.ttbBusyCnt*TTB_CLK_SEC*1e6/globEvNum, 0, 'g', 3):"");
//        else
//            busyItem->setText(getTimeStr(status.statistic.ttbBusyCnt*TTB_CLK_SEC*1e3));
//        for(int ch=0;ch<TTB_CH_COUNT;++ch){
//            bool linkEn = status.linkEn & (1<<ch);
//            bool linkUp = status.linkUp & (1<<ch);
//            QTableWidgetItem *chItem = ui->tableTtb->item(row, TTB_COLUMN_CH_0+ch);
//            chItem->setBackgroundColor(linkEn&&linkUp ? QColor("#98fb98") : Qt::white);
//            if(status.statistic.chBusyCnt[ch]) {
//                if(config.showAveragedBusy)
//                    chItem->setText(globEvNum?QString("    %1 us").arg(status.statistic.chBusyCnt[ch]*TTB_CLK_SEC*1e6/globEvNum, 0, 'g', 3):"");
//                else
//                    chItem->setText(getTimeStr(1.*status.statistic.chBusyCnt[ch]*TTB_CLK_SEC*1e3));
//            } else
//                chItem->setText("");
//        }
//    }
//    ui->tableTtb->setUpdatesEnabled(true);
//}

QVariant TtbModel::data_DisplayRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];
//    const auto &config = device.config;
    const auto &status = device.status;
    const auto &lldpInfo = device.lldp;

    switch (column) {
    case COL_TTB_ID: return index.getSerialStr();
    case COL_TTB_IP: return lldpInfo.common.ip_addr ? QHostAddress(lldpInfo.common.ip_addr).toString() : QVariant();
    case COL_TTB_TIME: return getDateTimeFromClock(status.systemTs).toString("hh:mm:ss");
    case COL_TTB_TEMP: return TempsInfo(status.getTemps()).brief();
    case COL_TTB_SPILL: return status.spillNum;
    case COL_TTB_EVENT_SPILL: return status.eventNR;
    case COL_TTB_EVENT_GLOB: return status.globEvNum;
    case COL_TTB_SPILL_TIME: return getTimeStr(status.statistic.spillCnt*TTB_CLK_SEC*1e3);
    case COL_TTB_BUSY_TIME:
    {
        const qint64 &globEvNum = status.globEvNum;
        if(showAveragedBusy)
            return globEvNum?QString("%1 us").arg(status.statistic.ttbBusyCnt*TTB_CLK_SEC*1e6/globEvNum, 0, 'g', 3):"";
        else
            return getTimeStr(status.statistic.ttbBusyCnt*TTB_CLK_SEC*1e3);
    }
    }
    return QVariant();
}

QVariant TtbModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (role != Qt::DisplayRole)
        return QVariant();
    if (orientation != Qt::Horizontal)
        return QVariant();

    if(isChCol(section)) {
        const auto ch = section-COL_TTB_CH_FIRST;
        return QString("Ch #%1").arg(ch);
    }

    switch (section) {
    case COL_TTB_ID: return tr("Id");
    case COL_TTB_IP: return tr("Ip");
    case COL_TTB_TIME: return tr("Time");
    case COL_TTB_TEMP: return tr("Temp");
    case COL_TTB_SPILL: return tr("Spill");
    case COL_TTB_EVENT_SPILL: return tr("Event");
    case COL_TTB_EVENT_GLOB: return tr("GlobEvent");
    case COL_TTB_SPILL_TIME: return tr("Spill time");
    case COL_TTB_BUSY_TIME: return tr("Busy time");
    }
    return QVariant();
}

bool TtbModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    auto column = index.column();
    auto row = index.row();
    auto deviceIndex = devices->indexOf(row);
    if (deviceIndex.isNull())
        return false;

    auto &device = (*devices)[deviceIndex];
    auto &config = device.config;

    if (role != Qt::CheckStateRole)
        return false;

    if(!isChCol(column))
        return false;
    const auto &ch = column-COL_TTB_CH_FIRST;
    config.ttb.chEn[ch] = (value == Qt::Checked);

    emit dataChanged(index, index);
    emit ttbDeviceConfigChanged(config);
    return true;
}
