#include "EventBuilderFsm.h"

#include <QDebug>
#include <QLoggingCategory>

#include "EvBConfig.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm.EvB")

const QMultiMap<FsmState, FsmState> STATE_WORK_LOOP{
    {FsmState::Ready, FsmState::Starting},
    {FsmState::Starting, FsmState::Configure},

    {FsmState::Configure, FsmState::Run},

    {FsmState::Run, FsmState::Stopping},
    {FsmState::Stopping, FsmState::PrepareStopping},
    {FsmState::PrepareStopping, FsmState::Ready},
    {FsmState::Stopping, FsmState::Ready},
};
}

EventBuilderFsm::EventBuilderFsm(QObject *parent):
    BaseFsm(parent)
{
    setObjectName("EventBuilderFsm");
    recreateInput(new EventBuilderFsmInput());
}

bool EventBuilderFsm::isTransitionAllowed(const FsmState &toState) const
{
    if(STATE_WORK_LOOP.contains(getState(), toState)) {
        emit doReqStateAction(toState);
        switch (toState) {
        case FsmState::Starting:
        case FsmState::Stopping:
            return true;
        case FsmState::PrepareStopping:
            return getInput()->combinerState == toState;
        default:
            return allInState(toState);
        }
    }
    return BaseFsm::isTransitionAllowed(toState);
}

void EventBuilderFsm::setCombinerState(FsmState cur_state)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << cur_state;

    getInput()->combinerState = cur_state;
    evaluate();
}

void EventBuilderFsm::setClientManagerFsmState(FsmState cur_state)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << cur_state;

    getInput()->clientManagerFsmState = cur_state;
    evaluate();
}

void EventBuilderFsm::setLookForStopEvNum(bool lookForStopEvNum)
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName() << lookForStopEvNum;

    getInput()->lookForStopEvNum = lookForStopEvNum;
    evaluate();
}

bool EventBuilderFsm::evaluate()
{
    if(BaseFsm::evaluate())
        return true;

    if(getState()==FsmState::Error) {
        if (allInState(FsmState::Ready)) {
            setState(FsmState::Ready);
            return true;
        }
    } else {
        if(hasError()) {
            printFsmInput();
            setState(FsmState::Error);
            return true;
        }

        if(!hasRequestState()) {
            if(!allInState(getState())) {
                if(getState()!= FsmState::Starting && getState()!= FsmState::Stopping
                    && getState()!= FsmState::PrepareStopping) {
                    if(getState() != FsmState::Ready) {
                        qWarning(fsm) << QString("%1 in st:%2, but not all subs in %2; go to err")
                                         .arg(objectName()).arg(fsmStateToString(getState()));
                    }
                    setState(FsmState::Error);
                    return true;
                }
            }
            switch (getState()) {
            case FsmState::Starting:
                setRequestState(FsmState::Configure);
                break;
            case FsmState::Configure:
                setRequestState(FsmState::Run);
                break;
            case FsmState::Stopping:
                setRequestState(getInput()->lookForStopEvNum
                                ? FsmState::PrepareStopping
                                : FsmState::Ready);
                break;
            case FsmState::PrepareStopping:
                setRequestState(FsmState::Ready);
                break;
            default:;
            }
        }
    }

    return false;
}

void EventBuilderFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

qint64 EventBuilderFsm::stateTimeoutMs(const FsmState &state) const
{
    static const qint64 EVB_SYNCT_SUBS_TIMEOUT = 500;
    static const qint64 EVB_RESET_TIMEOUT = 10e3;
    static const qint64 EVB_FILE_OPEN_TIMEOUT = 10e3;
    static const qint64 EVB_FLUSH_OUTPUTS_TIMEOUT = 10e3;

    if(hasRequestState()) {
        switch (state) {
        case FsmState::Starting:
            return EVB_RESET_TIMEOUT;
        case FsmState::Configure:
            return EVB_FILE_OPEN_TIMEOUT;
        case FsmState::PrepareStopping:
        case FsmState::Stopping:
            return EVB_FLUSH_OUTPUTS_TIMEOUT;
        default:
            return BaseFsm::stateTimeoutMs(state);
        }
    } else {
        switch (state) {
        case FsmState::Stopping:
            return EVB_FLUSH_OUTPUTS_TIMEOUT;
        case FsmState::Configure:
        case FsmState::PrepareStopping:
            return EVB_SYNCT_SUBS_TIMEOUT;
        default:
            return BaseFsm::stateTimeoutMs(state);
        }
    }
}

bool EventBuilderFsm::allInState(FsmState state) const
{
    auto i = getInput();
    if(i->combinerState != state ||
            i->clientManagerFsmState != state)
        return false;

    return true;
}

bool EventBuilderFsm::hasError() const
{
    auto i = getInput();
    if(i->combinerState == FsmState::Error ||
            i->clientManagerFsmState == FsmState::Error)
        return true;

    return false;
}

QDebug operator<<(QDebug dbg, const EventBuilderFsmInput &in)
{
    dbg << BaseFsmInput(in);
    dbg << "combiner:" << in.combinerState << ";";
    dbg << "clientManager:" << in.clientManagerFsmState << ";";
    dbg << "lookForStopEvNum:" << in.lookForStopEvNum << ";";
    return dbg;
}
