#include "Tlu40LvdsModel.h"

#include "Tlu40LvdsColumns.h"
#include "base/DeviceId.h"

#include <QTimer>

namespace {
inline Qt::CheckState boolToCheck(bool v)
{ return v ? Qt::Checked : Qt::Unchecked; }
QString getStatusChar(quint64 inputSt, int ch)
// { return (inputSt & (1<<ch)) ? QChar(0x2714) : QChar(0x2716); }
{ return (inputSt & (1<<ch)) ? "V" : "X"; }

Tlu40LvdsColumns getColumnByIndex(const QModelIndex &index)
{
    return static_cast<Tlu40LvdsColumns>(index.column());
}

static bool needReset(const BaseDeviceAppConfig &cfg1, const BaseDeviceAppConfig &cfg2)
{
    auto oldSet = cfg1.getDeviceIndexEnabledMap().getEnabled();
    auto newSet = cfg2.getDeviceIndexEnabledMap().getEnabled();
    if(oldSet != newSet)
        return true;
    for(auto index : oldSet) {
        auto it1 = cfg1.knownConfigs.find(index);
        auto it2 = cfg1.knownConfigs.find(index);
        Q_ASSERT(it1 != cfg1.knownConfigs.end());
        Q_ASSERT(it2 != cfg2.knownConfigs.end());
        if(it1->tlu40lvdsLuminosity.chEn.size() != it2->tlu40lvdsLuminosity.chEn.size())
            return true;
        if(it1->tlu40lvdsTofTrigger.modConfig.size() != it2->tlu40lvdsTofTrigger.modConfig.size())
            return true;
        for(auto i=0; i<it1->tlu40lvdsTofTrigger.modConfig.size(); ++i)
            if(it1->tlu40lvdsTofTrigger.modConfig[i].chEn.size() != it2->tlu40lvdsTofTrigger.modConfig[i].chEn.size())
                return true;
    }

    return false;
}

static bool hasChanges(const ModularDeviceConfig &c1, const ModularDeviceConfig &c2)
{
    if(c1.tlu40lvdsLuminosity != c1.tlu40lvdsLuminosity)
        return true;
    if(c1.tlu40lvdsTofTrigger != c1.tlu40lvdsTofTrigger)
        return true;
    return false;
}

static bool hasChanges(const BaseDeviceAppConfig &c1, const BaseDeviceAppConfig &c2)
{
    for (auto it1=c1.knownConfigs.cbegin(); it1!=c1.knownConfigs.cend(); ++it1) {
        if(!it1->enabled)
            continue;
        auto it2 = c2.knownConfigs.constFind(it1.key());
        Q_ASSERT(it2 !=c2.knownConfigs.cend() && it2->enabled);
        if (hasChanges(*it1, *it2))
            return true;
    }
    return false;
}
}

Tlu40LvdsModel::Tlu40LvdsModel(QObject *parent)
    : QAbstractItemModel(parent)
    , pendingTimer(new QTimer(this))
    , delayedSendConfigTimer(new QTimer(this))
{
    delayedSendConfigTimer->setInterval(10);
    delayedSendConfigTimer->setSingleShot(true);
    connect(delayedSendConfigTimer, &QTimer::timeout,
            this, &Tlu40LvdsModel::postCheckStateChange);

    connect(pendingTimer, &QTimer::timeout,
            this, &Tlu40LvdsModel::processPendingStatus);
    pendingTimer->setSingleShot(true);
    pendingTimer->setInterval(100);
}

QVariant Tlu40LvdsModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::ToolTipRole:
        return toolTipData(index);
    case Qt::TextAlignmentRole:
        return textAlignmentData(index);
    case Qt::CheckStateRole:
        return checkStateData(index);
    case Qt::UserRole:
        return limitsData(index);
    }
    return QVariant();
}

bool Tlu40LvdsModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!index.isValid())
        return false;

    switch (role) {
    case Qt::EditRole:
        return setEditRole(index, value);
    case Qt::CheckStateRole:
        return setCheckStateData(index, value);
    default:;
    }

    return false;
}

Qt::ItemFlags Tlu40LvdsModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);
    auto col = getColumnByIndex(index);
    switch (col) {
    case Tlu40LvdsColumns::COL_EN:
    case Tlu40LvdsColumns::COL_DEC_DES_RESET:
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
        break;
    default:;
    }
    if(!enable)
        f = f & (~Qt::ItemIsEnabled);
    return f;
}

QVariant Tlu40LvdsModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return {};
    if(role != Qt::DisplayRole)
        return {};
    switch (static_cast<Tlu40LvdsColumns>(section)) {
    case Tlu40LvdsColumns::COL_NAME: return "Name";
    case Tlu40LvdsColumns::COL_EN: return "En";
    case Tlu40LvdsColumns::COL_SHAPER: return "Shaper";
    case Tlu40LvdsColumns::COL_STATE: return "State";
    case Tlu40LvdsColumns::COL_CNT: return "Counter";
    case Tlu40LvdsColumns::COL_DEC_DES_RESET: return "DecDes Reset";
    default:;
    }
    return {};
}

QModelIndex Tlu40LvdsModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const auto *parentItem = parent.isValid() ?
        static_cast<Tlu40LvdsModelItem*>(parent.internalPointer()) :
        &rootItem;
    Tlu40LvdsModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex Tlu40LvdsModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const auto *childItem = static_cast<Tlu40LvdsModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    Tlu40LvdsModelItem *parentItem = childItem->parent;
    if(parentItem == &rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int Tlu40LvdsModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
        static_cast<Tlu40LvdsModelItem*>(parent.internalPointer()) :
        &rootItem;
    return parentItem->childCount();
}

int Tlu40LvdsModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return static_cast<int>(Tlu40LvdsColumns::COL_TOTAL);
}

void Tlu40LvdsModel::setEnabled(bool en)
{
    if(enable == en)
        return;

    enable = en;
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh);
}

int Tlu40LvdsModel::getDevRow(const DeviceIndex &di) const
{
    return -1;
}

void Tlu40LvdsModel::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if (config == newConfig)
        return;

    bool performReset = needReset(config, newConfig);
    bool hasAnyChanges = performReset ? true : hasChanges(config, newConfig);

    config = newConfig;
    if (performReset) {
        doReset();
    } else {
        if(hasAnyChanges)
            sendDataChanged();
    }
}

void Tlu40LvdsModel::setTluDesDecBp(const DeviceIndex &index, const TluDecDesBuildParams &bp)
{
    if(index.getDevId() != DEVICE_ID_TLU40LVDS)
        return;
    auto it = devBpDecDes.find(index);
    if(it != devBpDecDes.end() || *it == bp)
        return;
    devBpDecDes[index] = bp;

    auto devIt = config.knownConfigs.find(index);
    if(devIt == config.knownConfigs.end())
        return;
    auto &chEn = devIt->tluDecDes.chReset;
    auto curSize = chEn.size();
    if(curSize == bp.chNum)
        return;

    while(chEn.size() < bp.chNum)
        chEn.push_back(TluDecDesConfig::DEF_CH_RESET);
    chEn.resize(bp.chNum);

    doReset();
    postCheckStateChange();
}

void Tlu40LvdsModel::setTlu40LvdsLuminosityBp(const DeviceIndex &index, const Tlu40LvdsLuminosityBuildParams &bp)
{
    auto it = devBpLum.find(index);
    if(it != devBpLum.end() && *it == bp)
        return;

    devBpLum[index] = bp;
    auto devIt = config.knownConfigs.find(index);
    if(devIt == config.knownConfigs.end())
        return;
    auto &chEn = devIt->tlu40lvdsLuminosity.chEn;
    auto curSize = chEn.size();
    if(curSize == bp.chNum)
        return;

    while(chEn.size() < bp.chNum)
        chEn.push_back(Tlu40LvdsLuminosityConfig::DEF_CH_EN);
    chEn.resize(bp.chNum);

    doReset();
    postCheckStateChange();
}

void Tlu40LvdsModel::setTlu40LvdsTofTriggerBp(const DeviceIndex &index, const Tlu40LvdsTofTriggerBuildParams &bp)
{
    auto it = devBpTof.find(index);
    if(it != devBpTof.end() && *it == bp)
        return;

    devBpTof[index] = bp;
    auto devIt = config.knownConfigs.find(index);
    if(devIt == config.knownConfigs.end())
        return;

    bool modified = false;
    if(devIt->tlu40lvdsTofTrigger.modConfig.size() != bp.modCnt) {
        devIt->tlu40lvdsTofTrigger.modConfig.resize(bp.modCnt);
        modified = true;
    }

    for(Tlu40LvdsTofTriggerModConfig &modCfg : devIt->tlu40lvdsTofTrigger.modConfig) {
        auto &chEn = modCfg.chEn;
        if(chEn.size() != bp.chCntPerMod) {
            while(chEn.size() < bp.chCntPerMod)
                chEn.push_back(Tlu40LvdsTofTriggerModConfig::DEF_CH_EN);
            chEn.resize(bp.trigChCntPerMod);
            modified = true;
        }
    }

    if(modified) {
        doReset();
        postCheckStateChange();
    }
}

void Tlu40LvdsModel::setTlu40LvdsLuminosityStatus(const DeviceIndex &index, const Tlu40LvdsLuminosityStatus st)
{
    auto it = devStatusLum.constFind(index);
    if(it != devStatusLum.cend() && (*it== st))
            return;
    pendingDevStatusLum[index] = st;
    if(!pendingTimer->isActive())
        processPendingStatus();
}

void Tlu40LvdsModel::setTlu40LvdsTofTriggerStatus(const DeviceIndex &index, const Tlu40LvdsTofTriggerStatus &st)
{
    auto it = devStatusTof.find(index);
    if(it != devStatusTof.end() && *it == st)
        return;
    pendingDevStatusTof[index] = st;
    if(!pendingTimer->isActive())
        processPendingStatus();
}

void Tlu40LvdsModel::resetCounters()
{
    devStatusOffsetLum = devStatusLum;
    devStatusOffsetTof = devStatusTof;
    sendDataChanged();
}

void Tlu40LvdsModel::onDeviceConnected(const DeviceIndex &devIndex)
{
    devStatusOffsetLum.remove(devIndex);
    devStatusOffsetTof.remove(devIndex);
    sendDataChanged();
}

void Tlu40LvdsModel::postCheckStateChange()
{
    sendDataChanged(QVector<int>{Qt::CheckStateRole});
    emit configChanged(this->config);
}

void Tlu40LvdsModel::processPendingStatus()
{
    DeviceIndexSet modifiedSet;
    for(auto it=pendingDevStatusLum.cbegin(); it!=pendingDevStatusLum.cend(); ++it){
        devStatusLum[it.key()] = *it;
        modifiedSet << it.key();
    }
    pendingDevStatusLum.clear();

    for(auto it=pendingDevStatusTof.cbegin(); it!=pendingDevStatusTof.cend(); ++it) {
        devStatusTof[it.key()] = *it;
        modifiedSet << it.key();
    }
    pendingDevStatusTof.clear();

    sendDataChanged(QVector<int>{Qt::DisplayRole});

    if(!modifiedSet.isEmpty())
        pendingTimer->start();
}

void Tlu40LvdsModel::enqueueConfigSend()
{
    delayedSendConfigTimer->start();
}

void Tlu40LvdsModel::doReset()
{
    beginResetModel();
    rootItem.children.clear();
    const auto &setups = config.knownConfigs;
    if (config.hasEnabledDevices()) {
        auto allItem = rootItem.createChild();

        for (auto devIt = setups.cbegin(); devIt != setups.cend(); ++devIt) {
            if (!devIt->enabled)
                continue;
            if(devIt.key().getDevId() != DEVICE_ID_TLU40LVDS)
                continue;

            auto devItem = allItem->createChild();
            devItem->dev = devIt.key();
            int mod=0;
            for(auto it=devIt->tlu40lvdsTofTrigger.modConfig.cbegin();
                 it!=devIt->tlu40lvdsTofTrigger.modConfig.cend(); ++it) {
                auto modItem = devItem->createChild();
                modItem->isTof = true;
                modItem->mod = mod++;
                auto inputChPerMod = it->chEn.size()*2; // trigChCntPerMod -> chCntPerMod
                for (int ch = 0; ch < inputChPerMod; ch++) {
                    auto chItem = modItem->createChild();
                    chItem->ch = ch;
                }
            }
            {
                auto modItem = devItem->createChild();
                modItem->isLum = true;
                for (int ch = 0; ch < devIt->tlu40lvdsLuminosity.chEn.size(); ch++) {
                    auto chItem = modItem->createChild();
                    chItem->ch = ch;
                }
            }
        }
    }
    endResetModel();
}

void Tlu40LvdsModel::sendDataChanged(QVector<int> roles, const QModelIndex parent)
{
    auto rows = rowCount(parent);

    if(!rows)
        return;

    auto top = index(0,0, parent);
    auto last = index(rows-1, columnCount(top)-1, parent);
    emit dataChanged(top, last, roles);

    for(auto r=0; r<rows; ++r) {
        auto chIndex = index(r, 0, parent);
        sendDataChanged(roles, chIndex);
    }
}

QVariant Tlu40LvdsModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case Tlu40LvdsColumns::COL_NAME:
        if(item->dev.isNull())
            return "All";
        if(!item->isLum && !item->isTof)
            return item->dev.getSerialStr();

        if(item->ch == -1) {
            if(item->isTof)
                return QString("mod #%1").arg(item->mod);
            else
                return QString("Luminosity");
        }
        return QString("ch #%1").arg(item->ch);
    case Tlu40LvdsColumns::COL_STATE:
        if(item->ch == -1)
            return {};
        if(item->isTof) {
            auto it = devStatusTof.constFind(item->dev);
            if(it == devStatusTof.cend())
                return {};
            if(it->modSt.size() <= item->mod)
                return {};
            auto &modSt = it->modSt[item->mod];

            return getStatusChar(modSt.inputState, item->ch);
        } else {
            auto it = devStatusLum.constFind(item->dev);
            if(it == devStatusLum.cend())
                return {};
            return getStatusChar(it->inputSt, item->ch);
        }
    case Tlu40LvdsColumns::COL_CNT:
    {
        auto pair = getCounterPair(item);
        if(!pair.first)
            return {};
        return QString("%L1").arg(*pair.first - pair.second.value_or(0));
    }
    default:;
    }

    return {};
}

QVariant Tlu40LvdsModel::editData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case Tlu40LvdsColumns::COL_SHAPER:
        if(item->isTof) {
            if(item->ch != -1)
                return {};
            auto devIt = config.knownConfigs.constFind(item->dev);
            if(devIt == config.knownConfigs.cend())
                return {};

            const auto &modsCfg = devIt->tlu40lvdsTofTrigger.modConfig;
            if(modsCfg.size() <= item->mod)
                return {};

            return modsCfg[item->mod].shaperVal;
        } else if(item->isLum) {
            if(item->ch != -1)
                return {};
            auto devIt = config.knownConfigs.constFind(item->dev);
            if(devIt == config.knownConfigs.cend())
                return {};
            return devIt->tlu40lvdsLuminosity.shaperVal;
        }
    default:;
    }

    return {};
}

QVariant Tlu40LvdsModel::toolTipData(const QModelIndex &index) const
{

    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case Tlu40LvdsColumns::COL_STATE:
        return "Current input state";
    case Tlu40LvdsColumns::COL_CNT:
    {
        auto pair = getCounterPair(item);
        if(!pair.first)
            return {};
        auto ret = QString("%L1").arg(*pair.first);
        if(pair.second)
            ret += QString(" - %L2").arg(*pair.second);
        return ret;
    }
    default:;
    }

    return {};
}

QVariant Tlu40LvdsModel::textAlignmentData(const QModelIndex &index) const
{
    switch (getColumnByIndex(index)) {
    case Tlu40LvdsColumns::COL_STATE: return Qt::AlignHCenter;
    default:;
    }

    return {};
}

QVariant Tlu40LvdsModel::checkStateData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    if (!item)
        return {};

    switch (getColumnByIndex(index)) {
    case Tlu40LvdsColumns::COL_EN:
        if(item->isTof) {
            if(item->ch == -1)
                return {};
            auto devIt = config.knownConfigs.constFind(item->dev);
            if(devIt == config.knownConfigs.cend())
                return {};

            const auto &modsCfg = devIt->tlu40lvdsTofTrigger.modConfig;
            if(modsCfg.size() <= item->mod)
                return {};

            const auto &chEn = modsCfg[item->mod].chEn;
            auto ch = item->ch;
            if(chEn.size() <= ch)
                ch -= chEn.size();
            if(chEn.size() <= ch)
                return {};
            return boolToCheck(chEn[ch]);
        } else if(item->isLum) {
            if(item->ch == -1)
                return {};
            auto devIt = config.knownConfigs.constFind(item->dev);
            if(devIt == config.knownConfigs.cend())
                return {};

            const auto &chEn = devIt->tlu40lvdsLuminosity.chEn;
            if(chEn.size() <= item->ch)
                return {};
            return boolToCheck(chEn[item->ch]);
        }
        break;
    case Tlu40LvdsColumns::COL_DEC_DES_RESET:
    {
        if(item->dev.isNull() || item->isTof || item->isLum)
            return {};
        auto devIt = config.knownConfigs.constFind(item->dev);
        if(devIt == config.knownConfigs.cend())
            return {};

        const auto &chCfg = devIt->tluDecDes.chReset;
        if(chCfg.isEmpty())
            return {};
        return boolToCheck(chCfg[0]);
        break;
    }
    default:;
    }

    return {};
}

QVariant Tlu40LvdsModel::limitsData(const QModelIndex &index) const
{
    return std::numeric_limits<quint16>::max();
}

bool Tlu40LvdsModel::setEditRole(const QModelIndex &index, const QVariant &value)
{
    const auto item = getItemByIndex(index);
    auto col = getColumnByIndex(index);
    switch (col) {
    case Tlu40LvdsColumns::COL_SHAPER:
    {
        Q_ASSERT(item->ch == -1);
        Q_ASSERT(config.knownConfigs.contains(item->dev));
        if(!config.knownConfigs.contains(item->dev))
            return false;

        if(item->isLum) {
            config.knownConfigs[item->dev].tlu40lvdsLuminosity.shaperVal = value.toInt();
        } else if(item->isTof) {
            auto &tofCfg = config.knownConfigs[item->dev].tlu40lvdsTofTrigger;
            if(tofCfg.modConfig.size() <= item->mod)
                return false;
            tofCfg.modConfig[item->mod].shaperVal = value.toInt();
        } else {
            return false;
        }

        enqueueConfigSend();
        return true;
    }
    default:;
    }
    return false;
}

bool Tlu40LvdsModel::setCheckStateData(const QModelIndex &index, const QVariant &value)
{
    const auto item = getItemByIndex(index);
    if(item->ch == -1) {
        auto rows = rowCount(index);
        bool ok = true;
        for(int r=0; r<rows; ++r){
            ok &= setCheckStateData(this->index(r,index.column(),index), value);
        }
        return ok;
    } else {
        Q_ASSERT(!item->dev.isNull());
        auto col = getColumnByIndex(index);
        if(col != Tlu40LvdsColumns::COL_EN && col != Tlu40LvdsColumns::COL_DEC_DES_RESET)
            return false;

        Q_ASSERT(config.knownConfigs.contains(item->dev));
        if(!config.knownConfigs.contains(item->dev))
            return false;
        auto &devCfg = config.knownConfigs[item->dev];
        Q_ASSERT(value!=Qt::PartiallyChecked);

        if(col == Tlu40LvdsColumns::COL_EN) {
            if(item->isTof) {
                auto &tofCfg = devCfg.tlu40lvdsTofTrigger;
                if(tofCfg.modConfig.size() <= item->mod)
                    return false;
                tofCfg.modConfig[item->mod].shaperVal = value.toInt();

                auto &chEn = tofCfg.modConfig[item->mod].chEn;
                auto ch = item->ch;
                if(chEn.size() <= ch)
                    ch -= chEn.size();
                if(chEn.size() <= ch)
                    return false;
                chEn[ch] = (value==Qt::Checked) ? true : false;
            } else if(item->isLum) {
                auto &chEn = devCfg.tlu40lvdsLuminosity.chEn;
                if(chEn.size() <= item->ch)
                    return false;
                chEn[item->ch] = (value==Qt::Checked) ? true : false;
            } else
                return false;
        } else {
            auto &chCfg = devCfg.tluDecDes.chReset;
            if(chCfg.size() <= item->ch)
                return false;
            chCfg[item->ch] = (value==Qt::Checked) ? true : false;
        }
        enqueueConfigSend();
        return true;
    }
    return false;
}

Tlu40LvdsModelItem *Tlu40LvdsModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<Tlu40LvdsModelItem*>(index.internalPointer());
}

QVariant Tlu40LvdsModel::getRecursiveState(const QModelIndex &mIndex) const
{
    auto rows = rowCount(mIndex);
    if(rows) {
        QVariant allSt;
        for(int r=0; r<rows; ++r) {
            auto st = getRecursiveState(index(r,0,mIndex));
            if(allSt.isNull())
                allSt = st;
            else if(allSt != st)
                return Qt::PartiallyChecked;
        }
        return allSt.isNull() ? boolToCheck(TluDecDesConfig::DEF_CH_RESET) : allSt;
    } else {
        auto item = getItemByIndex(mIndex);
        const auto &cfgVector = config.knownConfigs.value(item->dev).tluDecDes.chReset;
        return boolToCheck(cfgVector.value(item->ch, TluDecDesConfig::DEF_CH_RESET));
    }
}

QPair<std::optional<quint32>, std::optional<quint32> > Tlu40LvdsModel::getCounterPair(Tlu40LvdsModelItem *item) const
{
    if(item->ch == -1)
        return {};
    QPair<std::optional<quint32>, std::optional<quint32> > ret;
    if(item->isTof) {
        auto it = devStatusTof.constFind(item->dev);
        if(it == devStatusTof.cend())
            return {};
        if(it->modSt.size() <= item->mod)
            return {};
        auto &modSt = it->modSt[item->mod];

        if(modSt.chInputCnt.size() <= item->ch)
            return {};
        ret.first = modSt.chInputCnt[item->ch];
        const auto &chInputCnt = devStatusOffsetTof.value(item->dev)
                                    .modSt.value(item->mod)
                                    .chInputCnt;
        if(chInputCnt.size() > item->ch)
            ret.second = chInputCnt[item->ch];
        return ret;
    } else {
        auto it = devStatusLum.constFind(item->dev);
        if(it == devStatusLum.cend())
            return {};
        if(it->chCnt.size() <= item->ch)
            return {};
        ret.first = it->chCnt.value(item->ch);
        const auto &chCnt = devStatusOffsetLum.value(item->dev).chCnt;
        if(chCnt.size() > item->ch)
            ret.second = chCnt[item->ch];
        return ret;
    }
}
