//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//    Copyright 2014-2015 Nikita Tarasov
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "MscMainWindow.h"
#include "ui_MscMainWindow.h"

#include <QDateTime>
#include <QDebug>
#include <QGridLayout>
#include <QMessageBox>
#include <QSettings>
#include <QSplitter>
#include <QTimer>

#include <cmath>
#include <numeric>

#include "base/DeviceId.h"
#include "base/gui_utils.h"
#include "base/singleton.h"
#include "config/MscConfigDispatcher.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-widgets/Ddr3StreamFifo/Ddr3StatusTableHelper.h"
#include "device-widgets/WrStatus/WrStatusTableHelper.h"
#include "mldiscover/DeviceDispatcher.h"
#include "widgets/DeviceStatusDialog.h"

#include "MscSignalRouter.h"
#include "MscThresholdDialog.h"
#include "MscHistDialog.h"
#include "MscHistWidget.h"
#include "MscHistModel.h"
#include "MscThresholdModel.h"

namespace {
    enum {COL_ID, COL_SERIAL, COL_FIRMWARE, COL_TEMP,
          COLUMN_WR_TIME,
          COL_STATE = COLUMN_WR_TIME+WrStatusTableHelper::COL_TOTAL,
          COL_CE_STATE,
          COL_DDR,
          COL_NUM = COL_DDR+Ddr3StatusTableHelper::COL_TOTAL};
}

MscMainWindow::MscMainWindow(QWidget *parent)
    : ReadoutAppMainWindow(parent)
    , ui(new Ui::MscMainWindow)
    , hist_model(new MscHistModel(this))
    , threshold_model(new MscThresholdModel(this))
    , thresholdDialog(new MscThresholdDialog(this))
    , histDialog(new MscHistDialog(this))
{
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);

    initGui();

    histDialog->setModel(hist_model);
    thresholdDialog->setModel(threshold_model);

    auto &deviceDispatcher = Singleton<DeviceDispatcher>::instance();
    deviceDispatcher.subscribeDevices(ui->m_stream_output);

    auto &configDispatcher = MscConfigDispatcher::instance();
    configDispatcher.subscribeConfig(thresholdDialog);
    configDispatcher.subscribeConfig(ui->streamCfgWidget);
    configDispatcher.subscribeCS(configuration_manager);

    auto &router = MscSignalRouter::instance();
    router.setMainWindow(this);
    router.setHistModel(hist_model);
    router.setTresholdModel(threshold_model);
    router.setMStreamStatisticWidget(ui->m_stream_output);
    router.setHistDialog(histDialog);
    configuration_manager->setDefaultConfig(BaseDeviceAppConfig());
    connect(ui->pushButtonFind, &QPushButton::clicked,
            this, &ReadoutAppMainWindow::actionDiscoverDialog);
}

MscMainWindow::~MscMainWindow()
{
    delete ui;
}

DeviceStatusTableViewFancy *MscMainWindow::getDeviceStatusTableView()
{
    return nullptr; // TODO: return ui->tableWidgetDevices;
}

void MscMainWindow::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    if(newConfig == config)
        return;

    const auto &oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    const auto &newSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    ReadoutAppMainWindow::setConfig(config_key, newConfig);

    const auto toDel = oldSet-newSet;
    const auto toAdd = newSet-oldSet;
    for (const auto &i : toDel) {
        removeDevice(i);
    }
    for (const auto &i : toAdd) {
        int row = getRowByDeviceIndex(i);
        if (row == -1)
            addDevice(i);
    }
    if (curDevice.isNull() || !config.contains(curDevice))
        setCurDevice(config.firstDeviceKey());
    if(!curDevice.isNull())
        addHists(curDevice);

    updateGui();
    updateVisiblePlots();
}

void MscMainWindow::statusUpdated(const MscStatus &status)
{
    emit setStatus(status);

    QTableWidget *tDev = ui->tableWidgetDevices;
    const int row = getRowByDeviceIndex(status.getDeviceIndex());
    if(row == -1)
        return;

    fill_col_state(row, status.connected);
    WrStatusTableHelper::updateDevWrStatus(tDev, COLUMN_WR_TIME,
                                           row, status.wrStatus);
    if(status.ddrStatus.has_value())
        Ddr3StatusTableHelper::updateDevStatus(tDev, COL_DDR,
                                               row, status.ddrStatus.value());

    QString ceText;
    if (status.inputStatus.has_value()) {
        for (bool st : status.inputStatus.value().ceInputStates())
            ceText += st ? 'V' : 'x';
    }
    ui->tableWidgetDevices->item(row, COL_CE_STATE)->setText(ceText);
    ui->tableWidgetDevices->item(row, COL_ID)->setText(status.getDeviceName());

    adjustTableSize(tDev);
}

void MscMainWindow::initGui()
{
    RecursiveWidgetSignalBlocker msb(this);

    ui->mainToolBar->hide();
    ui->comboBoxVisibleHist->addItem("All");
    ui->treeViewHist->setModel(hist_model);
    ui->treeViewThreshold->setModel(threshold_model);

    connect(hist_model, &QAbstractItemModel::modelReset, [=, this]() {
          auto t = ui->treeViewHist;
          for(int c=0; c<hist_model->columnCount(); ++c)
              t->resizeColumnToContents(c);
          t->expandAll();
    });

    connect(threshold_model, &QAbstractItemModel::modelReset, [=, this]() {
          auto t = ui->treeViewThreshold;
          for(int c=0; c<threshold_model->columnCount(); ++c)
              t->resizeColumnToContents(c);
          t->expandAll();
    });

    initTableDevice();
}

std::optional<int> getViewHist()
{
    const QSettings settings;
    bool ok = false;
    int v = settings.value("viewHist").toInt(&ok);
    if (ok)
        return v;
    return {};
}

void setViewHist(const QVariant &v)
{
    QSettings settings;
    settings.setValue("viewHist", v);
}

void MscMainWindow::updateGui()
{
    RecursiveWidgetSignalBlocker msb(this);

    ui->groupBoxStreamConfig->setEnabled(!curDevice.isNull());
    if(curDevice.isNull())
        return;

    auto &dc = config.findDevice(curDevice).mscCore;
    if(ui->comboBoxVisibleHist->count() != dc.hist.plot.size() + 1) {
        for(int i=0; i<dc.hist.plot.size(); ++i) {
            ui->comboBoxVisibleHist->insertItem(0, QString::number(i), i);
        }
    }

    int visibleHistIndex = ui->comboBoxVisibleHist->findData({}); // All index

    if (getViewHist().has_value()) {
        int index = ui->comboBoxVisibleHist->findData(getViewHist().value());
        if(index != -1) {
            visibleHistIndex = index;
        }
    }
    ui->comboBoxVisibleHist->setCurrentIndex(visibleHistIndex);
}

void MscMainWindow::setCurDevice(const DeviceIndex &i)
{
    curDevice = i;
    ui->streamCfgWidget->setCurDevice(curDevice);
}

void MscMainWindow::initTableDevice()
{
    QTableWidget *table = ui->tableWidgetDevices;
    table->setColumnCount(COL_NUM);
    table->setHorizontalHeaderItem(COL_ID, new QTableWidgetItem("Id"));
    table->setHorizontalHeaderItem(COL_SERIAL, new QTableWidgetItem("Serial"));
    table->setHorizontalHeaderItem(COL_STATE, new QTableWidgetItem("State"));
    table->setHorizontalHeaderItem(COL_CE_STATE, new QTableWidgetItem("CE state"));
    table->setHorizontalHeaderItem(COL_FIRMWARE, new QTableWidgetItem("f/w"));
    table->setHorizontalHeaderItem(COL_TEMP, new QTableWidgetItem("Temp"));
    WrStatusTableHelper::setHeaderWrStatus(table, COLUMN_WR_TIME);
    Ddr3StatusTableHelper::setHeaders(table, COL_DDR);
    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    adjustTableSize(table);
}

void MscMainWindow::addDevice(const DeviceIndex &index)
{
    setCurDevice(index);

    QTableWidget *table = ui->tableWidgetDevices;
    int row = table->rowCount();
    table->setRowCount(row+1);
    for(int col=0;col<COL_NUM; ++col) {
        QTableWidgetItem *i = new QTableWidgetItem();
        table->setItem(row, col, i);
    }
    adjustTableSize(table);

    table->item(row, COL_ID)->setText(index.getDeviceIdName());
    table->item(row, COL_SERIAL)->setText(index.getSerialStr());
}

void MscMainWindow::addHists(const DeviceIndex &index)
{
    if (!config.contains(index))
        return;
    const auto &hist = config.findDevice(curDevice).mscCore.hist;
    if(timeHistWidgets.size() == hist.plot.size())
        return;

    QGridLayout *g_lt_1 = ui->gridLayoutTimeHist;
    QGridLayout *g_lt_2 = ui->gridLayoutSpillPlots;

    for(int hid : hist.plot.keys()) {
        MscHistWidget *hist_time = new MscHistWidget(hid, true, this);
        hist_time->setXAxisTitle("dt, us");
        hist_time->setYAxisTitle("hits");
        timeHistWidgets[hid] = hist_time;
        g_lt_1->addWidget(hist_time, hid/4, hid%4);

        MscHistWidget *hist_plot = new MscHistWidget(hid, false, this);
        hist_plot->setXAxisTitle("time, s");
        hist_plot->setYAxisTitle("Rate, Hz");
        plotHistWidgets[hid] = hist_plot;
        g_lt_2->addWidget(hist_plot, hid/4, hid%4);
    }
}

void MscMainWindow::removeDevice(DeviceIndex index)
{
    int row = getRowByDeviceIndex(index);
    if(row == -1)
        return;
    ui->tableWidgetDevices->removeRow(row);
    adjustTableSize(ui->tableWidgetDevices);

    for(int i : timeHistWidgets.keys()){
        timeHistWidgets[i]->~MscHistWidget();
        timeHistWidgets.remove(i);
    }
    for(int i : plotHistWidgets.keys()){
        plotHistWidgets[i]->~MscHistWidget();
        plotHistWidgets.remove(i);
    }
}

void MscMainWindow::fill_col_state(int row, bool connected)
{
    QFont font;
    font.setBold(true);

    if(connected){
        ui->tableWidgetDevices->item(row, COL_STATE)->setText("Connected");
        ui->tableWidgetDevices->item(row, COL_STATE)->setFont(font);
        ui->tableWidgetDevices->item(row, COL_STATE)->setBackground(QBrush(QColor(34,139,34)));
        ui->tableWidgetDevices->item(row, COL_STATE)->setForeground(QBrush(QColor(255,255,255)));
    }
    else{
        ui->tableWidgetDevices->item(row, COL_STATE)->setText("Disconnected");
        ui->tableWidgetDevices->item(row, COL_STATE)->setFont(font);
        ui->tableWidgetDevices->item(row, COL_STATE)->setBackground(QBrush(QColor(230,57,0)));
        ui->tableWidgetDevices->item(row, COL_STATE)->setForeground(QBrush(QColor(255,255,255)));
    }
    adjustTableSize(ui->tableWidgetDevices);
}

DeviceIndex MscMainWindow::getDeviceIndexByRow(int row) const
{
    QTableWidget *table = ui->tableWidgetDevices;
    bool ok;
    quint64 serial = table->item(row, COL_SERIAL)->text()
            .remove('-')
            .toULongLong(&ok, 16);
    if(ok){
        DeviceIndex index(DeviceId::DEVICE_ID_MSC16VE_E, serial);
        return index;
    }
    return {};
}

int MscMainWindow::getRowByDeviceIndex(DeviceIndex index) const
{
    QTableWidget *table = ui->tableWidgetDevices;
    for(int row=0; row<table->rowCount(); ++row){
        if(getDeviceIndexByRow(row) == index)
            return row;
    }
    return -1;
}

void MscMainWindow::adjustTableSize(QTableWidget *table)
{
    int width=table->verticalHeader()->width();
    for(int i=0;i<table->columnCount();++i){
        if(table->isColumnHidden(i)) continue;
        width += table->columnWidth(i);
    }
    width += 5;

    int height=table->horizontalHeader()->height();
    for(int i=0;i<table->rowCount();++i){
        if(table->isRowHidden(i)) continue;
        height += table->rowHeight(i);
    }
    height += 5;

    QRect geo = table->geometry();
    geo.setWidth(width);
    geo.setHeight(height);
    table->setMinimumWidth(width);
    table->setMinimumHeight(height);
    table->setGeometry(geo);
}

void MscMainWindow::updateVisiblePlots()
{
    if(curDevice.isNull())
        return;
    const auto &hist = config.findDevice(curDevice).mscCore.hist;

    auto shown = [ &hist](int hId) {
        return (!getViewHist().has_value() ||  getViewHist().value() == hId)
                && hist.plot[hId].enable;
    };

    for(int hId : timeHistWidgets.keys()){
        MscHistWidget *time_hist = timeHistWidgets.value(hId, nullptr);
        if(time_hist){
            time_hist->setVisible(shown(hId));
        }
    }

    for(int hId : plotHistWidgets.keys()){
        MscHistWidget *plot = plotHistWidgets.value(hId, nullptr);
        if(plot)
            plotHistWidgets[hId]->setVisible(shown(hId));
    }
}

void MscMainWindow::deviceConnected(DeviceIndex index, QString fwStr)
{
    int row = getRowByDeviceIndex(index);
    if(row==-1)
        return;
    fill_col_state(row, true);
    ui->tableWidgetDevices->item(row, COL_FIRMWARE)->setText(fwStr);
}

void MscMainWindow::deviceDisconnected(DeviceIndex index)
{
    int row = getRowByDeviceIndex(index);
    if(row==-1)
        return;
    fill_col_state(row, false);
}

void MscMainWindow::histUpdated(const DeviceIndex &index, const MscDevHists &h)
{
    Q_UNUSED(index)
    for(auto chIt=h.hists.cbegin(); chIt!=h.hists.cend(); ++chIt) {
        auto hId = chIt.key();
        MscHistWidget *time_hist = timeHistWidgets.value(hId, nullptr);
        if(time_hist)
            time_hist->updateHistogram(chIt->time);

        MscHistWidget *plot = plotHistWidgets.value(hId, nullptr);
        if(plot)
            plot->updateHistogram(chIt->plot);
    }
}

//=== SLOTS FROM UI ===

void MscMainWindow::on_actionThreshold_dialog_triggered()
{
    thresholdDialog->show();
}

void MscMainWindow::on_actionHist_triggered()
{
    histDialog->show();
}

void MscMainWindow::on_comboBoxVisibleHist_currentIndexChanged(int index)
{
    auto d = ui->comboBoxVisibleHist->itemData(index);
    setViewHist(d);

    updateVisiblePlots();
    emit configChanged(config);
}

void MscMainWindow::saveAppGuiSettings(QSettings &s) const
{
    BaseMainWindow::saveAppGuiSettings(s);

    s.beginGroup("MainWindow");
    s.setValue("geometry", saveGeometry());
    s.setValue("windowState", saveState());
    s.setValue("tabIndex", ui->tabWidgetGr->currentIndex());
    s.endGroup();
}

void MscMainWindow::loadAppGuiSettings(const QSettings &s)
{
    BaseMainWindow::loadAppGuiSettings(s);

    QSettings settings;
    const QString prefix = "MainWindow/";
    restoreGeometry(settings.value(prefix+"geometry").toByteArray());
    restoreState(settings.value(prefix+"windowState").toByteArray());
    int tabCurrentIndex = settings.value(prefix+"tabIndex", 0).toInt();

    ui->tabWidgetGr->setCurrentIndex(tabCurrentIndex);
    adjustTableSize(ui->tableWidgetDevices);
}

// FIXME: duplicated code
void MscMainWindow::on_tableWidgetDevices_currentItemChanged(QTableWidgetItem *current, QTableWidgetItem *previous)
{
    if (!current)
        return;
    auto deviceIndex = getDeviceIndexByRow(current->row());
    deviceStatusDialog->setIndex(deviceIndex);
}

void MscMainWindow::on_tableWidgetDevices_itemDoubleClicked(QTableWidgetItem *item)
{
    if (!item)
        return;

    auto deviceIndex = getDeviceIndexByRow(item->row());
    deviceStatusDialog->setIndex(deviceIndex);
    deviceStatusDialog->show();
}
