/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RcConfigDispatcher.h"

#include <optional>
#include <QLoggingCategory>

#include "RcConfig.h"
#include "base/DeviceIndexEnabled.h"
#include "base/util.h"
#include "lib-common/BaseDeviceAppConfig.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(config)
Q_LOGGING_CATEGORY(config, "afi.rc.config")
}

class RcConfigDispatcherPrivate
{
public:
    std::optional<RcConfig> cache;
};

RcConfigDispatcher::RcConfigDispatcher(QObject *parent) :
    BaseConfigDispatcher(parent),
    d_ptr(new RcConfigDispatcherPrivate)
{
}

RcConfigDispatcher::~RcConfigDispatcher()
{
    delete d_ptr;
}

DeviceIndexEnabledMap RcConfigDispatcher::getDeviceIndexEnabledMap() const
{
    Q_D(const RcConfigDispatcher);
    if (!d->cache.has_value())
        return {};
    const auto &cfg = d->cache.value();
    const auto &deviceIndex = cfg.rcTrigConfig.index;
    DeviceIndexEnabledMap map;
    map.insert(deviceIndex, {deviceIndex, true});
    return map;
}

BaseDeviceAppConfig &RcConfigDispatcher::getAppConfigCache()
{
    static BaseDeviceAppConfig tmp; // FIXME: temporary stub
    return tmp;
}

void RcConfigDispatcher::sendCache()
{
    Q_D(const RcConfigDispatcher);
    if (!d->cache.has_value())
        return;
    BaseConfigDispatcher::sendCache();
    const auto &cfg = d->cache.value();
    emit configChanged(cfg);
    emit dlnpCalibrationTriggerConfigChanged(cfg.rcTrigConfig.dlnpCalibrationTrigger);
    emit slowControlConfigChanged(cfg.scConfig);
    emit spillCfgChanged(cfg.rcTrigConfig.spillControl);
    emit rcClientsConfigChanged(cfg.clients);
}

void RcConfigDispatcher::setConfig(const RcConfig &newConfig)
{
    Q_D(RcConfigDispatcher);
    if (d->cache.has_value()) {
        const auto &cache = d->cache.value();
        if (newConfig == cache)
            return;
        if (newConfig.rcTrigConfig.index.isNull()) {
            if (!cache.rcTrigConfig.index.isNull()) {
                qDebug(config) << "device removed:" << cache.rcTrigConfig.index.getIdent();
                qDebug(config) << "configChanged by" << qObjectAncestryStr(sender());
            }
        }
    }
    d->cache = newConfig;
    sendCache();
}

void RcConfigDispatcher::setRcClientsConfig(const RcClientConfigMap &newConfig)
{
    Q_D(RcConfigDispatcher);
    Q_ASSERT(d->cache.has_value());
    if (d->cache.has_value()) {
        auto &cache = d->cache.value();
        if (newConfig == cache.clients)
            return;
        cache.clients = newConfig;
        emit configChanged(d->cache.value());
        emit rcClientsConfigChanged(newConfig);
    }
}

void RcConfigDispatcher::setSlowControlConfig(const SlowControlConfig &newConfig)
{
    Q_D(RcConfigDispatcher);
    Q_ASSERT(d->cache.has_value());
    if (d->cache.has_value()) {
        auto &cache = d->cache.value();
        if (newConfig == cache.scConfig)
            return;
        cache.scConfig = newConfig;
        emit configChanged(d->cache.value());
        emit slowControlConfigChanged(newConfig);
    }
}

void RcConfigDispatcher::setSpillCfg(const SpillConfig &cfg)
{
    Q_D(RcConfigDispatcher);
    Q_ASSERT(d->cache.has_value());
    if (d->cache.has_value()) {
        auto &cache = d->cache.value();
        if (cache.rcTrigConfig.spillControl == cfg)
            return;
        cache.rcTrigConfig.spillControl = cfg;
        emit configChanged(d->cache.value());
        emit spillCfgChanged(cfg);
    }
}

void RcConfigDispatcher::setDlnpCalibrationTriggerConfig(const DlnpCalibrationTriggerConfig &cfg)
{
    Q_D(RcConfigDispatcher);
    Q_ASSERT(d->cache.has_value());
    bool changed = false;
    if (d->cache.has_value()) {
        auto &cache = d->cache.value();
        if (cache.rcTrigConfig.dlnpCalibrationTrigger == cfg)
            return;
        cache.rcTrigConfig.dlnpCalibrationTrigger = cfg;
        changed = true;
    }

    BaseConfigDispatcher::setDlnpCalibrationTriggerConfig(cfg);

    if(changed) {
        emit configChanged(d->cache.value());
        emit dlnpCalibrationTriggerConfigChanged(cfg);
    }
}
