/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TrigDeviceController.h"
#include "TrigDeviceController_p.h"

#include "RcConfig.h"
#include "RcSignalRouter.h"
#include "TrigDevice.h"
#include "TrigStatus.h"
#include "base/DeviceIndexEnabled.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "run-control-fsm/TrigDeviceMgrFsm.h"

static DeviceIndexSet getDeviceIndexEnabledSetFromMap(const DeviceIndexEnabledMap &map)
{
    DeviceIndexSet set;
    for (auto i : map)
        if (i.enabled)
            set.insert(i.index);
    return set;
}

static DeviceIndexEnabledMap deviceListFromConfig(const RcConfig &newConfig)
{
    DeviceIndexEnabledMap map;
    auto index = newConfig.rcTrigConfig.index;
    if (index.isNull())
        return map;
    map[index] = DeviceIndexEnabled(index, true);
    return map;
}



TrigDeviceController::TrigDeviceController(QObject *parent) :
    AbstractDeviceController(parent),
    d_ptr(new TrigDeviceControllerPrivate(this))
{
}

QSet<DeviceId> TrigDeviceController::getDeiveIdFilter()
{
    return {DEVICE_ID_U40VE_RC, DEVICE_ID_UT24VE};
}

QVariant TrigDeviceController::getStopEvNum() const
{
    Q_D(const TrigDeviceController);
    return d->stopEvNum;
}

void TrigDeviceController::setMegaMasterFlag(bool mm)
{
    Q_D(TrigDeviceController);
    d->megamaster_flg = mm;
}

void TrigDeviceController::setConfig(const RcConfig &newConfig)
{
    Q_D(TrigDeviceController);

    if (newConfig == d->config)
        return;

    const auto oldDeviceList = deviceListFromConfig(d->config);
    const auto newDeviceList = deviceListFromConfig(newConfig);

    bool rcCfgChanged = d->config.rcTrigConfig != newConfig.rcTrigConfig;
    d->config = newConfig;

    // update device list from config
    if (oldDeviceList != newDeviceList)
        emit deviceListChanged(newDeviceList);
    if(rcCfgChanged)
        d->onTrigDevConfigUpdate();
}

void TrigDeviceController::setDeviceList(const DeviceIndexEnabledMap &devices)
{
    static DeviceIndexEnabledMap old_devices;

    if (devices == old_devices)
        return;

    const auto old_set = getDeviceIndexEnabledSetFromMap(old_devices);
    const auto new_set = getDeviceIndexEnabledSetFromMap(devices);
    const auto toRemove = old_set - new_set;
    const auto toAdd = new_set - old_set;

    for (auto i : toRemove)
        removeTrigDevice(i);

    for (auto i : toAdd)
        addTrigDevice(i);

    old_devices = devices;
    Q_D(TrigDeviceController);
    d->fsm->setDeviceList(devices);
    d->onTrigDevConfigUpdate();
}

void TrigDeviceController::onRunStart(const RcRunId &, const QVariant &startEvNum)
{
    Q_D(TrigDeviceController);
    const auto &curSt = d->fsm->getState();
    d->startEvNum = startEvNum;
    if(d->megamaster_flg) {
        //In megamaster we need just set startEvNum, for future usage
        return;
    }
    if(curSt==FsmState::CloseGlobTrig || curSt==FsmState::OpenGlobTrig) {
        //TODO: explude somehow this slot invokation in megamaster device
        qFatal("add debug here"); // todel
        return;
    }

    if(curSt==FsmState::Ready) {
        d->fsm->setRequestState(FsmState::Starting);
    } else {
        qWarning() << QString("Got %1 in incorrect state %2 (megamaster_flg=%3)")
                      .arg(__PRETTY_FUNCTION__).arg(fsmStateToString(curSt))
                      .arg(d->megamaster_flg);
        d->fsm->setRequestState(FsmState::Error);
    }}

void TrigDeviceController::onRunStop(const QVariant &evNum, bool force)
{
    Q_UNUSED(evNum)
    Q_D(TrigDeviceController);
    const auto &curSt = d->fsm->getState();
    if(force) {
        // Not correct but simpliest way
        if(curSt != FsmState::Ready)
            d->fsm->setRequestState(FsmState::Error);
        return;
    }
    if(d->megamaster_flg)
        return;
    if(curSt==FsmState::Ready) {
        //TODO: explude somehow this slot invokation in megamaster device
        qFatal("add debug here"); // todel
        return;
    }
    d->fsm->setRequestState(FsmState::Stopping);
}

void TrigDeviceController::onPrepareStop()
{
    Q_D(TrigDeviceController);
    Q_ASSERT(d->megamaster_flg);
    d->fsm->setRequestState(FsmState::Stopping);
}

void TrigDeviceController::onCloseGlobTrig()
{
    Q_D(TrigDeviceController);
    Q_ASSERT(d->megamaster_flg);
    d->fsm->setRequestState(d->fsm->getState()==FsmState::Ready ? FsmState::Starting : FsmState::Error);
}

void TrigDeviceController::onOpenGlobTrig()
{
    Q_D(TrigDeviceController);
    Q_ASSERT(d->megamaster_flg);
    d->onOpenGlobTrig();
}

void TrigDeviceController::addTrigDevice(const DeviceIndex &index)
{
    Q_D(TrigDeviceController);

    bool existed = (index == d->config.rcTrigConfig.index);
    d->config.rcTrigConfig.index = index;
    if (!existed)
        emit configChanged(d->config);

    if (d->trigThreads.contains(index))
        return;

    auto dev = new TrigDevice(index, nullptr);
    d->trigThreads.insert(index, TrigDeviceProxy(dev));
    connect(dev, &TrigDevice::updateStatus, d, &TrigDeviceControllerPrivate::setTrigStatus);
    connect(dev, &TrigDevice::deviceConfigWritten, d, &TrigDeviceControllerPrivate::setupWritten);
    connect(d, &TrigDeviceControllerPrivate::writeSetup, dev, &TrigDevice::writeSetup);
    connect(this, &TrigDeviceController::printFsmInput, d, &TrigDeviceControllerPrivate::printFsmInput);
    connect(d, &TrigDeviceControllerPrivate::preRunReset, dev, &TrigDevice::preRunReset);

    RcSignalRouter::instance().connect_TrigDevice(dev);
    // device should have setHostAddress set up at this point,
    // subscribe sends first signal here
    deviceDispatcher->subscribeHostAddressUpdates(index, dev);

    // emit deviceAdded(index); // to model
    addMStreamDump(index);
}

void TrigDeviceController::addMStreamDump(const DeviceIndex &index)
{
    Q_D(TrigDeviceController);

    MStreamDumpParameters dump_parameters;
    dump_parameters.deviceIndex = index;
    dump_parameters.bigFragId = true;
//    dump_parameters.hwBufSize = devStatus.mstreamHwBufSize;
//    dump_parameters.ackSizeLimit = devStatus.mstreamMultiAck ? devStatus.mstreamHwBufSize : 1;
    dump_parameters.preRunInit();

    auto mstream = new MStreamDump(dump_parameters, false);
    connect(mstream, &MStreamDump::closeProgram, this, &TrigDeviceController::closeProgram);
    connect(mstream, &MStreamDump::updateProgramDescription,
            this, &TrigDeviceController::updateProgramDescription);
//    connect(mstream, &MStreamDump::stat_updated,
//            this, &TrigDeviceController::msDumpStatUpdated);

    LldpDiscovery::subscribe(mstream, {index.getDevId()});
//    connect(mstream, &MStreamDump::stat_from_receiver_updated,
//            this, &TrigDeviceController::stat_from_receiver_updated);

//    connect(this, &TrigDeviceController::msdump_reset, mstream, &MStreamDump::msdump_reset);
//    connect(mstream, &MStreamDump::reseted, this, &TrigDeviceController::msdumpReseted);
    connect(d, &TrigDeviceControllerPrivate::sendJsonBlock,
            mstream, &MStreamDump::sendJsonBlock);

    mstream->setObjectName("MSD_"+index.getSerialStr());
    d->msDumpThreads.insert(index, MStreamDumpProxy(mstream));
}

void TrigDeviceController::removeTrigDevice(const DeviceIndex &index)
{
    Q_D(TrigDeviceController);

    d->trigThreads.remove(index);
    d->trigStatus.remove(index);
    d->config.rcTrigConfig.index = DeviceIndex();
    emit configChanged(d->config);
    // emit deviceRemoved(index); // to model
}
