#include "TrcDevice.h"

#include <QDateTime>
#include <QElapsedTimer>
#include <QFile>
#include <QTextStream>
#include <QVector>
#include <math.h>

#include "LegacyDeviceModule.h"
#include "TrcCore/PhysicalTrigger/PhysicalTriggerModule.h"
#include "TrcCore/TrcErc/TrcErcModule.h"
#include "TrcRegs.h"
#include "regio/regop.h"

TrcDevice::TrcDevice(const DeviceIndex &index, QObject *parent)
    : ModularMregDevice(index, status, parent)
    , status(index)
    , legacyDevice (createModule<LegacyDeviceModule>())
{
    qRegisterMetaType<ModularDeviceConfig>();
    qRegisterMetaType<BaseDevCountersMap>();
    setObjectName(getIdent());
}

bool TrcDevice::writeConfig()
{
    bool ok = true;
    if(!phyTrigMod)
        return false;
    ok &= phyTrigMod->writeConfig(config.trcCore.phyTrig);
    if(evReadoutMod)
        ok &= evReadoutMod->writeConfig(config.trcCore.phyTrig.oscEn);

    return ok & legacyDevice->regWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg());
}

void TrcDevice::createModules()
{
    ModularMregDevice::createModules();
    phyTrigMod = createSingleModule<PhysicalTriggerModule>();
    evReadoutMod = createSingleModule<TrcErcModule>();
}

void TrcDevice::resetAll()
{
    if(!isConnected())
        return;
    RegOpVector r;
    r.RegWrite16(REG_DEVICE_CTRL, 1);
    r.RegWrite16(REG_DEVICE_CTRL, 0);
    legacyDevice->regOpExecRebased(r);
}

void TrcDevice::softClear()
{
    if(!isConnected())
        return;

    if(phyTrigMod)
        phyTrigMod->softClear();
}

bool TrcDevice::readStatus()
{
    return
        ModularMregDevice::readStatus() &&
        readTrcStatus();
}

bool TrcDevice::updateCounters()
{
    RegOpVector r;
    r.RegWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg());
    r.RegWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg() | REG_DEVICE_CTRL_BIT_LOCK_COUNTERS);
    return regOpExec(r);
}

bool TrcDevice::readTrcStatus()
{
    if (!updateCounters())
        return {};
    if (!phyTrigMod)
        return {};

    if (!phyTrigMod->readStatus(&status.phyTrig))
        return {};
    if(evReadoutMod) {
        quint64 globEvNum;
        if (!evReadoutMod->readGlobEvNum(&globEvNum))
            return {};
        status.globEvNum = globEvNum;
    }

    emit statusUpdated(status);
    emit countersUpdated(status.phyTrig.rcCounters);
    return true;
}

void TrcDevice::on_deviceConnected()
{
    ModularMregDevice::on_deviceConnected();
    resetAll();

    if(phyTrigMod) {
        auto bp = phyTrigMod->getBuildParam();
        if(bp)
            emit deviceInfoUpdated(getDeviceIndex(), bp->auxChNum, bp->phyChNum);
    }
}

void TrcDevice::on_deviceDisconnected()
{
    ModularMregDevice::on_deviceDisconnected();
    if(phyTrigMod) {
        phyTrigMod->deinit();
        phyTrigMod.clear();
    }
    if(evReadoutMod) {
        evReadoutMod->deinit();
        evReadoutMod.clear();
    }
}

quint16 TrcDevice::getDeviceCtrlReg() const
{
    return configCache.trcCore.phyTrig.trig.trigOpen ? REG_DEVICE_CTRL_BIT_RUN : 0;
}

void TrcDevice::preRunReset(const QVariant &startEvNum)
{
    ModularMregDevice::preRunActions(startEvNum);

    clearStatistics();
    bool ok;
    quint64 evNum = startEvNum.toULongLong(&ok);
    if(!ok)
        evNum = 1;
    if(evReadoutMod)
        evReadoutMod->setGlobEvNum(evNum);
}

void TrcDevice::clearStatistics()
{
    RegOpVector r;
    r.RegWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg() | REG_DEVICE_CTRL_BIT_SOFTCLEAR);
    r.RegWrite16(REG_DEVICE_CTRL, getDeviceCtrlReg());
    legacyDevice->regOpExecRebased(r);
}
