#include "HptdcVoltageTest.h"

#include <QCoreApplication>
#include <QDebug>
#include <QFile>
#include <QFileInfo>
//#include <QHash>
#include <QTextStream>
#include <QSettings>
#include <QStandardPaths>

#include "histogramm/AnalyzerStat.h"
#include "histogramm/HistMngrCfg.h"
#include "histogramm/Histogramm.h"

void HptdcVoltageTest::print(QString dataFile,
                             const AnalyzerStat &lastAnalStat,
                             QHash<HistKey, Histogramm> hc,
                             HistMngrCfg *histCfg,
                             const InlResults &inlRes)
{
    updatePowerInfo(dataFile);
    DeviceIndex stopIndex = histCfg->stop_dev_index;
    quint16 stopCh = histCfg->stop_ch;
    auto stopHits = lastAnalStat.dev_stat.value(stopIndex).dev_ch_hits.value(stopCh).tdc_hits;

    QFile file;
    QString filePath = QStandardPaths::locate(QStandardPaths::HomeLocation, QString(), QStandardPaths::LocateDirectory)
            .append(QString("volTest_%1.txt").arg(source_fn));
    file.setFileName(filePath);
    if(!file.open(QIODevice::WriteOnly | QIODevice::Text | QIODevice::Append)) {
        qCritical() << "Failed to open file for INL-check result:"
                    << filePath << file.errorString();
        return;
    }
    QTextStream textStream(&file);
    QStringList line;

    textStream << "File\t" << dataFile << "\n";
    textStream << "EvParsed\t" << lastAnalStat.eventDecoded << "\n";
    textStream << "\n";
    textStream << "\n";

    bool printHeader=true;

    auto key_list = hc.keys();
    std::sort(key_list.begin(), key_list.end());
    DeviceIndex cur_dev_index;
    for(HistKey hist_key : key_list) {
        if(hist_key.hist_id == HistId::inlCheck){
            DeviceIndex index = hist_key.index;
            if(index != cur_dev_index) {
                const auto &errorMap = lastAnalStat.dev_stat[index].err_cnt;
                for(quint32 errWord : errorMap.keys()) {
                      textStream << "HPTDC_ERR\t" << hex << errWord << dec;
                      quint32 cnt=0;
                      for(auto range : errorMap[errWord]) {
                          textStream << "\t"<< range.first << "\t"<< range.second ;
                          cnt += range.second+1-range.first;
                      }
                      textStream << "\tTot:" << cnt << "\n";
                  }

                const auto &minTemp = inlRes[index].minTemp;
                const auto &maxTemp = inlRes[index].maxTemp;
                for(int id : minTemp.keys()){
                    textStream << QString("Temp_%1\t").arg(id) << minTemp[id] << "\t" << maxTemp[id] << "\n";
                }
                const auto &keys = min[index].sensors.keys();
                for (auto i : keys) {
                    if (! i.contains("TDC", Qt::CaseInsensitive)) continue;
                    textStream << i << ", mV\t" << min[index].sensors[i].vol << "\t" << max[index].sensors[i].vol << "\n";
                }
                cur_dev_index = index;
            }

            qint64 chHits = lastAnalStat.dev_stat.value(index).dev_ch_hits.value(hist_key.ch).tdc_hits;
            const Histogramm &hist = hc[hist_key];
            if(hist.getEntries()==0) continue;

            DistributionParams histDist = hist.getDistribution();

            if(printHeader) {
                printHeader = false;
                line << "stopCh"
                     << "ch"
                     << "cnt"
                     << "mean ns"
                     << "RMS ps"
                     << "min, ns"
                     << "max, ns"
                     << "max-min, ps"
                     << "d_hits";
                textStream << line.join('\t') << '\n';
                line.clear();
            }
            line << QString::number(stopCh)
                 << QString::number(hist_key.ch)
                 << QString::number(histDist.cnt)
                 << QString::number(histDist.mean)
                 << QString::number(histDist.rms*1e3, 'f', 0)
                 << QString::number(histDist.min)
                 << QString::number(histDist.max)
                 << QString::number((histDist.max-histDist.min)*1e3, 'f', 0)
                 << QString::number(chHits-stopHits);
            textStream << line.join('\t') << '\n';
            line.clear();
        }
    }
}

void HptdcVoltageTest::updatePowerInfo(QString fn)
{
    min.clear();
    max.clear();
    QFileInfo fi(fn);
    if(!fi.exists()) return;
    source_fn = fi.fileName().remove(".data");
    qInfo() << fn;
    QString crStr = fn.mid(fn.lastIndexOf("mpd")+4, 15);

    QDateTime crDate = QDateTime::fromString(crStr, "yyyyMMdd_HHmmss");
    updatePowerInfo(crDate, fi.lastModified());
}

void HptdcVoltageTest::updatePowerInfo(QDateTime start, QDateTime stop)
{
    qInfo() << start.toString();
    qInfo() << stop.toString();
    qint64 startSec = start.toSecsSinceEpoch();
    qint64 stopSec = stop.toSecsSinceEpoch();
    min.clear();
    max.clear();
    QSettings s(QCoreApplication::organizationName(), "tdc2_power");
    for(QString group : s.childGroups()) {
        bool ok1,ok2;
        QStringList strList = group.split("_");
        if(strList.size()!=2) continue;
        DeviceIndex i(static_cast<DeviceId>(strList.first().toUShort(&ok1, 16)),
                      strList.last().toUInt(&ok2, 16));
        if(i.isNull() || !ok1 || !ok2) continue;
        s.beginGroup(group);
        for(QString ts : s.childGroups()) {
            qint64 sec = ts.toLongLong(&ok1);
            if(!ok1 || sec<=startSec || sec>=stopSec) continue;
            s.beginGroup(ts);
            for(QString param : s.childKeys()) {
                bool cur = param.startsWith("cur");
                quint16 id = param.right(1).toUShort(&ok1);
                quint16 val = s.value(param).toString().toUShort(&ok2);
                if(!ok1 || !ok2) continue;
                /* TODO
                auto &mapMin = cur ? min[i].cur : min[i].vol;
                if(mapMin.contains(id))
                    mapMin[id] = qMin(mapMin[id], val);
                else
                    mapMin[id] = val;
                auto &mapMax = cur ? max[i].cur : max[i].vol;
                if(mapMax.contains(id))
                    mapMax[id] = qMax(mapMin[id], val);
                else
                    mapMax[id] = val;
                */
            }
            s.endGroup();
        }
        s.endGroup();
    }
}
