#include "BmcStatus.h"
#include "base/DeviceId.h"
#include "bmc/fpga_mcu_regs_addr.h"

// 14-bit signed value, 1/16 degrees Celsius
double decodeTemp_v1(qint32 val)
{
    val &= 0x1FFF;
    val = (val << 19) >> 19;
    return val / 32.0;
}
double decodeTemp_v2(quint16 val)
{
    double v = static_cast<qint16>(val);
    return v / 32.0;
}

int BmcStatus::tempCount() const
{
    switch (deviceId) {
    case DEVICE_ID_ADC64VE2:
        return 3;
    case DEVICE_ID_TDC64VHLE:
        return 2;
    case DEVICE_ID_TDC64VLE:
        return 4;
    case DEVICE_ID_TDC72VXS:
        return 4;
    case DEVICE_ID_TQDC16:
    case DEVICE_ID_TQDC16VS_E:
        if (pcb_ver < fw_version_t(2, 0, 0))
            return 2;
        else
            return 6;
    case DEVICE_ID_TTVXS:
    case DEVICE_ID_CRU16WR:
        return 3;
    default:;
    }
    return 0;
}

QString BmcStatus::tempLabel(int index) const
{
    switch (deviceId) {
    case DEVICE_ID_ADC64VE2:
        if (index == 0) return "Board";
        if (index == 1) return "FPGA";
        if (index == 2) return "VCXO";
        break;
    case DEVICE_ID_TDC64VHLE:
        if (index == 0) return "TDC-A";
        if (index == 1) return "TDC-B";
        break;
    case DEVICE_ID_TDC64VLE:
        if (index == 0) return "Board";
        if (index == 1) return "FPGA";
        if (index == 2) return "VCXO";
        if (index == 3) return "TDC";
        break;
    case DEVICE_ID_TDC72VXS:
        if (index == 0) return "PLL";
        if (index == 1) return "TDC-A";
        if (index == 2) return "TDC-B";
        if (index == 3) return "TDC-C";
        break;
    case DEVICE_ID_TQDC16:
    case DEVICE_ID_TQDC16VS_E:
        if (pcb_ver < fw_version_t(2, 0, 0)) {
            if (index == 0) return "PLL";
            if (index == 1) return "Analog";
        } else {
            if (index == 0) return "TMP421";
            if (index == 1) return "FPGA";
            if (index == 2) return "VCXO";
            if (index == 3) return "TDC-A";
//            if (index == 4) return "TDC-B";
//            if (index == 5) return "Analog";
        }
        break;
    case DEVICE_ID_TTVXS:
    case DEVICE_ID_CRU16WR:
        if (index == 0) return "Board";
        if (index == 1) return "FPGA";
        if (index == 2) return "VCXO";
        break;
    default:;
    }
    return {};
}

int BmcStatus::powerCount() const
{
    switch (deviceId) {
    case DEVICE_ID_ADC64VE2:
        return 1;
    case DEVICE_ID_TDC64VHLE:
        if (pcb_ver >= fw_version_t(2, 0, 0))
            return 6;
        else
            return 0;
    case DEVICE_ID_TDC64VLE:
        return 3;
    case DEVICE_ID_TDC72VXS:
        if (pcb_ver >= fw_version_t(4, 0, 0))
            return 5;
        else
            return 0;
    case DEVICE_ID_TQDC16:
    case DEVICE_ID_TQDC16VS_E:
        if (pcb_ver >= fw_version_t(2, 0, 0))
            return 3;
        else
            return 0;
    case DEVICE_ID_TTVXS:
    case DEVICE_ID_CRU16WR:
        return 2;
    default:;
    }
    return 0;
}

QString BmcStatus::powerLabel(int index) const
{
    switch (deviceId) {
    case DEVICE_ID_ADC64VE2:
        if (index == 0) return "Backplane 5V";
    case DEVICE_ID_TDC64VHLE:
    case DEVICE_ID_TDC72VXS:
        if (index == 0) return "TDC-A";
        if (index == 1) return "TDC-B";
        if (index == 2) return "TDC-C";
        if (index == 3) return "Backplane 5V";
        if (index == 4) return "Backplane 3.3V";
        if (index == 5) return "TDC-D";
        break;
    case DEVICE_ID_TDC64VLE:
        if (index == 0) return "TDC-A";
        if (index == 1) return "TDC-B";
        if (index == 2) return "Backplane 5V";
        break;
    case DEVICE_ID_TQDC16:
    case DEVICE_ID_TQDC16VS_E:
        if (index == 0) return "TDC";
        if (index == 1) return "Backplane 5V";
        if (index == 2) return "Backplane 3.3V";
        break;
    case DEVICE_ID_TTVXS:
    case DEVICE_ID_CRU16WR:
        if (index == 0) return "Backplane 5V";
        if (index == 1) return "Backplane 3.3V";
        break;
    default:;
    }
    return {};
}

static const quint16 v1_temp_sensor_error_value = 0xFFFF;
static const quint16 v2_temp_sensor_error_value = 0x8000;

BmcStatus::BmcStatus(const DeviceId &deviceId, const fw_version_t &pcb_ver, const QMap<quint16, quint16> &regs)
    : deviceId(deviceId), pcb_ver(pcb_ver)
{
    if (deviceId == DEVICE_ID_UNKNOWN)
        return;
    bool v1 = (regs[REG_BMC_VERSION] == 0 && regs[REG_BMC_REVISION] == 0);
    if (v1) {
        mcuVer = fw_version_t(regs[REG_BMC_V1_VERSION], regs[REG_BMC_V1_REVISION]);
        quint16 onboard_pll_status = regs[REG_BMC_V1_ONBOARD_PLL_STATUS];
        pll_locked = (onboard_pll_status & 0x08) == 0;
        pll_source = (onboard_pll_status & 0x07) >> 1;
        pllUnlockCnt = regs[REG_BMC_V1_PLL_UNLOCK_CNT]; // & 0x0FFF
        for (int i=0; i<std::min(4, tempCount()); i++) {
            quint16 rawTemp = regs[quint16(REG_BMC_V1_TEMP_1 + i)];
            if (rawTemp != v1_temp_sensor_error_value)
                temp[tempLabel(i)] = decodeTemp_v1(rawTemp);
        }
    } else {
        mcuVer = fw_version_t(regs[REG_BMC_VERSION], regs[REG_BMC_REVISION]);
        //deviceId = static_cast<DeviceId>(regs[REG_BMC_DEV_ID]);
        quint16 onboard_pll_status = regs[REG_BMC_ONBOARD_PLL_STATUS];
        pll_source = (onboard_pll_status & 0x07) >> 1;
        pllUnlockCnt = regs[REG_BMC_PLL_UNLOCK_CNT];
        overall_status = static_cast<SensorStatus>(regs[REG_BMC_SYSTEM_STATUS] & 0x3);
        power_status   = static_cast<SensorStatus>(regs[REG_BMC_POWER_STATUS] & 0x3);
        pll_status     = static_cast<SensorStatus>(regs[REG_BMC_PLL_STATUS] & 0x3);
        for (int i=0; i<std::min(4, tempCount()); i++) {
            quint16 rawTemp = regs[quint16(REG_BMC_TEMP_1 + i)];
            if (rawTemp != v2_temp_sensor_error_value)
                temp[tempLabel(i)] = decodeTemp_v2(rawTemp);
        }
        for (int i=0; i<powerCount(); i++) {
            PowerSensor sensor;
            sensor.vol = regs[quint16(REG_BMC_VOL_TDC_A + 2*i)];
            sensor.cur = regs[quint16(REG_BMC_VOL_TDC_A + 2*i+1)];
            power.sensors.insert(powerLabel(i), sensor);
        }
    }
}

QString BmcPower::getSummary(bool withCurrent, QString filter) const
{
    QStringList list;
    for (auto it = sensors.begin(); it != sensors.end(); ++it) {
        auto name = it.key();
        if (!filter.isNull() && ! name.contains(filter, Qt::CaseInsensitive)) continue;
        const auto &sensor = it.value();
        QString s = QString("%1: %2V")
                        .arg(name).arg(sensor.vol * 1e-3, 0, 'g', 3);
        if(withCurrent)
            s += QString(" %3A").arg(sensor.cur * 1e-3, 0, 'g', 3);
        list << s;
    }
    return list.join('\n');
}

bool BmcPower::isCriticalCurrent() const
{
    if(setupInProgress) return false;
    for (auto it = sensors.begin(); it != sensors.end(); ++it) {
        auto name = it.key();
        if (! name.contains("TDC", Qt::CaseInsensitive)) continue;
        const auto &sensor = it.value();
        if (sensor.cur > 1500)
            return true;
    }
    return false;
}

QMap<QString, SensorStatus> BmcStatus::getSubsystemStatus() const
{
    QMap<QString, SensorStatus> statusMap;
    statusMap["overall"] = get_overall_status();
    statusMap["power"] = get_power_status();
    statusMap["PLL"] = get_pll_status();
    return statusMap;
}
