/*
**    Copyright 2009-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "BruteFT.h"

#include <cassert>
#include <cstddef>
#include <cmath>
#include <iostream>

namespace dsplib {

// Discrete Fourier Transform, order N, Real form
// symmetry: X[k] = X*[N-k]
void BruteFT::DFT (const std::vector<double> &x, std::vector<double> *S, std::vector<double> *C)
{
    assert(S);
    assert(C);
    const size_t n = x.size();
    const size_t out_size = n/2+1;
    S->resize(out_size);
    C->resize(out_size);
    for (size_t k = 0; k < out_size; k++) {
        (*S)[k] = 0;
        (*C)[k] = 0;
        long double s=0, c=0;
        for (size_t i = 0; i < n; i++) {
            const double arg = 2 * M_PI * k * i / n;
            const double is = -sin(arg);
            const double ic = cos(arg);
            s += is * x[i];
            c += ic * x[i];
        }
        (*S)[k] = s;
        (*C)[k] = c;
        //        (*S)[k] /= sqrt (n / 2);
        //        (*C)[k] /= sqrt (n / 2);
    }
    //  if (n%2 == 0) /* N is even */ C[n/2] *= .5; /* Nyquist freq. */
}

// Inverse Discrete Fourier Transform, order n, Real form
std::vector<double> BruteFT::IDFT (const std::vector<double> &S, const std::vector<double> &C)
{
    // BUG: the result is not consistent with fftw
    //    const size_t N = (C.size()-1)*2;
    //    std::vector<double> out(N);
    //    fftw_plan p = fftw_plan_r2r_1d(N, &x[0], &out[0], FFTW_HC2R, FFTW_ESTIMATE);
    //    Q_ASSERT(p);
    //    fftw_execute(p);
    if (S.size() < 1 || C.size() < 1)
        return std::vector<double> ();
    assert(S.size() <= C.size());
    const size_t n = C.size() + S.size() - 1;
    std::cerr << "IDFT: n = " << n << std::endl;
    std::vector<double> out(n);

    for (size_t k = 0; k < n; k++) {
        long double x = 0;
        for(size_t i=1; i<S.size(); i++) {
            const long double a = (long double)(2. * i * k) / n;
            const long double arg = M_PI*a;
            x += sinl(arg)*S[i];
        }
        for(size_t i=0; i<C.size(); i++) {
            const long double a = (long double)(2. * i * k) / n;
            const long double arg = M_PI*a;
            long double m = 2;
            if ((i==0) || i+1==C.size()) m=1;
            //            if (i==0) m=1;
            //            if (((i+1)==C.size())) m=2;
            x += cosl(arg)*C[i]*m;
        }
        /*
        for (size_t i = 0; i < (n / 2 + 1); i++) {
            c = C[i];
//            if ((i == 0) || (i == n / 2))
//                c *= 0.5;
            arg = 2 * pi * i * k / n;
            is = sin(arg);
            ic = cos(arg);
            x += (is * S[i] + ic * c);
            //      x[k] += (sin(2*pi*i*k/n)*S[i] + cos(2*pi*i*k/n)*c);
        }
*/
        //        x[k] /= sqrt (n / 2);
        out[k] = x;
    }
    return out;
}

// Calculate power from rectangular FFT
std::vector<double> BruteFT::RECT2POW (const std::vector<double> &S, const std::vector<double> &C)
{
    std::vector<double> out;
    RECT2POW(S, C, &out);
    return out;
}

void BruteFT::RECT2POW (const std::vector<double> &S, const std::vector<double> &C, std::vector<double> *out)
{
    const size_t n = S.size();
    assert(n == C.size());
    if (out->size() != n)
        out->resize(n);
    for (size_t i = 0; i < n; i++)
        (*out)[i] = hypot (S[i], C[i]);
}

// Calculate phase from rectangular FFT
std::vector<double> BruteFT::RECT2PHA (const std::vector<double> &S, const std::vector<double> &C)
{
    assert(S.size() == C.size());
    std::vector<double> out(S.size());
    for (size_t i = 0; i < S.size(); i++)
        out[i] = atan2 (S[i], C[i]);
    return out;
}

// Compatibility
void BruteFT::DFT_tmp (const std::vector<double> &x, std::vector<double> *S, std::vector<double> *C)
{
    assert(S);
    assert(C);
    const size_t n = x.size();
    const size_t out_size = n/2+1;
    S->resize(out_size);
    C->resize(out_size);
    for (size_t k = 0; k < out_size; k++) {
        (*S)[k] = 0;
        (*C)[k] = 0;
        for (size_t i = 0; i < n; i++) {
            double arg = 2 * M_PI * k * i / n;
            double is = sin(arg);
            double ic = cos(arg);
            (*S)[k] += is * x[i];
            (*C)[k] += ic * x[i];
        }
        (*S)[k] /= sqrt (n / 2 + .5);
        (*C)[k] /= sqrt (n / 2 + .5);
    }
    //  if (n%2 == 0) /* N is even */ C[n/2] *= .5; /* Nyquist freq. */
}

void BruteFT::IDFT (std::vector<double> *x, const std::vector<double> &S, const std::vector<double> &C)
{

    // Inverse Discrete Fourier Transform, order n
    // Real form

    size_t n = (std::min(S.size(), C.size()) - 1) * 2;
    if (x->size() != n)
        x->resize(n);

    for (size_t k = 0; k < n; k++) {
        (*x)[k] = 0;
        for (size_t i = 0; i < (n / 2 + 1); i++) {
            double c = C[i];
            if ((i == 0) || (i == n / 2))
                c *= 0.5;
            double arg = 2 * M_PI * i * k / n;
            double is = sin(arg);
            double ic = cos(arg);
            (*x)[k] += (is * S[i] + ic * c);
            //      x[k] += (sin(2*pi*i*k/n)*S[i] + cos(2*pi*i*k/n)*c);
        }
        (*x)[k] /= sqrt (n / 2 + .5);
    }
}

} // namespace dsplib
