/*
**    Copyright 2009-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "FirDesign.h"
#include "RFFTEngine.h"
#include "dsplib.h"
#include <algorithm>
#include <cassert>

namespace dsplib {

static const double eps = 1e-100;

std::vector<double> FirDesign::create_deconvolution_dblFIR (const std::vector<double> &dsrc, const std::vector<double> &dreq)
{
    // pad with zeroes to make lengths equal
    std::vector<double> src(dsrc);
    std::vector<double> req(dreq);
    auto max_size = std::max(src.size(), req.size());
    if (max_size < 2)
        return {};

    src.resize(max_size);
    req.resize(max_size);

    const size_t inlen = max_size;

    assert(inlen <= 10000);

    if ((inlen > 10000)) {
        //        logger.err("Bad input");
        return {};
    }

    const unsigned int deconvkernlen = inlen;
    const unsigned int deconvkernhalflen = deconvkernlen / 2 + 1;

    // Calculate signal DC component
    // const double dcsrc = dcdbl (src);
    const double dcreq = dcdbl (req);
    // Calculate signal RMS power
    // const double rmssrc = rmsdbl (src, dcsrc);
    const double rmsreq = rmsdbl (req, dcreq);
    if (rmsreq < eps) {
        return {};
    }

    RFFTEngine fft;
    const auto &src_dft = fft.DFT(src);
    const auto &req_dft = fft.DFT(req);
    auto imp_dft(src_dft);
    const double N = src.size();
    for (size_t i = 0; i < N/2+1; i++) {
        imp_dft[i] = req_dft[i] / src_dft[i] / N;
        // amplification limit
        auto mag = abs(imp_dft[i]);
        const double amp_limit = 1.0;
        if (mag > amp_limit) // FIXME
            imp_dft[i] /= (mag/amp_limit);
    }
    for (size_t i = N/2+1; i < imp_dft.size(); i++) {
        imp_dft[i] = 0;
    }
    std::vector<double> imp_mag(imp_dft.size());
    for (size_t i = 0; i < imp_dft.size(); i++) {
        imp_mag[i] = abs(imp_dft[i]);
    }
    auto max_amplif = std::max_element(imp_mag.cbegin(), imp_mag.cend());

    const auto &imp = fft.IDFT(imp_dft);

    std::vector<double> result;
    for (unsigned int i = 0; i < deconvkernlen; i++) {
        result.push_back(imp[(i + deconvkernhalflen) % deconvkernlen]);
    }
    return result;
}

} // namespace dsplib
