/*
**    Copyright 2021 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

// see https://www.jsonrpc.org/specification

#include "JsonRpc.h"

#include <QDeadlineTimer>
#include <QTimer>
#include <QLoggingCategory>

Q_DECLARE_LOGGING_CATEGORY(lc_jsonrpc)
Q_LOGGING_CATEGORY(lc_jsonrpc, "afi.jsonrpc")

static RpcRequestId rpcSeq = 0;
const int pollTimeout = 100;

class RpcRequestQueueItem
{
public:
    explicit RpcRequestQueueItem(const JsonRpcRequest &j, int timeoutMs)
        : request(j), timer(timeoutMs) {}
    bool hasExpired() const { return timer.hasExpired(); }
    JsonRpcRequest getRequest() const { return request; }
private:
    JsonRpcRequest request;
    QDeadlineTimer timer;
};

class JsonRpcPrivate
{
public:
    QMap<RpcRequestId, RpcRequestQueueItem> queue;
    QTimer pollTimer;
};

static QJsonObject createJsonRpcError(JsonRpcError err)
{
    static const QMap<JsonRpcError, QString> errorStr =
        {
            {JsonRpcError::Timeout, "Timeout"},
            {JsonRpcError::ParseError, "Parse error"},
            {JsonRpcError::InvalidRequest, "Invalid request"},
            {JsonRpcError::MethodNotFound, "Method not found"},
            {JsonRpcError::InvalidParams, "Invalid parameters"},
            {JsonRpcError::InternalError, "Internal error"},
            {JsonRpcError::ApplicationError, "Application error"},
            {JsonRpcError::SystemError, "System error"},
            {JsonRpcError::TransportError, "Transport error"},
        };
    return QJsonObject
        {{"code", QJsonValue{static_cast<int>(err)}},
         {"message", QJsonValue{errorStr[err]}}};
}

JsonRpc::JsonRpc(QObject *parent)
    : QObject{parent},
    d_ptr(new JsonRpcPrivate)
{
    Q_D(JsonRpc);
    connect(&d->pollTimer, &QTimer::timeout, this, &JsonRpc::pollQueue);
    d->pollTimer.setInterval(pollTimeout);
    d->pollTimer.setSingleShot(true);
    d->pollTimer.start();

    connect(this, &JsonRpc::rpcError,
            this, &JsonRpc::onRpcError);
}

JsonRpc::~JsonRpc()
{
    delete d_ptr;
}

JsonRpcRequest JsonRpc::findRequest(const RpcRequestId &id) const
{
    Q_D(const JsonRpc);
    auto it = d->queue.constFind(id);
    if (it == d->queue.constEnd())
        return {};
    else
        return it.value().getRequest();
}

void JsonRpc::messageReceived(const QJsonObject &json)
{
    const auto &id = json["id"].toVariant().toLongLong();
    const auto &jsonrpc = json.value("jsonrpc").toString();
    if (jsonrpc != "2.0") {
        rpcRespondError(id, JsonRpcError::InvalidRequest);
        return;
    }
    // resuest: method
    // response: result or error
    const auto &method = json["method"];
    const auto &result = json["result"];
    const auto &error = json["error"];
    if (!error.isUndefined()) {
        dequeue(id);
        emit rpcError(id, error);
        return;
    }
    if (!result.isUndefined()) {
        dequeue(id);
        emit rpcComplete(id, result);
        return;
    }
    if (!method.isUndefined()) {
        auto params = json["params"];
        emit rpcReceived(id, JsonRpcRequest{method.toString(), params});
        return;
    }
    rpcRespondError(id, JsonRpcError::InvalidRequest);
}

RpcRequestId JsonRpc::rpcCall(const JsonRpcRequest &request, int timeoutMs)
{
    Q_D(JsonRpc);
    const auto id = rpcSeq++;
    auto jrpc = QJsonObject
        {{"jsonrpc", "2.0"},
         {"method", request.method},
         {"params", request.params},
         {"id", QVariant::fromValue(id).toJsonValue()}};
    auto item = RpcRequestQueueItem(request, timeoutMs);
    if (!request.method.startsWith("notify")) {
        d->queue.insert(id, item);
    }
    emit sendMessage(jrpc);
    return id;
}

void JsonRpc::rpcRespondResult(const RpcRequestId &id, const QJsonValue &result)
{
    auto response = QJsonObject
        {{"jsonrpc", "2.0"},
         {"id", QVariant::fromValue(id).toJsonValue()},
         {"result", result}
        };
    emit sendMessage(response);
}

void JsonRpc::rpcRespondError(const RpcRequestId &id, JsonRpcError error)
{
    auto response = QJsonObject
        {{"jsonrpc", "2.0"},
         {"id", QVariant::fromValue(id).toJsonValue()},
         {"error", createJsonRpcError(error)}
         };
    emit sendMessage(response);
}

void JsonRpc::onRpcError(const RpcRequestId &id, const QJsonValue &error)
{
    auto req = findRequest(id);
    qWarning(lc_jsonrpc).noquote().nospace() <<
        QString("RPC #%1 (%2) error: %3 (%4)")
            .arg(id)
            .arg(req.method, error["message"].toString())
            .arg(error["code"].toInt());
}

void JsonRpc::pollQueue()
{
    Q_D(JsonRpc);
    QMutableMapIterator<RpcRequestId, RpcRequestQueueItem> i(d->queue);
    while (i.hasNext()) {
        auto m = i.next();
        auto id = m.key();
        auto item = m.value();
        if (item.hasExpired()) {
            emit rpcError(id, createJsonRpcError(JsonRpcError::Timeout));
            i.remove();
        }
    }
    d->pollTimer.start();
}

void JsonRpc::dequeue(const QVariant &id)
{
    Q_D(JsonRpc);
    d->queue.remove(id.toLongLong());
}
