/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceEnableTableModel.h"

int DeviceEnableTableModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return devices.size();
}

int DeviceEnableTableModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return COL_TOTAL;
}

DeviceIndex DeviceEnableTableModel::deviceIndexOf(int row) const
{
    if (row < 0)
        return {};
    int r = 0;
    for (auto it = devices.cbegin(); it != devices.cend(); ++it)
        if (r++ == row)
            return it.key();
    return {};
}

int DeviceEnableTableModel::rowOf(const DeviceIndex &deviceIndex) const
{
    return std::distance(devices.begin(), devices.find(deviceIndex));
}

static const double ResetModelRatioThreshold = 0; // always use resetModel

void DeviceEnableTableModel::setDeviceList(const DeviceIndexEnabledMap &map)
{
    const auto &newSet = map.getEnabled();
    const auto &oldSet = makeDeviceIndexSet(devices.keys());
    const auto toRemove = oldSet-newSet;
    const auto toAdd = newSet-oldSet;
    bool reset = false;
    reset |= (toRemove.size() >= devices.size() * ResetModelRatioThreshold);
    reset |= (toAdd.size() >= devices.size() * ResetModelRatioThreshold);
    if (reset) {
        beginResetModel();
        for(auto i : toRemove)
            devices.remove(i);
        for(auto i : toAdd)
            devices.insert(i, map[i]);
        endResetModel();
    } else {
        for(auto i : toRemove) {
            removeRow(rowOf(i));
            devices.remove(i);
        }
        for(auto i : toAdd) {
            devices.insert(i, map[i]);
            insertRow(rowOf(i));
        }
    }
}

Qt::ItemFlags DeviceEnableTableModel::flags(const QModelIndex & index) const
{
    auto f = QAbstractTableModel::flags(index);
    switch (index.column()) {
    case COL_ENABLE:
        return  f | Qt::ItemIsUserCheckable;
    default:
        return f;
    }
}

bool DeviceEnableTableModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    const auto &deviceIndex = deviceIndexOf(index.row());
    if (deviceIndex.isNull())
        return {};

    const auto it = devices.find(deviceIndex);
    if (it == devices.end())
        return false;

    if (role == Qt::CheckStateRole) {
        if (index.column() == COL_ENABLE) {
            bool state = (value == Qt::Checked);
            if (it->enabled == state)
                return true;
            it->enabled = state;
            emit dataChanged(index, index);
            emit deviceListChanged(devices);
            return true;
        }
    }
    if (role == Qt::DisplayRole) {
        if (index.column() == COL_ENABLE) {
            bool state = value.value<bool>();
            if (it->enabled == state)
                return true;
            it->enabled = state;
            emit dataChanged(index, index);
            emit deviceListChanged(devices);
            return true;
        }
    }
    return false;
}

QVariant DeviceEnableTableModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    const auto &deviceIndex = deviceIndexOf(index.row());
    if (deviceIndex.isNull())
        return {};

    const auto &device = devices[deviceIndex];

    switch(column) {
    case COL_ENABLE:
        switch (role) {
        case Qt::EditRole: return device.enabled;
        case Qt::CheckStateRole: return device.enabled ? Qt::Checked : Qt::Unchecked;
        }
        break;
    case COL_DEVICE_ID:
        switch (role) {
        case Qt::DisplayRole: return deviceIndex.getDeviceIdName();
        }
        break;
    case COL_DEVICE_SERIAL:
        switch (role) {
        case Qt::DisplayRole: return deviceIndex.getSerialStr();
        }
        break;
    }
    return {};
}

QVariant DeviceEnableTableModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (static_cast<Columns>(section)) {
        case COL_ENABLE: return tr("En");
        case COL_DEVICE_ID: return tr("Type");
        case COL_DEVICE_SERIAL: return tr("Serial");
        case COL_TOTAL: return {};
        }
    }
    return DeviceIndexTableModel::headerData(section, orientation, role);
}

QMap<int, bool> DeviceEnableTableModel::defaultColumnState() const
{
    return  {
        {COL_ENABLE, false},
        {COL_DEVICE_ID, false},
        {COL_DEVICE_SERIAL, false}
    };
}
