/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MergeColumnsTableModel.h"

#include <optional>
#include <QDebug>

MergeColumnsTableModel::MergeColumnsTableModel(QObject *parent)
    : QAbstractTableModel(parent)
{
}

int MergeColumnsTableModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return rowKeys.size();
//    int v = 0;
//    QSet<QVariant> keySet;
//    for(auto mi : modelList) {
//        for(auto r=0; r<mi.model->rowCount(); ++r)
//            keySet << mi.model->data(mi.model->index(r, mi.mergeCol), mi.mergRole);
//    }

//    return v;
}

int MergeColumnsTableModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    int v = 0;
    for (auto &m : modelList)
        v += m.model->columnCount();

    return v;
}

int MergeColumnsTableModel::getColumnFromSource(QAbstractItemModel *model, int col) const
{
    int colOffset = 0;

    for (auto &mi : modelList) {
        if (model == mi.model)
            return col + colOffset;
        colOffset += mi.model->columnCount();
    }

    return -1;
}

MergeColumnsTableModel::ModelWithColumn MergeColumnsTableModel::mapColumnToSource(int column) const
{
    if (column < 0)
        return {};

    for (auto &mi : modelList) {
        const auto &modCol = mi.model->columnCount();
        if (column < modCol)
            return {mi.model, column};
        column -= modCol;
    }
    return {};
}

QModelIndex MergeColumnsTableModel::mapFromSource(const QModelIndex &source) const
{
    int colOffset = 0;

    for (auto &mi : modelList)
        if (source.model() == mi.model) {
            const auto &key = mi.getKeyByRow(source.row());
            auto r = rowKeys.indexOf(key);
            if(r == -1)
                return QModelIndex();
            return index(r, source.column() + colOffset);
        } else
            colOffset += mi.model->columnCount();

    return QModelIndex();
}

QModelIndex MergeColumnsTableModel::mapToSource(const QModelIndex & proxy) const
{
    if (! proxy.isValid())
        return {};
    int row = proxy.row();
    int column = proxy.column();
    if (row >= rowKeys.size())
        return {};
    const auto &key = rowKeys[row];

    for (auto &mi : modelList) {
        const auto &modCol = mi.model->columnCount();
        if(column < modCol) {
            auto innerRow=mi.getRowByKey(key);
            if(innerRow == -1)
                return QModelIndex();
            return mi.model->index(innerRow, column);
        } else
            column -= modCol;
    }
    return {};
}

Qt::ItemFlags MergeColumnsTableModel::flags(const QModelIndex &index) const
{
    if (!index.isValid())
        return {};
    auto sourceIndex = mapToSource(index);
    if (!sourceIndex.isValid())
        return {};
    return sourceIndex.flags();
}

QVariant MergeColumnsTableModel::data(const QModelIndex &index, int role) const
{
    return mapToSource(index).data(role);
}

bool MergeColumnsTableModel::setData(const QModelIndex& index, const QVariant& value, int role)
{
    auto sourceIndex = mapToSource(index);
    QAbstractItemModel * sourceModel = nullptr;
    for (auto &mi : modelList) {
        if (sourceIndex.model() == mi.model) {
            sourceModel = mi.model;
            break;
        }
    }
    if (!sourceModel)
        return {};
    return sourceModel->setData(sourceIndex, value, role);
}

QVariant MergeColumnsTableModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Vertical) {
        if (section < 0 || section >= rowKeys.size())
            return {};

        const auto &key = rowKeys[section];

        for (auto &mi : modelList) {
            if (primaryModel && mi.model != primaryModel)
                continue;
            int row = mi.getRowByKey(key);
            if (row == -1)
                continue;
            const auto &data = mi.model->headerData(row, orientation, role);
            if (!data.isNull())
                return data;
        }
        return {};
    }
    if (orientation == Qt::Horizontal) {
        for (auto &mi : modelList) {
            const auto &modCol = mi.model->columnCount();
            if (section < modCol)
                return mi.model->headerData(section, orientation, role);
            else
                section -= modCol;
        }
    }
    return {};
}

void MergeColumnsTableModel::connectSourceModel(QAbstractItemModel *model)
{
    connect(model, &QAbstractItemModel::dataChanged,
            this, &MergeColumnsTableModel::onDataChanged);
    connect(model, &QAbstractItemModel::rowsInserted,
            this, &MergeColumnsTableModel::onRowsInserted);
//    connect(model, &QAbstractItemModel::rowsAboutToBeRemoved,
//            this, &MergeColumnsTableModel::onRowsAboutToBeRemoved);
    connect(model, &QAbstractItemModel::rowsRemoved,
            this, &MergeColumnsTableModel::onRowsRemoved);
    connect(model, &QAbstractItemModel::modelReset,
            this, &MergeColumnsTableModel::sourceModelReset);
}

void MergeColumnsTableModel::updateRowKeys()
{
    rowKeys.clear();
//    QSet<QVariant> keySet;

    for (auto &mi : modelList) {
        for(auto r=0; r<mi.model->rowCount(); ++r) {
            const auto &key = mi.getKeyByRow(r);
//            keySet << key;
            if(!rowKeys.contains(key))
                rowKeys << key;
        }
    }
//    rowKeys = keySet.values().toVector();
}

bool MergeColumnsTableModel::sourceKeyExists(const QVariant &key, const QAbstractItemModel *excludeModel) const
{
    bool keyExists = false;
    for (auto &mi : modelList) {
        if (!mi.model || (mi.model == excludeModel))
            continue;
        if (mi.getRowByKey(key) != -1) {
            keyExists = true;
            break;
        }
    }
    return keyExists;
}

void MergeColumnsTableModel::addSourceModel(QAbstractItemModel *model, Qt::ItemDataRole mergeRole)
{
    beginResetModel();
    modelList.append(MergeModelInfo(model, mergeRole));
    connectSourceModel(model);
    updateRowKeys();
    endResetModel();
    emit modelReset(); // endResetModel() doesn't emit this signal
}

void MergeColumnsTableModel::setPrimarySourceModel(QAbstractItemModel *model)
{
    if (primaryModel == model)
        return;
    beginResetModel();
    primaryModel = model;
    endResetModel();
}

QList<QAbstractItemModel *> MergeColumnsTableModel::sourceModels() const
{
    QList<QAbstractItemModel *> result;
    for (auto &mi : modelList) {
        if (mi.model)
            result.push_back(mi.model);
    }
    return result;
}

void MergeColumnsTableModel::onDataChanged(const QModelIndex &topLeft,
                                           const QModelIndex &bottomRight,
                   const QVector<int> &roles)
{
    for(auto r=topLeft.row(); r<=bottomRight.row(); ++r) {
        const auto &left = mapFromSource(topLeft.siblingAtRow(r));
        const auto &right = mapFromSource(bottomRight.siblingAtRow(r));
        emit dataChanged(left, right, roles);
    }
}

void MergeColumnsTableModel::onRowsInserted(const QModelIndex &parent, int first, int last)
{
    auto prevRowKeys = rowKeys;
    updateRowKeys();
    if (prevRowKeys != rowKeys)
        emit modelReset();
}

//void MergeColumnsTableModel::onRowsAboutToBeRemoved(const QModelIndex &parent, int first, int last)
//{
//    // find source model
//    auto sourceModel = qobject_cast<QAbstractItemModel *>(sender());
//    std::optional<MergeModelInfo> info;
//    for (auto &mi : modelList) {
//        if (mi.model == sourceModel) {
//            info = mi;
//            break;
//        }
//    }
//    if (!info.has_value())
//        return;
//    QVector<QVariant> keysToRemove;
//    for (int r = first; r <= last; r++) {
//        keysToRemove.push_back(info->getKeyByRow(r));
//    }
//    for (auto it = keysToRemove.begin(); it != keysToRemove.end(); ++it) {
//        const auto &key = *it;
//        // check if other models have same key
//        if (sourceKeyExists(key, sourceModel))
//            continue;
//        int row = rowKeys.indexOf(key);
//        if (row == -1)
//            continue;
//        beginRemoveRows({}, row, row);
//    }
//    qDebug() << __PRETTY_FUNCTION__ << first << last;
//}

void MergeColumnsTableModel::onRowsRemoved(const QModelIndex &parent, int first, int last)
{
    auto prevRowKeys = rowKeys;
    updateRowKeys();
    // endRemoveRows();
    if (prevRowKeys != rowKeys)
        emit modelReset();
}

void MergeColumnsTableModel::sourceModelReset()
{
    updateRowKeys();
    emit modelReset();
}

QVariant MergeColumnsTableModel::MergeModelInfo::getKeyByRow(int row) const
{
    Q_ASSERT(model);
    if (!model)
        return {};
    return model->headerData(row, Qt::Vertical, mergeRole);
}

int MergeColumnsTableModel::MergeModelInfo::getRowByKey(const QVariant &key) const
{
    Q_ASSERT(model);
    for(auto row=0; model && row<model->rowCount(); ++row) {
        if(getKeyByRow(row) == key)
            return row;
    }

    return -1;
}
