/*
**    Copyright 2010-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "FirDesignDialog.h"
#include "ui_FirDesignDialog.h"

#include <cmath>
#include <cstdlib>
#include <cstring>
#include <algorithm>
#include <span>

#include <QDebug>
#include <QMessageBox>
#include <QComboBox>
#include <QLabel>
#include <QSpinBox>
#include <QSettings>
#include <qwt_scale_engine.h>

#include "FirDesignDialogPrivate.h"
#include "base/gui_utils.h"
#include "dsplib/Convolver.h"
#include "dsplib/FitterGaus.h"
#include "dsplib/FirDesign.h"
#include "dsplib/RFFTEngine.h"
#include "dsplib/WindowFunction.h"
#include "dsplib/dsplib.h"
#include "dsplib/fir.h"
#include "qxw/WaveformData.h"
#include "qxw/WaveformPlot.h"
#include "widgets/FileDialog.h"

auto snf_db_min = -120;
auto snf_db_max = 0;

static constexpr const auto k_firFileName = "firFileName";
static constexpr const auto k_firSelectedFilter = "firSelectedFilter";
static constexpr const auto k_shapeFileName = "shapeFileName";
static constexpr const auto k_shapeSelectedFilter = "shapeSelectedFilter";
static constexpr const auto k_textFileFilter = "Text files (*.txt);;All files (*)";

static constexpr const auto settingsSection = "FirDesignDialog";
static const int calibrationColorIndex = 4;
static const int residualsColorIndex = 0;
static const int referenceColorIndex = 1;
static const int filterColorIndex = 6;
static const int testColorIndex = 4;

static const std::initializer_list<std::pair<int, const char *>> QTextStreamStatusMap {
    {QTextStream::Ok, "Ok"},
    {QTextStream::ReadPastEnd, "Read past end"},
    {QTextStream::ReadCorruptData, "Read corrupt data"},
    {QTextStream::WriteFailed, "Write failed"}
};

template<typename T>
static void setSettings(const QString &name, const T &value)
{
    QSettings settings;
    settings.setValue(QString(settingsSection) + "/" + name, value);
}
static QVariant getSettings(const QString &name)
{
    const QSettings settings;
    return settings.value(QString(settingsSection) + "/" + name);
}

using namespace dsplib;

FirDesignDialog::FirDesignDialog (QWidget* parent)
    : QDialog(parent)
    , ui(new Ui::FirDesignDialog)
    , d_ptr(new FirDesignDialogPrivate)
{
    ui->setupUi(this);

    {
        RecursiveWidgetSignalBlocker blocker(this);
        constexpr const std::array<std::pair<const char *, WindowFunction::fwinf_t>, 11> items {{
            {"Rectangle", WindowFunction::WF_RECT},
            {"Blackman", WindowFunction::WF_BLACKMAN},
            {"Hamming", WindowFunction::WF_HAMMING},
            {"Hanning", WindowFunction::WF_HANN},
            {"Barlett", WindowFunction::WF_BARLETT},
            {"Tukey 0.75", WindowFunction::WF_TUKEY_075},
            {"Tukey 0.5", WindowFunction::WF_TUKEY_050},
            {"Tukey 0.25", WindowFunction::WF_TUKEY_025},
            {"Gauss 0.25", WindowFunction::WF_GAUSS_025},
            {"Gauss 0.5", WindowFunction::WF_GAUSS_050},
            {"Gauss 0.75", WindowFunction::WF_GAUSS_075}
        }};
        for (const auto &it : items) {
            ui->comboBoxWindowingFunction->addItem(it.first, it.second);
        }
        //    win_type = WF_BLACKMAN;             // default
        //    comboBoxWindowingFunction->setCurrentIndex(win_id_by_type((fwinf_t)d->newDecoderConfig.fwinf[0]));

        ui->sbfirnp->setRange(2, 16);
        ui->sbfirnp->setValue(16);

        ui->sbwid->setRange(0.1, 100);
        ui->sbwid->setSingleStep(0.1);
        ui->sbwid->setDecimals(2);

        ui->sbpos->setRange(-1000, 1000);
        ui->sbpos->setDecimals(2);
        ui->sbpos->setSingleStep(0.1);

        ui->sbrnf->setRange(snf_db_min, snf_db_max);
        ui->sbrnf->setDecimals(1);
        ui->sbsnf->setRange(snf_db_min, snf_db_max);
        ui->sbsnf->setDecimals(1);
    }

    Restore();
}

FirDesignDialog::~FirDesignDialog ()
{
    delete d_ptr;
    delete ui;
}

DecoderConfig FirDesignDialog::getDecoderConfig() const
{
    Q_D(const FirDesignDialog);
    return d->newDecoderConfig;
}

void FirDesignDialog::setAdcConfig(const FirDesignConfig &r)
{
    Q_D(FirDesignDialog);
    if (r == d->firDesignConfig)
        return;
    d->firDesignConfig = r;
}

void FirDesignDialog::setDecoderConfig(const DecoderConfig &r)
{
    Q_D(FirDesignDialog);
    if (r == d->newDecoderConfig)
        return;
    d->newDecoderConfig = r;
    Restore();
}

WindowFunction::fwinf_t FirDesignDialog::win_type() const
{
    Q_D(const FirDesignDialog);
    return d->firDesignConfig.fwinf;
}

std::vector<double> FirDesignDialog::getCalAvgY() const
{
    Q_D(const FirDesignDialog);
    const auto &data = d->newDecoderConfig.calavg.samples;
    return std::vector<double>(data.begin(), data.end());
}

double FirDesignDialog::getCalAvgAmplitude() const
{
    const auto &calAvg = getCalAvgY();
    if (calAvg.empty())
        return 0;
    return std::max(1.0, *std::max_element(calAvg.begin(), calAvg.end()));
}

size_t FirDesignDialog::getRefSize() const
{
    Q_D(const FirDesignDialog);
    return d->newDecoderConfig.calavg.samples.size();
}

std::vector<double> FirDesignDialog::getRefY_n(double amplitudeScaler, int n) const
{
    const auto lat = data_latency();
    std::vector<double> ref;
    ref.reserve(n);
    for (int i = 0; i < n; i++) {
        ref.push_back(amplitudeScaler * reqfunscaled(i - lat));
    }
    // Filter

    return ref;
}

//std::vector<double> FirDesignDialog::getRefY(int channel, double amplitudeScaler) const
//{
//    return getRefY_n(channel, amplitudeScaler, getRefSize());
//}

std::vector<double> FirDesignDialog::getRefY() const
{
    return getRefY_n(getCalAvgAmplitude(), getRefSize());
}

std::vector<double> FirDesignDialog::getRefXInNs() const
{
    const auto lat = data_latency();
    const size_t n = getRefSize();
    std::vector<double> x;
    x.reserve(n);
    for (size_t i = 0; i < n; i++) {
        x.push_back(i - lat);
    }
    return x;
}

std::vector<double> FirDesignDialog::getCalAvgXInNs() const
{
    return getRefXInNs();
}

void FirDesignDialog::replotINP ()
{
    WaveformPlot *wp = ui->plotINP; // Input signal

    auto vx = getCalAvgXInNs();
    auto vy0 = getCalAvgY();
    auto vy1 = getRefY();
    size_t n = std::min(vx.size(), std::min(vy0.size(), vy1.size()));
    if (n == 0) {
        wp->clear();
        wp->replot();
        return;
    }
    vx.resize(n);
    vy0.resize(n);
    vy1.resize(n);
    WaveformDataAll wda;
    wda.push_back(WaveformData(vx, vy0, tr("Calibration data"), calibrationColorIndex));
    wda.push_back(WaveformData(vx, vy1, tr("Reference data"), referenceColorIndex));
    wp->setWaveformData(wda);
    wp->enableLegend(false);

    QwtText xTitle(tr("Time"));
    xTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->replot();
}

std::vector<double> FirDesignDialog::magnitudesDb(const std::vector<std::complex<double>> &v)
{
    std::vector<double> result;
    result.reserve(v.size());
    for (const auto &i : v) {
        result.push_back(Decibels::val2db(std::abs(i)));
    }
    return result;
}

std::vector<std::complex<double>> FirDesignDialog::makeRealDFT(const std::vector<double> &req) const
{
    Q_D(const FirDesignDialog);

    if (req.empty())
        return {};
    auto reqDFT = d->fft.DFT(req);
    const auto N = req.size();
    reqDFT.resize(N/2+1); // Real-to-Complex DFT result is symmetric
    const double dftScaler = sqrt(2) / (N);
    for (auto i=0u; i<reqDFT.size(); i++)
        reqDFT[i] *= dftScaler;
    return reqDFT;
}

std::vector<double> FirDesignDialog::fftFrequencies(int timePoints)
{
    int freqPoints = timePoints / 2 + 1;
    std::vector<double> v;
    v.reserve(freqPoints);
    for (int i = 0; i < freqPoints; i++) {
        v.push_back(double(i) / timePoints);
    }
    return v;
}

void FirDesignDialog::replotINPFFT ()
{
    Q_D(const FirDesignDialog);

    WaveformPlot *wp = ui->plotINPFFT; // Input signal FFT
    const auto &inp = getCalAvgY();
    const auto &req = getRefY();

    const auto &inpDftPy = magnitudesDb(makeRealDFT(inp));
    const auto &reqDftPy = magnitudesDb(makeRealDFT(req));
    const auto timePoints = std::min(inp.size(), req.size());
    const auto &dftPx = fftFrequencies(timePoints);

    WaveformDataAll wda;
    wda.push_back(WaveformData(inpDftPy, tr("Source FFT"), calibrationColorIndex));
    wda.push_back(WaveformData(reqDftPy, tr("Target FFT"), referenceColorIndex));

    if (!dftPx.empty()) {
        const double snfdb = d->firDesignConfig.snf; // - fullScaleDb;
        const double rnfdb = d->firDesignConfig.pnf; // - fullScaleDb;
        WaveformPlot::MarkerData snfMarker;
        snfMarker.lineStyle = QwtPlotMarker::HLine;
        snfMarker.yvalue = snfdb;
        snfMarker.pen = QPen(WaveformPlot::standardColor(calibrationColorIndex), 1, Qt::DotLine);
        snfMarker.label = QString("input %1 dB").arg(snfdb);
        WaveformPlot::MarkerData rnfMarker;
        rnfMarker.lineStyle = QwtPlotMarker::HLine;
        rnfMarker.yvalue = rnfdb;
        rnfMarker.pen = QPen(WaveformPlot::standardColor(referenceColorIndex), 1, Qt::DotLine);
        rnfMarker.label = QString("target %1 dB").arg(rnfdb);
        wp->setMarkers({snfMarker, rnfMarker});
    }
    wp->setWaveformData(wda);

    //  wp->setAxisScaleEngine(QwtPlot::yLeft, new QwtLog10ScaleEngine);
    //  wp->axisScaleEngine(QwtPlot::yLeft)->setAttribute(QwtScaleEngine::Floating);
    //  wp->setAxisOptions (QwtPlot::yLeft, QwtAutoScale::Floating | QwtAutoScale::Logarithmic);
    //  wp->setAxisScale (QwtPlot::yLeft, INPFFT_MIN, INPFFT_MAX);
    //  wp->setxrange (0, H);

    //  printf("%edB  %e   %edB  %e\n", snf, snfdb, rnf, rnfdb);
    //  printf("%e:%e - %e:%e\n", px[0], p1[0], px[1], p1[1]);
    //  printf("%e:%e - %e:%e\n", px[0], p2[0], px[1], p2[1]);
    wp->enableLegend(false);

    QwtText xTitle(tr("Frequency"));
    QwtText yTitle(tr("dBFS"));
    xTitle.setFont(font());
    yTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->setAxisTitle(QwtPlot::yLeft, yTitle);
    wp->replot();
}

void FirDesignDialog::replotFIR ()
{
    Q_D(FirDesignDialog);

    WaveformPlot *wp = ui->plotFIR; // FIR impulse

    const auto &fir = d->newDecoderConfig.channels.fir;
    if (fir.empty()) {
        wp->clear();
        ui->tlklen->clear();
        ui->tloffs->clear();
        ui->tlsum->clear();
        return;
    }

    const auto klen = fir.size();
    std::vector<double> px, FIRpy, WFpy;
    for (size_t i = 0; i < fir.size(); i++) {
        px.push_back(i);
        FIRpy.push_back(fir[i]);
        WFpy.push_back(WindowFunction::value(win_type(), i, klen));
    }

    WaveformDataAll wda;
    wda.push_back(WaveformData(px, FIRpy, tr("Filter Impulse Response"), filterColorIndex));
    wda.push_back(WaveformData(px, WFpy, tr("Windowing"), filterColorIndex+2));
    wp->setWaveformData(wda);
    wp->enableLegend(false);

    QwtText xTitle(tr("Time"));
    xTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->replot();

    ui->tlklen->setText (tr("Size: %1").arg(klen));
    double sum = 0;
    for (size_t  i = 0; i < klen; i++)
        sum += fir[i];
    ui->tlsum->setText (tr("Sum: %1").arg(sum));
}

static std::vector<double> phaseUnwrap(const std::vector<double> &v)
{
    std::vector<double> result;
    result.reserve(v.size());
    result.push_back(v[0]);
    for (size_t i=1; i < v.size(); i++) {
        double delta = v[i] - v[i-1];
        if (delta > M_PI)
            delta -= 2*M_PI;
        if (delta < -M_PI)
            delta += 2*M_PI;
        result.push_back(result[i-1] + delta);
    }
    return result;
}

static std::vector<double> groupDelay(const std::vector<double> &phase)
{
    const auto &v = phaseUnwrap(phase);
    std::vector<double> result;
    result.reserve(v.size());
    result.push_back(0);
    for (size_t i=1; i < v.size(); i++) {
        result.push_back(v[i] - v[i-1]);
    }
    return result;
}

void FirDesignDialog::replotFIRFFT ()
{
    Q_D(FirDesignDialog);

    WaveformPlot *wp = ui->plotFIRFFT; // FIR impulse FFT
    const auto &fir = d->newDecoderConfig.channels.fir;

    if (fir.empty()) {
        wp->clear();
        return;
    }
    const std::vector<double> firCoef(fir.begin(), fir.end());

    const size_t N = firCoef.size();
    std::vector<std::complex<double> > firDft = d->fft.DFT(firCoef);
    firDft.resize(N/2+1); // result is symmetric
    const double dftScaler = 1. / sqrt(N);
    std::vector<double> px;
    std::vector<double> FirDftMagnitude, FirDftPhase;
    for (size_t i = 0; i < firDft.size(); i++) {
        px.push_back(double(i) / N);
        FirDftMagnitude.push_back(Decibels::val2db(std::abs(firDft[i]) * dftScaler));
        FirDftPhase.push_back(std::arg(firDft[i]));
    }
    WaveformDataAll wda;
    wda.push_back(WaveformData(px, FirDftMagnitude, tr("Filter FFT"), filterColorIndex));
    if (false) {
        const auto &FirDftPhaseUnwrap = phaseUnwrap(FirDftPhase);
        const auto &FirDftGroupDelay = groupDelay(FirDftPhaseUnwrap);
        wda.push_back(WaveformData(px, FirDftGroupDelay, tr("Group delay"), filterColorIndex+2));
        wda.push_back(WaveformData(px, FirDftPhaseUnwrap, tr("Filter phase"), filterColorIndex+4));
    }
    wp->setWaveformData(wda);
    wp->enableLegend(false);

    QwtText xTitle(tr("Frequency"));
    QwtText yTitle(tr("Amplitude, dB"));
    xTitle.setFont(font());
    yTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->setAxisTitle(QwtPlot::yLeft, yTitle);
    wp->replot();
}

struct gauss_param_t {
    double a = 0;
    double t = 0;
    double w = 1;
    double aerr = 1;
    double terr = 1;
    double werr = 1;
    double error = 1;
};

gauss_param_t FirDesignDialog::find_hit(const std::vector<double> &testResult) const
{
    if (testResult.size() < 3)
        return {};
    Q_D(const FirDesignDialog);
    const FitterGausProfile &fitprof = d->newDecoderConfig.fitprof;
    gauss_param_t p;
    FitterGaus fitterGaus(testResult.size());

    int peakIndex = 0;
    double peakValue = testResult[0];
    for (size_t i=0; i<testResult.size(); i++) {
        if (testResult[i] > peakValue) {
            peakIndex = i;
            peakValue = testResult[peakIndex];
        }
    }
    //double err[3] = { 0, 0, 0 };
    // const CalAvgChannelConfig &calAvgChannel = d->newDecoderConfig.calavg.channels;
    int fs = fitprof.get_fit_half_np(d->firDesignConfig.pwid);
    const auto N = testResult.size();
    int xl = peakIndex - fs;
    int xr = peakIndex + fs;
    if (xl < 0) xl = 0;
    if (xr >= int(N)) xr = int(N) - 1;
    qDebug() << QString("peakIndex=%1, fit range %2:%3").arg(peakIndex).arg(xl).arg(xr);
    //    fitstatus_t fitstatus;
    //    fitstatus.niter = 100;
    //    FindGausParam (0, TSTpy, rlen, par, &fitstatus);
    std::vector<double> fitData;
    for (int i=xl; i<xr; i++)
        fitData.push_back(testResult[i]);

    //    GausPar par = GaussianApproximator::FindGausApproxShifted (fitData);
    //    p.a = par.a;
    //    p.t = xl + par.x0;
    //    p.w = par.sigma;

    {
        double par[3]={0,0,0};
        fitstatus_t fitstatus;
        fitstatus.niter = fitprof.get_dec_fititer();
        fitterGaus.fit (fitprof.get_dec_fitfixpar(), &fitData[0], fitData.size(), par, &fitstatus);
        p.a = par[0];
        p.t = par[1];
        p.w = par[2];
        double sum2 = 0;
        double sumerr2 = 0;
        for (size_t i=0; i<fitData.size(); i++) {
            double y = fitData[i];
            double yfit = p.a * gauss(i - p.t, p.w);
            sum2 += pow(y, 2);
            sumerr2 += pow(y - yfit, 2);
        }
        sum2 /= fitData.size();
        sumerr2 /= fitData.size();
        p.error = sqrt(sumerr2 / sum2);
        qDebug() << sum2 << sumerr2 << p.error;
    }
    p.t += xl;

    //    QuickFitGaus (&TSTpy[xl], xr-xl, par, &fitstatus);
    //    printf("(2) par = %f, %f, %f; xp=%d, fs=%d, newrange=%d-%d\n", par[0], par[1]+xl, par[2], xp, fs, xl, xr);
    //    printf("(3) par err = %f, %f, %f\n", (apar[0]-par[0]), (apar[1]-(par[1]+xl)), (apar[2]-par[2]));

    //	printf ("Expected latency: %f\n", (double)lat);
    //	printf ("Measured latency: %f\n", par[1]+xl);
    return p;
}

void FirDesignDialog::replotTEST ()
{
    Q_D(FirDesignDialog);

    WaveformPlot *wp = ui->plotTEST; // Test impulse

    const auto &fir = d->newDecoderConfig.channels.fir;
    if (fir.empty()) {
        wp->clear();
        return;
    }


    const auto &inp = getCalAvgY();

    auto testResult = Convolver::convolve (fir, inp);
    if (testResult.size() >= fir.size())
        testResult.resize(testResult.size() - fir.size()); // trim tail (remove edge effects)

    const auto lat = data_latency() + fir.size()/2;

    //    QwtText yTitle(tr("Amplitude"));
    //    yTitle.setFont(font());
    //    wp->setAxisTitle(QwtPlot::yLeft, yTitle);

    //printf("replot ok\n");

//    const auto &calAvgChannel = d->newDecoderConfig.calavg;
//    if (calAvgChannel.samples.empty() || testResult.empty()) {
//        ui->textLabelFitStatus->clear ();
//        wp->clear();
//        wp->replot();
//        return;
//    }

    // find peak value
    const gauss_param_t &par = find_hit(testResult);
    const QString plusMinus = QString::fromUtf8("\u00B1");
    QString text_a = tr("Ampl: %1 %2 %3")
                       .arg(par.a, 0, 'f', 1)
                       .arg(plusMinus)
                       .arg(par.aerr, 0, 'f', 1);
    QString text_t = tr("Pos: %1 %2 %3")
                         .arg((par.t/* + xl*/ - lat), 0, 'f', 1)
                         .arg(plusMinus)
                         .arg(par.terr, 0, 'f', 2);
    QString text_w = tr("Width: %1 %2 %3")
                         .arg(GausFWHM * par.w, 0, 'f', 2)
                         .arg(plusMinus)
                         .arg(par.werr, 0, 'f', 2);
    QString text_err = tr("Rel. error: %1")
                         .arg(par.error, 0, 'e', 2);
    ui->textLabelFitStatus->setText (text_a + "<br>" +  text_t + "<br>" + text_w + "<br>" + text_err);

    std::vector<double> px;
    std::vector<double> py;
    std::vector<double> perr;
    for (size_t i = 0; i < testResult.size(); i++) {
        px.push_back(i - lat);
        double y = testResult[i];
        py.push_back(y);
        double yfit = par.a * gauss(i - par.t, par.w);
        perr.push_back((y - yfit) * 100);
    }

    WaveformDataAll wda;
    wda.push_back(WaveformData(px, py, tr("Result waveform"), testColorIndex));
    wda.push_back(WaveformData(px, perr, tr("Residuals waveform * 100"), residualsColorIndex));
    wp->setWaveformData(wda);
    wp->enableLegend(false);

    QwtText xTitle(tr("Time"));
    xTitle.setFont(font());
    wp->setAxisTitle(QwtPlot::xBottom, xTitle);
    wp->replot();
}

void FirDesignDialog::Restore ()
{
    Q_D(FirDesignDialog);
    if (restoring)
        return;

    restoring = 1;
    {
        RecursiveWidgetSignalBlocker blocker(this);

        ui->sbpos->setValue (d->firDesignConfig.ppos);
        ui->sbwid->setValue (d->firDesignConfig.pwid);
        ui->sbexp->setValue (d->firDesignConfig.expfall);
        /*
  CreateFIR();
  calcRES();

        */

        //    ui->plotINPFFT->setAxisScale (QwtPlot::yLeft, INPFFT_MIN, INPFFT_MAX);
        //    QwtScaleEngine *scaleEngine = ui->plotINPFFT->axisScaleEngine(QwtPlot::yLeft);
        //    scaleEngine->setReference(val2db(INPFFT_MIN));
        ui->sbsnf->setRange (snf_db_min, snf_db_max);
        ui->sbrnf->setRange (snf_db_min, snf_db_max);
        ui->sbsnf->setValue (d->firDesignConfig.snf);
        ui->sbrnf->setValue (d->firDesignConfig.pnf);

        for (int i=0; i<ui->comboBoxWindowingFunction->count(); i++) {
            if (ui->comboBoxWindowingFunction->itemData(i) == win_type())
                ui->comboBoxWindowingFunction->setCurrentIndex(i);
        }
    }

    on_pushButtonGenerateFir_clicked ();
    Redraw ();
    restoring = 0;
    //  printf("Restore... Ok\n");
}

void FirDesignDialog::Redraw ()
{
    //printf("Redraw\n");

    replotINP ();
    replotINPFFT ();
    replotFIR ();
    replotFIRFFT ();
    replotTEST ();
}

void FirDesignDialog::clearFir()
{
    Q_D(FirDesignDialog);

    d->newDecoderConfig.channels.fir.clear();
//    d->newDecoderConfig.channels.deconv.fir.coef.push_back(1.0);
    replotFIR();
    replotFIRFFT();
    replotTEST();
}

static std::vector<double> truncateSymmetric(const std::vector<double> &in, int length)
{
    // FIR truncation. FIR will have fixed number of samples (np)
    auto df(in);
    std::vector<double> result;
    const int olen = df.size();

    int np = std::clamp(length, 0, olen);
    int i1 = (olen - np) / 2;
    int i2 = (olen + np) / 2;

    for (int i = i1; i < i2; i++) {
        df[i - i1] = df[i];
    }

    df.resize(i2 - i1); // np
    return df;
}

static std::vector<double> applyWindowFunction(const std::vector<double> &v, WindowFunction::fwinf_t windowFunction)
{
    std::vector<double> result(v.size());
    for (unsigned int i = 0; i < result.size(); i++) {
        result[i] = v[i] * WindowFunction::value(windowFunction, i, v.size() - 1);
    }
    return result;
}

void FirDesignDialog::on_pushButtonGenerateFir_clicked ()
{
    Q_D(FirDesignDialog);

    if (restoring)
        return;
    //printf("CreateFIR\n");

    if (d->newDecoderConfig.calavg.samples.empty()) {
        //    log_notice ("ch %d, No average waveform, please calibrate average first", ch);
        return;
    }
    const auto &calAvg = getCalAvgY();
    const auto n = calAvg.size();
    auto req = getRefY_n(getCalAvgAmplitude(), n);
    if (req.size() > calAvg.size())
        req.resize(calAvg.size());
    Q_ASSERT(calAvg.size() == req.size());

    // double orig_thr = db2val (d->newDecoderConfig.channels[ch].deconv.snf);
    // double req_thr = db2val (d->newDecoderConfig.channels[ch].deconv.pnf);

    auto fullFir = FirDesign::create_deconvolution_dblFIR(calAvg, req);

    auto df = truncateSymmetric(fullFir, ui->sbfirnp->value());

    df = applyWindowFunction(df, win_type());

    // After windowing, first and last samples may be near zero
    // TODO

    // Copy new fir to cfg
//    intfir_t newFIR;
//    newFIR.coef.resize(df.klen);
//    newFIR.offs = df.offs;
//    newFIR.bits = FIR_COEFF_BITS;
//    Make_Discrete (&df.coef[0], &newFIR.coef[0], df.size(), FIR_COEFF_BITS);
    d->newDecoderConfig.channels.fir = df; // newFIR;

    /*
  if (!restoring)
    Restore ();
*/
    //  Restore(); -- is called from Restore() --
    replotFIR ();
    replotFIRFFT ();
    replotTEST ();
}

double FirDesignDialog::reqf (double i) const
{
    double ampl = 1; // 0.5 * BITS2MAXUINT (adcConfig.bits);
    return ampl * reqfunscaled(i);
}

double FirDesignDialog::reqfunscaled(double i) const
{
    Q_D(const FirDesignDialog);

    const int nl = 10;

    double pos = d->firDesignConfig.ppos;
    double wid = d->firDesignConfig.pwid;
    double value = 0;
    for (int l = 0; l < nl; l++)
        value += 1. * gauss(i + (1. * l / nl - .5) - pos, wid / GausFWHM);
    value /= nl;
    return value;
}

//--------- Slot actions -----------------------------------------------------------------

void FirDesignDialog::on_comboBoxWindowingFunction_activated (int item)
{
    Q_D(FirDesignDialog);

    const auto new_win_type = static_cast<WindowFunction::fwinf_t>(ui->comboBoxWindowingFunction->itemData(item).toInt());
    if (new_win_type == win_type()) return;
    d->firDesignConfig.fwinf = new_win_type;
    on_pushButtonGenerateFir_clicked ();
}

void FirDesignDialog::on_sbfirnp_valueChanged (int)
{
    //  if (restoring) return;
    on_pushButtonGenerateFir_clicked ();
}


void FirDesignDialog::on_sbsnf_valueChanged (double v)
{
    Q_D(FirDesignDialog);
    if (restoring)
        return;
    d->firDesignConfig.snf = v;
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
    setButtons();
}

void FirDesignDialog::on_sbrnf_valueChanged (double v)
{
    Q_D(FirDesignDialog);
    if (restoring)
        return;
    d->firDesignConfig.pnf = v;
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
    setButtons();
}

void FirDesignDialog::on_sbpos_valueChanged (double v)
{
    Q_D(FirDesignDialog);
    d->firDesignConfig.ppos = v;
    //  Redraw();
    replotINP ();
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
}

void FirDesignDialog::on_sbwid_valueChanged (double v)
{
    Q_D(FirDesignDialog);
    d->firDesignConfig.pwid = v;
    //  Redraw();
    replotINP ();
    replotINPFFT ();
    on_pushButtonGenerateFir_clicked ();
}

void FirDesignDialog::on_sbexp_valueChanged (double v)
{
    Q_D(FirDesignDialog);
    d->firDesignConfig.expfall = v;

    on_pushButtonGenerateFir_clicked ();
    //  Redraw();
    replotINP ();
    replotINPFFT ();
    // replotTEST ();
}

double FirDesignDialog::data_latency () const
{
    Q_D(const FirDesignDialog);
    return d->newDecoderConfig.calavg.offset;
}

template<class T>
QStringList vectorToStringList(const QVector<T> &values)
{
    QStringList sl;
    for (const auto &value : values) {
        sl += QString("%1").arg(value);
    }
    return sl;
}

//--------------- Button actions --------------------------------------------------------

void FirDesignDialog::writeFirFile()
{
    QString title("Save FIR filter...");
    QString fileName = getSettings(k_firFileName).toString();
    QString selectedFilter = getSettings(k_firSelectedFilter).toString();
    fileName = FileDialog::getSaveFileName(
        this, title, fileName, k_textFileFilter, &selectedFilter);
    if (fileName.isNull())
        return;

    if (!fileName.endsWith(".txt"))
        fileName += ".txt";

    Q_D(const FirDesignDialog);
    writeNumbersFile(fileName, d->newDecoderConfig.channels.fir);
    setSettings(k_firFileName, fileName);
    setSettings(k_firSelectedFilter, selectedFilter);
}

void FirDesignDialog::readFirFile ()
{
    Q_D(FirDesignDialog);

    QString title("Open FIR filter...");
    QString fileName = getSettings(k_firFileName).toString();
    QString selectedFilter = getSettings(k_firSelectedFilter).toString();
    fileName = FileDialog::getOpenFileName(
        this, title, fileName, k_textFileFilter, &selectedFilter);
    if (fileName.isNull())
        return;
    setSettings(k_firFileName, fileName);
    setSettings(k_firSelectedFilter, selectedFilter);
    const auto &fir = readNumbersFile(fileName);
    if (fir.empty()) {
        return;
    }
    d->newDecoderConfig.channels.fir = std::vector<double>(fir.begin(), fir.end());
    Restore ();
    auto text = QString("Read file\n'%1':\n\n%2")
                    .arg(fileName)
                    .arg(vectorToStringList(fir).join(", "));
    QMessageBox::information(this, title, text, QMessageBox::Ok | QMessageBox::Default, Qt::NoButton);
}

template<typename Container>
bool FirDesignDialog::writeNumbersFile(const QString &fileName, const Container &data)
{
    QFile file(fileName);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
        QMessageBox::warning(this, "Error", file.errorString());
        return {};
    }
    QTextStream out(&file);
    for (auto i : data) {
        out << i << "\n";
    }
    if (out.status() != QTextStream::Ok) {
        static const QMap map{QTextStreamStatusMap};
        QMessageBox::warning(this, "Error", QString("Error writing file %1\n\n%2")
                                                    .arg(fileName).arg(map[out.status()]));
        return false;
    }
    return true;
}

QVector<double> FirDesignDialog::readNumbersFile(const QString fileName)
{
    QFile file(fileName);
    if (!file.open(QIODevice::ReadOnly)) {
        QMessageBox::warning(this, "Error", file.errorString());
        return {};
    }
    QTextStream in(&file);
    QVector<double> v;
    while (!in.atEnd()) {
        double d;
        in >> d;
        if (in.atEnd())
            break;
        if (in.status() != QTextStream::Ok) {
            static const QMap map{QTextStreamStatusMap};
            QMessageBox::warning(this, "Error", QString("Error reading file %1\n\n%2")
                                                        .arg(fileName).arg(map[in.status()]));
            break;
        }
        v.push_back(d);
    }
    return v;
}

void FirDesignDialog::makeTestShape()
{
    int len = 100;
    QVector<double> py;
    double dt = 1.;
    double rt = 1.5;
    double ft = 15.;
    double pulse_t = ranuni(3.0, len - 15.0); // pulse time offset inside waveform slide
    for (int i = 0; i < len; i++) {
        double d;
        d = 0.0 + detpulse (1.0, i*dt-pulse_t, rt, ft) + 1e-6 * rangaus(1.);
        // d = sin(2 * M_PI * i / len * 3) + 1e-6 * rangaus(1.);
        d = (int) (d * 32768.0) / 32768.0;
        py.push_back(d);
    }

    CalAvgConfig channel;
    channel.samples = py;
    channel.offset = pulse_t; // dt*len;
    Q_D(FirDesignDialog);
    d->newDecoderConfig.calavg = channel;
    Restore ();
}

void FirDesignDialog::readShapeFile()
{
    Q_D(FirDesignDialog);

    QString title("Load average shape");
    QString fileName = getSettings(k_shapeFileName).toString();
    QString selectedFilter = getSettings(k_shapeSelectedFilter).toString();
    fileName = FileDialog::getOpenFileName(
        this, title, fileName, k_textFileFilter, &selectedFilter);
    if (fileName.isNull())
        return;
    setSettings(k_shapeFileName, fileName);
    setSettings(k_shapeSelectedFilter, selectedFilter);
    const auto &shape = readNumbersFile(fileName);
    if (shape.empty()) {
        return;
    }
    CalAvgConfig channel;
    channel.samples = shape;
    channel.offset = 5.0;
    d->newDecoderConfig.calavg = channel;
    Restore ();
    auto text = QString("Read file\n'%1':\n\n%2")
                    .arg(fileName)
                    .arg(vectorToStringList(shape).join(", "));
    QMessageBox::information(this, title, text, QMessageBox::Ok | QMessageBox::Default, Qt::NoButton);
}

void FirDesignDialog::writeShapeFile ()
{
    QString title("Save average shape");
    QString fileName = getSettings(k_shapeFileName).toString();
    QString selectedFilter = getSettings(k_shapeSelectedFilter).toString();
    fileName = FileDialog::getSaveFileName(
        this, title, fileName, k_textFileFilter, &selectedFilter);
    if (fileName.isNull())
        return;

    if (!fileName.endsWith(".txt"))
        fileName += ".txt";

    Q_D(const FirDesignDialog);
    writeNumbersFile(fileName, d->newDecoderConfig.calavg.samples);
    setSettings(k_shapeFileName, fileName);
    setSettings(k_shapeSelectedFilter, selectedFilter);
}

bool FirDesignDialog::modified() const
{
    Q_D(const FirDesignDialog);
    return d->currentDecoderConfig != d->newDecoderConfig;
}

void FirDesignDialog::apply()
{
    Q_D(FirDesignDialog);
    emit decoderConfigChanged(d->newDecoderConfig);
    d->currentDecoderConfig = d->newDecoderConfig;
    setButtons();
}

void FirDesignDialog::setButtons ()
{
    Q_D(const FirDesignDialog);
    ui->buttonBox->button(QDialogButtonBox::Apply)->setEnabled(modified());
    ui->buttonBox->button(QDialogButtonBox::RestoreDefaults)->setEnabled(d->newDecoderConfig != DecoderConfig());
}

void FirDesignDialog::restoreDefaults()
{
    Q_D(FirDesignDialog);
    d->newDecoderConfig = {};
    setButtons();
    Restore();
}

void FirDesignDialog::on_buttonBox_clicked(QAbstractButton* button)
{
    switch(ui->buttonBox->standardButton(button)) {
    case QDialogButtonBox::Ok :
    case QDialogButtonBox::Apply :
        apply();
        break;
    case QDialogButtonBox::RestoreDefaults :
        restoreDefaults();
        break;
    default:;
    }
}

void FirDesignDialog::on_buttonBoxShape_clicked(QAbstractButton *button)
{
    switch (ui->buttonBoxShape->standardButton(button)) {
    case QDialogButtonBox::Reset:
        makeTestShape();
        return;
    case QDialogButtonBox::Open:
        readShapeFile();
        return;
    case QDialogButtonBox::Save:
        writeShapeFile();
        return;
    default:
        Q_UNREACHABLE();
    }
}

void FirDesignDialog::on_buttonBoxFir_clicked(QAbstractButton *button)
{
    switch (ui->buttonBoxFir->standardButton(button)) {
    case QDialogButtonBox::Reset:
        clearFir();
        return;
    case QDialogButtonBox::Open:
        readFirFile();
        return;
    case QDialogButtonBox::Save:
        writeFirFile();
        return;
    default:
        Q_UNREACHABLE();
    }
}
