/*
**    Copyright 2020 Ivan Filippov
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceTableModel.h"

#include <QHostAddress>
#include <QPalette>

#include "app/AppStyle.h"

static const int cacheSizeLimit = 1000; // fail-safe

void DeviceTableModel::lldpCacheUpdated(const LldpInfoCache &cache)
{
    if (cache.size() > cacheSizeLimit)
        return;
    if (lldpCache == cache)
        return;
    const DeviceIndexSet &oldKeys = devices.keys();
    const DeviceIndexSet &newKeys = cache.keys().toSet();
    const auto toAdd = newKeys - oldKeys;
    bool needReset = !toAdd.empty();
    if (needReset)
        beginResetModel();
    lldpCache = cache;
    for (auto it = cache.begin(); it != cache.end(); ++it) {
        DeviceTableEntry ent(
                    it.key(),
                    it.value(),
                    DeviceTableEntry::Dynamic,
                    DeviceTableEntry::EnableDefault
                    );
        if (!devices.contains(ent.index));
        devices.insertDynamic(ent);
        if (!needReset) {
            const auto topLeft = createIndex(rowOf(ent.index), 0);
            const auto botRight = createIndex(rowOf(ent.index), columnCount() - 1);
            emit dataChanged(topLeft, botRight);
        }
    }
    if (needReset)
        endResetModel();
}

QVariant DeviceTableModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid())
        return {};

    if (index.row() >= devices.size())
        return {};

    auto entIt = devices.find(index.row());
    if (entIt == devices.end())
        return {};
    auto dev = *entIt;

    switch (role) {
    case Qt::CheckStateRole: {
        switch (index.column()) {
        case BaseDeviceTableModel::COL_ENABLE:
            return dev.enabled != DeviceTableEntry::Disable ? Qt::Checked
                                                            : Qt::Unchecked;
        default:
            return QVariant();
        }
    }
    case Qt::BackgroundRole:
        if (dev.lldpInfo.hasExpired())
            return AppStyle::tableBg(Severity::Unknown);
    case Qt::DisplayRole:
        return getDisplayData(index);
    case Qt::ToolTipRole:
        return getToolTipData(index);
    case Qt::UserRole:
        return getUserData(index);
    }

    return QVariant();
}

QVariant DeviceTableModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (static_cast<Columns>(section)) {
        case COL_ENABLE:
            return "En";
        case COL_ONLINE:
            return "Online";
        case COL_DYNAMIC:
            return "Static";
        case COL_DEVICE_INDEX:
            return "Index";
        case COL_DEVICE_ID:
            return "Type";
        case COL_SERIAL:
            return "Serial";
        case COL_FIRM_WARE:
            return "Firmware";
        case COL_SLOT:
            return "Slot";
        case COL_IP:
            return "IP Address";
        case COL_MAC:
            return "MAC Address";
        case COL_MASTER:
            return "Master";
        case COL_M_STREAM:
            return "MStream";
        case COL_TOTAL: return {};
        }
    }
    return DeviceIndexTableModel::headerData(section, orientation, role);
}

bool DeviceTableModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    const auto it = devices.find(index.row());
    if(it == devices.end())
        return false;

    if(role == Qt::CheckStateRole) {
        if(index.column()==COL_ENABLE) {
            it->enabled = (value == Qt::Checked) ? DeviceTableEntry::Enable
                                                 : DeviceTableEntry::Disable;
            emit dataChanged(index, index);
            return true;
        }
    }
    if(role == Qt::DisplayRole) {
        if(index.column()==COL_DYNAMIC) {
            it->dynamic = value.value<DeviceTableEntry::DynamicFlag>();
            emit dataChanged(index, index);
            return true;
        }
        if(index.column()==COL_ENABLE) {
            it->enabled = value.value<DeviceTableEntry::EnableFlag>();
            emit dataChanged(index, index);
            return true;
        }
    }
    return false;
}

QVariant DeviceTableModel::getDisplayData(const QModelIndex &index) const
{
    auto entIt = devices.find(index.row());
    if (entIt == devices.end())
        return {};

    const auto &deviceIndex = entIt->index;
    const auto &lldp = entIt->lldpInfo.valid ? entIt->lldpInfo : LldpInfo();

    // no LLDP info = return empty
    //    if (1
    //            && index.column() != COL_STATE
    //            && index.column() != COL_TYPE_NUM
    //            && index.column() != COL_SERIAL_NUM) {
    //        if (lldp.isNull())
    //            return QVariant();
    //    }

    switch(index.column()) {
    case COL_ENABLE:
        return entIt->enabled;
    case COL_ONLINE:
        if (entIt->lldpInfo.hasExpired()) return "Offline";
        return QString::number(entIt->lldpInfo.msToExpire()*1e-3, 'f', 1);
    case COL_DYNAMIC:
        return entIt->dynamic;
    case COL_DEVICE_INDEX:
        return deviceIndex.getIdent();
    case COL_DEVICE_ID:
    {
        QString modelStr = lldp.getModelVerStr();
        if (lldp.med.modelname_str.isEmpty()) {
            // modelStr = deviceTypeMap.values(lldp.device_id).join("/");
            modelStr = deviceIndex.getDeviceIdName();
        }
        return modelStr;
    }
    case COL_SERIAL:
        if (lldp.med.sn_str.isEmpty())
            return deviceIndex.getSerialStr();
        return lldp.med.sn_str;
    case COL_FIRM_WARE:
        return lldp.med.fw_str;
    case COL_SLOT:
        if (!lldp.ext.chassis_slot) // slot '0' is not valid
            return QVariant();
        return lldp.ext.chassis_slot;
    case COL_IP:
    {
        if (!lldp.common.ip_addr)
            return QVariant();
        return QHostAddress(lldp.common.ip_addr).toString();
    }
    case COL_MAC:
    {
        if (!lldp.common.haddr)
            return QVariant();
        QString mac_adress = QString("%1").arg(lldp.common.haddr, 12, 16, QChar('0'));
        mac_adress = mac_adress.toUpper();
        const int step = 2;
        for(int i=step; i< mac_adress.size(); i+= step+1){
            mac_adress.insert(i, ":");
        }
        return mac_adress;
    }
    case COL_MASTER:
        return makeIpPortStr(lldp.ext.master_locked, lldp.ext.master.ip, lldp.ext.master.port);
    case COL_M_STREAM:
        return makeIpPortStr(lldp.ext.mstream_locked, lldp.ext.mstream.ip, lldp.ext.mstream.port);
    }

    return QVariant();
}

QVariant DeviceTableModel::getToolTipData(const QModelIndex &index) const
{
    return QVariant();
}

QVariant DeviceTableModel::getUserData(const QModelIndex &index) const
{
    auto entIt = devices.find(index.row());
    if(entIt == devices.end())
        return QVariant();

    const auto &deviceIndex = entIt->index;

    switch(index.column()) {
    case COL_ONLINE: {
        const auto &info = entIt->lldpInfo;
        return info.valid && !info.hasExpired();
    }
    case COL_DEVICE_INDEX:
        return QVariant::fromValue(deviceIndex);
    case COL_DEVICE_ID:
        return deviceIndex.getDevId();
    case COL_SERIAL:
        return deviceIndex.getSerial();
    default:
        return QVariant();
    }
}

QString DeviceTableModel::makeIpPortStr(bool locked, quint32 ip, quint16 port) const
{
    if (locked) {
        auto ipStr = QHostAddress(ip).toString();
        return QString("%1:%2").arg(ipStr).arg(port);
    }
    return QString();
}
