/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "LldpDiscoveryTableModel.h"

#include <QHostAddress>
#include <QColor>
#include <QTimer>

#include "app/AppStyle.h"

static const int cacheSizeLimit = 1000; // fail-safe
static const int pollDataTimeout = 300;

LldpDiscoveryTableModel::LldpDiscoveryTableModel(QObject *parent)
    : DeviceIndexTableModel(parent)
    , pollDataTimer(new QTimer(this))
{
    pollDataTimer->setSingleShot(true);
    pollDataTimer->start(pollDataTimeout);
    connect(pollDataTimer, &QTimer::timeout, this, &LldpDiscoveryTableModel::pollData);
}

void LldpDiscoveryTableModel::insertRow(const DeviceIndex &deviceIndex, const LldpInfo &data)
{
    int newRow = 0;
    for (auto it = lldpData.cbegin(); it != lldpData.cend(); ++it) {
        newRow++;
        if (deviceIndex < it.key()) {
            break;
        }
    }
    qDebug() << __PRETTY_FUNCTION__ << newRow;
    beginInsertRows({}, newRow, newRow);
    lldpData[deviceIndex] = data;
    endInsertRows();
}

void LldpDiscoveryTableModel::removeRow(const DeviceIndex &deviceIndex)
{
    int row = rowOf(deviceIndex);
    beginRemoveRows({}, row, row);
    lldpData.remove(deviceIndex);
    endRemoveRows();
}

void LldpDiscoveryTableModel::lldpCacheUpdated(const LldpInfoCache &input)
{
    if (input.size() > cacheSizeLimit)
        return;
    pendingData = input;
}

void LldpDiscoveryTableModel::checkPendingData()
{
    if (!pendingData.has_value())
        return;
    const auto &newData = pendingData.value();

    const auto &oldKeys = lldpData.keys();
    const auto &newKeys = newData.keys();
    if (oldKeys != newKeys) {
        // add/remove rows, update
        if (false) {
            beginResetModel();
            lldpData = newData;
            endResetModel();
            return;
        } else {
            const DeviceIndexSet oldSet(oldKeys.begin(), oldKeys.end());
            const DeviceIndexSet newSet(newKeys.begin(), newKeys.end());
            const auto toAdd = newSet - oldSet;
            const auto toRemove = oldSet - newSet;
            for (auto i : toRemove) {
                removeRow(i);
            }
            for (auto i : toAdd) {
                insertRow(i, newData[i]);
            }
        }
    }

    // update existing rows
    for (auto it = newData.begin(); it != newData.end(); ++it) {
        const auto &deviceIndex = it.key();
        if (lldpData[deviceIndex] == it.value())
            continue;
        lldpData[deviceIndex] = it.value();
        invalidateRow(deviceIndex);
    }
}

void LldpDiscoveryTableModel::pollData()
{
    checkPendingData();
    checkForExpire();
    pollDataTimer->start();
}

void LldpDiscoveryTableModel::checkForExpire()
{
    for (auto it = lldpData.begin(); it != lldpData.end(); ++it) {
        const auto &deviceIndex = it.key();
        bool newState = it.value().hasExpired();
        if (!newState) {
            // update countdown
            int row = rowOf(deviceIndex);
            auto indexFrom = createIndex(row, COL_ONLINE);
            emit dataChanged(indexFrom, indexFrom);
        }
        if (newState != expiredState[deviceIndex]) {
            expiredState[deviceIndex] = newState;
            invalidateRow(deviceIndex);
        }
    }
}

int LldpDiscoveryTableModel::rowCount(const QModelIndex &parent) const
{
    return lldpData.size();
}

int LldpDiscoveryTableModel::columnCount(const QModelIndex &parent) const
{
    return COL_TOTAL;
}

DeviceIndex LldpDiscoveryTableModel::deviceIndexOf(int row) const
{
    int r = 0;
    for (auto it = lldpData.begin(); it != lldpData.end(); ++it)
        if (r++ == row)
            return it.key();
    return {};
}

int LldpDiscoveryTableModel::rowOf(const DeviceIndex &deviceIndex) const
{
    auto it = lldpData.find(deviceIndex);
    if (it == lldpData.end())
        return -1;
    auto row = std::distance(lldpData.begin(), it);
    return static_cast<int>(row);
}

QVariant LldpDiscoveryTableModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid())
        return {};
    int row = index.row();
    if (row < 0)
        return {};
    const auto &deviceIndex = deviceIndexOf(row);
    const auto &data = lldpData[deviceIndex];
    int column = index.column();

    switch (role) {
    case Qt::BackgroundRole:
        if (data.hasExpired())
            return AppStyle::tableBg(Severity::Unknown);
    case Qt::DisplayRole:
        return getDisplayData(column, deviceIndex);
    case Qt::ToolTipRole:
        return getToolTipData(column, deviceIndex);
    case Qt::UserRole:
        return getUserData(column, deviceIndex);
    }

    return {};
}

QVariant LldpDiscoveryTableModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch(section) {
        case COL_ONLINE:
            return "Online";
        case COL_DEVICE_INDEX:
            return "Index";
        case COL_DEVICE_ID:
            return "Type";
        case COL_SERIAL:
            return "Serial";
        case COL_HW_VER:
            return "h/w";
        case COL_FIRM_WARE:
            return "Firmware";
        case COL_SLOT:
            return "Slot";
        case COL_IP:
            return "IP Address";
        case COL_MAC:
            return "MAC Address";
        case COL_MASTER:
            return "Master";
        case COL_M_STREAM:
            return "MStream";
        }
    }

    return DeviceIndexTableModel::headerData(section, orientation, role);
}

QMap<int, bool> LldpDiscoveryTableModel::defaultColumnState() const
{
    return {
        {COL_ONLINE, false},
        {COL_DEVICE_INDEX, false},
        {COL_DEVICE_ID, false},
        {COL_SERIAL, false},
        {COL_HW_VER, false},
        {COL_FIRM_WARE, false},
        // {COL_SLOT, false},
        {COL_IP, false},
        {COL_MAC, false},
        {COL_MASTER, false},
        {COL_M_STREAM, false}
    };
}

QVariant LldpDiscoveryTableModel::getDisplayData(int column, const DeviceIndex &deviceIndex) const
{
    const auto &lldp = lldpData[deviceIndex];

    switch(column) {
    case COL_ONLINE:
        if (lldp.hasExpired()) return "Offline";
        return QString::number(lldp.msToExpire()*1e-3, 'f', 1);
    case COL_DEVICE_INDEX:
        return deviceIndex.getIdent();
    case COL_DEVICE_ID:
    {
        QString modelStr = lldp.getModelVerStr();
        if (lldp.med.modelname_str.isEmpty()) {
            // modelStr = deviceTypeMap.values(lldp.device_id).join("/");
            modelStr = deviceIndex.getDeviceIdName();
        }
        return modelStr;
    }
    case COL_SERIAL:
        if (lldp.med.sn_str.isEmpty())
            return deviceIndex.getSerialStr();
        return lldp.med.sn_str;
    case COL_HW_VER:
        return lldp.med.hw_str;
    case COL_FIRM_WARE:
        return lldp.med.fw_str;
    case COL_SLOT:
        if (!lldp.ext.chassis_slot) // slot '0' is not valid
            return QVariant();
        return lldp.ext.chassis_slot;
    case COL_IP:
    {
        if (!lldp.common.ip_addr)
            return QVariant();
        return QHostAddress(lldp.common.ip_addr).toString();
    }
    case COL_MAC:
    {
        if (!lldp.common.haddr)
            return QVariant();
        QString mac_adress = QString("%1").arg(lldp.common.haddr, 12, 16, QChar('0'));
        mac_adress = mac_adress.toUpper();
        const int step = 2;
        for(int i=step; i< mac_adress.size(); i+= step+1){
            mac_adress.insert(i, ":");
        }
        return mac_adress;
    }
    case COL_MASTER:
        return makeIpPortStr(lldp.ext.master_locked, lldp.ext.master.ip, lldp.ext.master.port);
    case COL_M_STREAM:
        return makeIpPortStr(lldp.ext.mstream_locked, lldp.ext.mstream.ip, lldp.ext.mstream.port);
    }

    return {};
}

QVariant LldpDiscoveryTableModel::getToolTipData(int column, const DeviceIndex &deviceIndex) const
{
    return {};
}

QVariant LldpDiscoveryTableModel::getUserData(int column, const DeviceIndex &deviceIndex) const
{
    switch(column) {
    case COL_DEVICE_INDEX:
        return QVariant::fromValue(deviceIndex);
    case COL_DEVICE_ID:
        return deviceIndex.getDevId();
    case COL_SERIAL:
        return deviceIndex.getSerial();
    default:
        return {};
    }
}

QString LldpDiscoveryTableModel::makeIpPortStr(bool locked, quint32 ip, quint16 port) const
{
    if (locked) {
        auto ipStr = QHostAddress(ip).toString();
        return QString("%1:%2").arg(ipStr).arg(port);
    }
    return {};
}
