/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef LLDPDISCOVERY_H
#define LLDPDISCOVERY_H

#include <QMap>
#include <QMetaType>
#include <QObject>
#include <QPair>

#include <memory>

#include "base/DeviceIndex.h"
#include "LldpInfo.h"
#include "LldpDiscoveryFilter.h"

typedef enum {
    DeviceDiscovered,
    DeviceExpired,
    DeviceBusy
} DeviceState;

Q_DECLARE_METATYPE(DeviceState)

class LldpInfoCache: public QMap<DeviceIndex, LldpInfo> {};
class DeviceStatePair: public QPair<DeviceIndex, DeviceState> {
public:
    DeviceStatePair() = default;
    DeviceStatePair(const DeviceIndex&i, const DeviceState &s)
        : QPair<DeviceIndex, DeviceState>(i,s) {}
};

class LldpDiscoveryPrivate;

class LldpDiscovery : public QObject
{
    Q_OBJECT
public:
    // Meyer's Singleton, thread-safe since C++11
    static LldpDiscovery & instance()
    {
        static LldpDiscovery obj;
        return obj;
    }

    static LldpInfo lookup(const DeviceIndex &index, bool *Ok = nullptr) {
        return instance().lookup_internal(index, Ok);
    }
    LldpInfo lookup_internal(const DeviceIndex &index, bool *Ok = nullptr) const;

    // subscribe with permanent filter (post-processing after variable filter)
    template<class T>
    static void subscribe(const T *obj, const QSet<DeviceId> &id = {}) {
        instance().subscribe_internal(obj, id);
    }
    template<class T>
    void subscribe_internal(const T *obj, const QSet<DeviceId> &id = {}) {
        if (id.isEmpty()) {
            connect(this, &LldpDiscovery::cacheUpdated,
                    obj, &T::lldpCacheUpdated,
                    Qt::UniqueConnection);
        } else {
            auto filter = new LldpDiscoveryFilter(id, this);

            connect(this, &LldpDiscovery::cacheUpdated,
                    filter, &LldpDiscoveryFilter::lldpCacheUpdated,
                    Qt::UniqueConnection);

            connect(filter, &LldpDiscoveryFilter::lldpCacheUpdatedFiltered,
                    obj, &T::lldpCacheUpdated,
                    Qt::UniqueConnection);

            connect(obj, &T::destroyed,
                    filter, &QObject::deleteLater);
        }
    }

signals:
    void cacheUpdated(const LldpInfoCache &);
    void deviceStateChanged(const DeviceStatePair &);
    void setDeviceIdFilter(const DeviceIdSet &filter); // run-time configurable filter

protected:
    LldpDiscovery();
    ~LldpDiscovery() override;

private:
    Q_DISABLE_COPY_MOVE(LldpDiscovery)
    Q_DECLARE_PRIVATE(LldpDiscovery)
    LldpDiscoveryPrivate * const d_ptr;
};

Q_DECLARE_METATYPE(LldpInfoCache)
Q_DECLARE_METATYPE(DeviceStatePair)

#endif // LLDPDISCOVERY_H
