#include "AdcDataTestModule.h"

#include "base/fw_version.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

#include "AdcDataTestStatus.h"

namespace {
enum {
    REG_ID                 = 0x00,
    REG_CONTROL            = 0x02,
    REG_PATTERN            = 0x03,
    REG64_TEST_COUNT        = 0x04, // RO
    REG64_FAILED_MASK_LO   = 0x08, // RO
    REG64_FAILED_MASK_HI   = 0x0C  // RO
};

typedef union reg_control_t {
    quint16 all;
    struct {
        quint32 enable: 1; // [0]
        quint32 pause: 1; // [1]
        quint32 reserved: 14; // [15:2]
    } bit;
} reg_control_t;
constexpr const quint16 CTRL_BIT_ENABLE = 1;
constexpr const quint16 CTRL_BIT_PAUSE  = 2;
constexpr const quint16 CTRL_BIT_PRBS_EN = 4;
} // namespace

QList<AFI_SDB_ID> AdcDataTestModule::getSupportedSdbIds()
{
    return {AFI_SDB_ADC_DATA_TEST};
}

void AdcDataTestModule::post_init()
{
    regCtrl = 0;
    regWrite16(REG_CONTROL, regCtrl);
}

bool AdcDataTestModule::supportPRBS() const
{
    return getVersion() >= fw_version_t{1,1,0};
}

std::optional<AdcDataTestStatus> AdcDataTestModule::readStatus()
{
    AdcDataTestStatus result;
    if (!readStatus(&result))
        return {};
    return result;
}

bool AdcDataTestModule::readStatus(AdcDataTestStatus *result)
{
    AdcDataTestStatus status;
    RegOpVector r;
    r.RegRead32(REG_ID, &status.id);
//    r.RegWrite16(REG_CONTROL, CTRL_BIT_ENABLE | CTRL_BIT_PAUSE);
//    r.RegRead64(REG64_TEST_COUNT, &status.testCnt);
//    r.RegRead64(REG64_FAILED_MASK_LO, &status.failedChMask_lo);
//    r.RegRead64(REG64_FAILED_MASK_HI, &status.failedChMask_hi);
//    r.RegWrite16(REG_CONTROL, CTRL_BIT_ENABLE);
    if (result)
        *result = status;
    return regOpExecRebased(r);
}

bool AdcDataTestModule::readPatternStatus(AdcPatternStatus *result)
{
    AdcPatternStatus st;

    RegOpVector r;
    regCtrl |= CTRL_BIT_PAUSE;
    r.RegWrite16(REG_CONTROL, regCtrl);
    r.RegRead64(REG64_TEST_COUNT, &st.testCnt);
    r.RegRead64(REG64_FAILED_MASK_LO, &st.failedChMask_lo);
    r.RegRead64(REG64_FAILED_MASK_HI, &st.failedChMask_hi);
    regCtrl &= ~CTRL_BIT_PAUSE;
    r.RegWrite16(REG_CONTROL, regCtrl);

    if (!regOpExecRebased(r))
        return {};
    if (result)
        *result = st;
    return true;
}

bool AdcDataTestModule::setTestPatternAndEnable(quint16 pattern)
{
    RegOpVector r;
    r.RegWrite16(REG_CONTROL, 0);
    r.RegWrite16(REG_PATTERN, pattern);
    regCtrl = CTRL_BIT_ENABLE;
    r.RegWrite16(REG_CONTROL, regCtrl);
    return regOpExecRebased(r);
}

bool AdcDataTestModule::setPatternTestEn(bool en)
{
    RegOpVector r;
    r.RegWrite16(REG_CONTROL, 0);
    regCtrl = en ? CTRL_BIT_ENABLE : 0;
    r.RegWrite16(REG_CONTROL, regCtrl);
    return regOpExecRebased(r);
}

bool AdcDataTestModule::setPRBSEn(bool en)
{
    RegOpVector r;
    r.RegWrite16(REG_CONTROL, 0);
    regCtrl = en ? CTRL_BIT_PRBS_EN : 0;
    r.RegWrite16(REG_CONTROL, regCtrl);
    return regOpExecRebased(r);
}

quint64 AdcDataTestModule::getPatternTestCount()
{
    quint64 result;
    regRead64(REG64_TEST_COUNT, &result);
    return result;
}

quint64 AdcDataTestModule::getPatternTestFailedChWithStop()
{
    quint64 readval;
    RegOpVector r;
    r.RegWrite16(REG_CONTROL, 0);
    r.RegRead64(REG64_FAILED_MASK_LO, &readval);
    if (!regOpExecRebased(r))
        return {};
    return readval;
}

bool AdcDataTestModule::getPatternTestFailedChWithPause(quint64 *result)
{
    quint64 readval2;
    RegOpVector r;
    regCtrl |= CTRL_BIT_PAUSE;
    r.RegWrite16(REG_CONTROL, regCtrl);
    r.RegRead64(REG64_FAILED_MASK_LO, result);
    r.RegRead64(REG64_FAILED_MASK_HI, &readval2); //ignored at this moment
    regCtrl &= ~CTRL_BIT_PAUSE;
    r.RegWrite16(REG_CONTROL, regCtrl);
    return regOpExecRebased(r);
}
