#include "AdcMpdTrigModule.h"

#include "AdcMpdTrigConfig.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"
#include "CrosspointSwitch/CrosspointSwitchConfig.h"

namespace {
enum {
    REG_BASE_CSR                = 0x0,
    REG_BASE_ADC_STATIC_COMP    = 0x80,
    REG_BASE_CROSSPOINT_SWITCH  = 0x100,
    REG_BASE_TRIG_GROUP_THR     = 0x180,
};

// CSR
enum {
    REG_ID                  = 0x0,
    REG_COTROL              = 0x2,
    // REG_LOCK                = 0x3,
    REG_TRIG_GROUP_NUM      = 0x4,
    REG_CH_NUM_PER_GROUP    = 0x5,
    REG_ADC_CH_NUM          = 0x6,
    REG_TRIG_CH_EN          = 0x8, // 128-bits, R/W
    REG_TRIG_CH_POL         = 0x10, // 128-bits, R/W
};

} // namespace

AdcMpdTrigModule::AdcMpdTrigModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
}

QList<AFI_SDB_ID> AdcMpdTrigModule::getSupportedSdbIds()
{
    return {AFI_SDB_ADC_MPD_TRIG};
}

bool AdcMpdTrigModule::writeConfig(const AdcMpdTrigConfig &config)
{
    if(!buildParams || !crSwMod)
        return false;

    CrosspointSwitchConfig crSwCfg;
    auto bp = crSwMod->getBuildParam();
    if(!bp)
        return false;
    crSwCfg.out.resize(bp->outSize);

    RegOpVector r;
    QVector<quint16> dataChEn;
    if(buildParams->adcChNum>0)
        dataChEn.resize((buildParams->adcChNum-1)/16 + 1);
    auto dataChPol = dataChEn;

    for(auto group=0; group<buildParams->trigGroupNum; ++group) {
        auto grCfg = config.groups.value(group);
        for(auto ch=0; ch<buildParams->chNumPerGroup; ++ch) {
            auto chCfg = grCfg.channels.value(ch);
            auto absCh = group*buildParams->chNumPerGroup + ch;
            if(chCfg.en)
                dataChEn[absCh/16] |= 1<<(absCh%16);
            if(chCfg.pol)
                dataChPol[absCh/16] |= 1<<(absCh%16);
            crSwCfg.out[absCh] = chCfg.adcCh;
        }
        r.RegWrite32(REG_BASE_TRIG_GROUP_THR+group*2, grCfg.trigThr);
    }

    for(auto ch=0; ch<config.adcStaticCompensation.size(); ++ch)
        r.RegWrite16(REG_BASE_ADC_STATIC_COMP+ch, config.adcStaticCompensation.value(ch,0));

    for(auto i=0; i<(buildParams->adcChNum/16); ++i) {
        r.RegWrite16(REG_BASE_CSR+REG_TRIG_CH_EN+i, dataChEn[i]);
        r.RegWrite16(REG_BASE_CSR+REG_TRIG_CH_POL+i, dataChPol[i]);
    }

    bool ok = crSwMod->writeConfig(crSwCfg);
    ok &= regOpExecRebased(r);

    return ok;
}

void AdcMpdTrigModule::post_init()
{
    buildParams.reset();
    quint32 id;
    AdcMpdTrigBuildParams bp;

    RegOpVector r;
    r.RegRead32(REG_ID, &id);
    r.RegRead16(REG_TRIG_GROUP_NUM, &bp.trigGroupNum);
    r.RegRead16(REG_CH_NUM_PER_GROUP, &bp.chNumPerGroup);
    r.RegRead16(REG_ADC_CH_NUM, &bp.adcChNum);


    if(regOpExecRebased(r) && getSupportedSdbIds().contains(static_cast<AFI_SDB_ID>(id))) {
        buildParams = bp;
        MregBusInterface busI(bus.dev);
        busI.setAddrRange({baseAddr() + REG_BASE_CROSSPOINT_SWITCH, 0x7f});
        crSwMod.reset(new CrosspointSwitchModule(busI));
        crSwMod->init();
        auto crSwModBp = crSwMod->getBuildParam();
        Q_ASSERT(crSwModBp &&
                 crSwModBp->inSize == bp.adcChNum &&
                 crSwModBp->outSize == (bp.chNumPerGroup*bp.trigGroupNum));
        if(!crSwModBp ||
            crSwModBp->inSize != bp.adcChNum &&
                crSwModBp->outSize != (bp.chNumPerGroup*bp.trigGroupNum)) {
            crSwMod.reset();
        }
    } else {
        crSwMod.clear();
    }
}
