#include "DeserializersControlModule.h"

#include "base/DeviceId.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

#include "AdcSerDesDelay.h"

namespace {
enum {
    REG_DES_CTRL                = 0x00,
    REG_DES_STATUS              = 0x01,
    REG_DES_IDELAY_TAP_VAL      = 0x03,
    REG_DES_IDELAY_LOAD_MASK    = 0x04,
    REG_DES_ADC_CH_INFO         = 0x05, // RO
};

const quint16 ADC_CLK_CTRL_BIT_ADC_RESET      = 0x8000;
const quint16 ADC_CLK_CTRL_BIT_DES_RESET      = 0x4000;
const quint16 ADC_CLK_CTRL_BIT_STATUS_RESET   = 0x2000;
const quint16 ADC_CLK_CTRL_BIT_ADC_SYNC       = 0x1000; // to sync ADC after reconfiguration
const quint16 ADC_CLK_CTRL_BIT_RESET          = 0x0100; // unused

} // namespace

DeserializersControlModule::DeserializersControlModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
    if (!isAvailable()) {
        if (getDeviceId() == DEVICE_ID_PHOS_FEC)
            setBaseAddr(0x4000, 0xFF);
    }
}

QList<AFI_SDB_ID> DeserializersControlModule::getSupportedSdbIds()
{
    return {AFI_SDB_Deserializers_Contr};
}

void DeserializersControlModule::post_init()
{
    BaseDeviceModule::post_init();
    buildParams.reset();

    {
        DeserializersControlBuildParams bp;
        quint16 val;
        regRead16(REG_DES_ADC_CH_INFO, &val);
        bp.desCnt = val&0xFF;
        bp.chPerChip = val>>8;
        if (getDeviceId() == DEVICE_ID_TQDC16VS_E) {
            // remove it after introducing normal TADC module to TQDC16VS
            bp.desCnt = 4;
            bp.chPerChip = 4;
        }
        if(bp.isValid())
            buildParams = bp;
    }
}

RegOpVector DeserializersControlModule::getClockDelaySetRegOp(quint16 clockDelay) const
{
    RegOpVector r;
    r.RegWrite16(REG_DES_IDELAY_TAP_VAL, clockDelay);
    r.RegWrite16(REG_DES_IDELAY_LOAD_MASK, 0x5555);
    r.rebase(baseAddr());
    return r;
}

RegOpVector DeserializersControlModule::getDataDelaySetRegOp(quint16 dataDelay) const
{
    RegOpVector r;
    r.RegWrite16(REG_DES_IDELAY_TAP_VAL, dataDelay);
    r.RegWrite16(REG_DES_IDELAY_LOAD_MASK, 0xAAAA);
    r.rebase(baseAddr());
    return r;
}

RegOpVector DeserializersControlModule::getStatusRegOp(quint16 *result) const
{
    RegOpVector r;
    r.RegRead16(REG_DES_STATUS, result);
    return r;
}

std::optional<DeserializersControlStatus> DeserializersControlModule::readStatus()
{
    if(!buildParams)
        return {};

    quint16 result;
    auto r = getStatusRegOp(&result);
    if (!regOpExecRebased(r))
        return {};
    DeserializersControlStatus desSt;
    desSt.desCnt = buildParams->desCnt;
    desSt.setStatus(result);
    return desSt;
}

bool DeserializersControlModule::setDeserReset(bool en)
{
    updateDesCtrlVal();
    if(en)
        curDesCtrlVal |= ADC_CLK_CTRL_BIT_DES_RESET;
    else
        curDesCtrlVal &= ~ADC_CLK_CTRL_BIT_DES_RESET;

    return regWrite16(REG_DES_CTRL, curDesCtrlVal);
}

bool DeserializersControlModule::ResetAdcPin()
{
    return pulseCtrlBit(ADC_CLK_CTRL_BIT_ADC_RESET);
}

bool DeserializersControlModule::setAdcSync()
{
    return pulseCtrlBit(ADC_CLK_CTRL_BIT_ADC_SYNC);
}

bool DeserializersControlModule::ResetDeser()
{
    if (!pulseCtrlBit(ADC_CLK_CTRL_BIT_DES_RESET))
        return false;
    QThread::usleep(10);
    return true;
}

bool DeserializersControlModule::changeDesDelay(quint16 channel_mask, quint16 taps)
{
    RegOpVector r;
    r.RegWrite16(REG_DES_IDELAY_TAP_VAL, taps);
    r.RegWrite16(REG_DES_IDELAY_LOAD_MASK, channel_mask);
    return regOpExecRebased(r);
}

bool DeserializersControlModule::ResetAdcClkDelay()
{
    // this bit is unused in f/w at least since 10 Jun 2021
    return pulseCtrlBit(ADC_CLK_CTRL_BIT_RESET);
}

bool DeserializersControlModule::updateDesCtrlVal()
{
    return regRead16(REG_DES_CTRL, &curDesCtrlVal);
}

bool DeserializersControlModule::pulseCtrlBit(quint16 bit)
{
    if (!updateDesCtrlVal())
        return false;
    RegOpVector r;
    curDesCtrlVal |= bit;
    r.RegWrite16(REG_DES_CTRL, curDesCtrlVal);
    curDesCtrlVal &= ~bit;
    r.RegWrite16(REG_DES_CTRL, curDesCtrlVal);
    return regOpExecRebased(r);
}

bool DeserializersControlModule::ResetAdcStatus()
{
    return pulseCtrlBit(ADC_CLK_CTRL_BIT_STATUS_RESET);
}

bool DeserializersControlModule::setClockDelay(quint16 clkDelay)
{
    return changeDesDelay(0x5555, clkDelay);
}

bool DeserializersControlModule::setDataDelay(quint16 dataDelay)
{
    return changeDesDelay(0xAAAA, dataDelay);
}

bool DeserializersControlModule::changeDesDelay(const AdcSerDesDelayMap &delay)
{
    if (!ResetAdcClkDelay())
        return false;

    for(auto it=delay.map.cbegin(); it!=delay.map.cend(); ++it) {
        const auto &delay = it.value();
        quint16 mask = 1 << (it.key()*2);
        if (!changeDesDelay(mask, delay.clock))
            return false;
        mask <<= 1;
        if (!changeDesDelay(mask, delay.data))
            return false;
    }
    return true;
}

bool DeserializersControlModule::setSerDesDelay(const AdcSerDesDelayMap &delay)
{
    bool ok = changeDesDelay(delay) &&
        ResetDeser() &&
        ResetAdcStatus();
    if (!ok)
        return false;
    QThread::msleep(1);
    return true;
}
