#include "FeLinkDeviceUtil.h"

#include "regio/RegOpStream.h"
#include "regio/regop.h"

#include "FeLinkConfig.h"
#include "FeLinkBuildParams.h"
#include "FeLinkStatus.h"
#include "FeLinkRegs.h"

namespace {
typedef union {
    quint16 all;
    struct {
        unsigned int chEn: 1;               // [0]
        unsigned int manualPhaseAdjEn: 1;   // [1]
        unsigned int autoPhaseAdjEn: 1;     // [2]
        unsigned int rcvTrigEn: 1;          // [3]
        unsigned int unused2: 10;           // [13-4]
        unsigned int resetStat: 1;          // [14]
        unsigned int resetSoftPll: 1;       // [15]
    } bit;
} reg_channel_cfg;

typedef union {
    quint16 all;
    struct {
        unsigned int manualPhyReset: 1;     // [0]
        unsigned int autoPhyResetEn: 1;     // [1]
        unsigned int selfReleasePhyReset: 1;// [2]
        unsigned int unused: 13;            // [3-15]
    } bit;
} reg_phy_level_ctrl;

typedef union {
    quint32 all;
    struct {
        unsigned int chNum: 8;                // [0-7]
        unsigned int unused1: 8;                // [8-15]
        unsigned int vxsSlot: 8;                // [16-23]
        unsigned int envCode: 4;                // [24-27]
        unsigned int unused2: 3;                // [28-30]
        unsigned int master: 1;                // [31]
    } bit;
} reg_channel_info;

typedef union {
    quint16 all;
    struct {
        unsigned int gtxReset: 1;       // [0]
        unsigned int remoteGtxReset: 1; // [1]
        unsigned int unused: 14;        // [2-15]
    } bit;
} reg_gtx_ctrl;

FeLinkChannelInfo feLinkChannelInfoFromReg(reg_channel_info regInfo) {
    FeLinkChannelInfo i;
    i.ch = regInfo.bit.chNum;
    i.slotNum = regInfo.bit.vxsSlot;
    i.isMaster = regInfo.bit.master;
//    st.communication_evn = static_cast<fe_link_com_env_code>(chInfo.bit.envCode);
    return i;
}
}

RegOpVector getStatusLockROV(bool block)
{
    RegOpVector r;
    r.RegWrite16(REG_GLOB_LOCK, block ? 0x1 : 0);
    return r;
}

mlink::RegOpVector getFeLinkStatusROV(int ch)
{
    RegOpVector r;

    quint16 base = ch*0x80;

    r.RegWrite16(base+REG_FELINK_CTRL, REG_FELINK_CTRL_BIT_CNT_LOCK);
    r.RegRead16(base+REG_FELINK_STATUS);
    r.RegRead16(base+REG_FELINK_RX_BITSLIDE);
    r.RegRead16(base+REG_FELINK_REMOTE_VER);
    r.RegRead64(base+REG64_FELINK_REMOTE_ID);
    r.RegRead16(base+REG_FELINK_SYNC_RESEND_TIME);
    r.RegRead16(base+REG_FELINK_TIME_FRAME_TIMEOUT);
    r.RegRead32(base+REG32_FELINK_TAI_NSEC);
    r.RegRead64(base+REG64_FELINK_TAI_SEC);
    r.RegRead32(base+REG32_FELINK_LINK_STATE_TIME);
    r.RegRead32(base+REG32_FELINK_CNT_RX_CODE_ERROR);
    r.RegRead32(base+REG32_FELINK_CNT_RX_CRC_ERROR);
    r.RegRead32(base+REG32_FELINK_CNT_TIME_FRAMES_SND);
    r.RegRead32(base+REG32_FELINK_CNT_TRIG_FRAMES_SND);
    r.RegRead32(base+REG32_FELINK_CNT_TIME_FRAMES_RCV);
    r.RegRead32(base+REG32_FELINK_CNT_TRIG_FRAMES_RCV);
    r.RegRead32(base+REG32_FELINK_CNT_UNKNOWN_FRAMES);
    r.RegWrite16(base+REG_FELINK_CTRL, 0x0);

    return r;
}

mlink::RegOpVector getFeLinkStatusROV_v2(quint16 baseAddr)
{
    RegOpVector r;

    r.RegRead32(baseAddr+REG32_CH_INFO);
    r.RegRead16(baseAddr+REG_CH_VER);
    r.RegRead16(baseAddr+REG_CH_STATUS);
    r.RegRead16(baseAddr+REG_CH_RX_BITLIDE);
    r.RegRead16(baseAddr+REG_CH_REMORE_VER);
    r.RegRead64(baseAddr+REG64_CH_REMORE_ID);
    r.RegRead32(baseAddr+REG32_CH_REMORE_INFO);
    r.RegRead64(baseAddr+REG64_CH_TAI_SEC);
    r.RegRead32(baseAddr+REG32_CH_TAI_NSEC);
    r.RegRead32(baseAddr+REG32_CH_LINK_STATE_TIME);
    r.RegRead16(baseAddr+REG_CH_SYNC_RESEND_TIME);
    r.RegRead16(baseAddr+REG_CH_TIMEOUT_TIME_FRAME);
    r.RegRead32(baseAddr+REG32_CH_GTX_STATUS);

    r.RegRead32(baseAddr+REG32_CH_CNT_RX_CODE_ERROR);
    r.RegRead32(baseAddr+REG32_CH_CNT_RX_CRC_ERROR);
    r.RegRead32(baseAddr+REG32_CH_CNT_SYNC_FRAMES_SND);
    r.RegRead32(baseAddr+REG32_CH_CNT_SYNC_FRAMES_RCV);
    r.RegRead32(baseAddr+REG32_CH_CNT_TIME_FRAMES_SND);
    r.RegRead32(baseAddr+REG32_CH_CNT_TIME_FRAMES_RCV);
    r.RegRead32(baseAddr+REG32_CH_CNT_TRIG_FRAMES_SND);
    r.RegRead32(baseAddr+REG32_CH_CNT_TRIG_FRAMES_RCV);
    r.RegRead32(baseAddr+REG32_CH_CNT_XOFF_FRAMES_SND);
    r.RegRead32(baseAddr+REG32_CH_CNT_XOFF_FRAMES_RCV);
    r.RegRead32(baseAddr+REG32_CH_CNT_DATA_FRAMES_SND);
    r.RegRead32(baseAddr+REG32_CH_CNT_DATA_FRAMES_RCV);

    r.RegRead32(baseAddr+REG32_CH_CNT_UNKNOWN_FRAMES);
    r.RegRead32(baseAddr+REG32_CH_CNT_RESET_PHY);

    return r;
}

FeLinkStatus getFeLinkStatus(const mlink::RegOpVector &regRes, int shift)
{
    const auto &chunk = RegOpVector(regRes.begin() + shift, regRes.end());
    RegOpStream in(chunk);
    FeLinkStatus st;
    quint16 r16;
    quint64 r32;
    quint64 r64;

    in >> r16; // skip CTRL write (lock counters)
    in >> r16;
    st.baseStatus = FeLinkBaseStatus(reg_fe_link_base_status{r16});
    st.txInfo.isMaster = (r16>>15) & 0x1;
    in >> st.RX_Bitslide;
    in >> r16; st.remoteVer = FeLinkProtocolVersion(r16);
    in >> r64; st.remoteIndex = DeviceIndex(static_cast<DeviceId>((r64>>48)&0xFF),
                                 r64&0xffffFFFFffff);
    in >> st.REG_FELINK_SYNC_RESEND_TIME;
    in >> st.REG_FELINK_TIME_FRAME_TIMEOUT;

    in >> r32; st.timeCode.valid = r32 & 0x2; st.timeCode.ns = r32>>2;
    in >> st.timeCode.sec; // warn: 64-bit to 32-bit seconds

    in >> st.Current_link_state_time;

    in >> st.counter.RX_code_error;
    in >> st.counter.RX_CRC_error;
    in >> st.counter.time_fr_sent;
    in >> st.counter.trigger_fr_sent;
    in >> st.counter.time_fr_received;
    in >> st.counter.trigger_fr_received;
    in >> st.counter.unknown_received;
    in >> r16; // skip CTRL write (unlock counters)

    if (!in.isValid())
        return {};
    return st;
}

FeLinkStatus getFeLinkStatus_v2(const mlink::RegOpVector &regRes, int shift)
{
    const auto &chunk = RegOpVector(regRes.begin() + shift, regRes.end());
    RegOpStream in(chunk);
    FeLinkStatus st;
    reg_channel_info chInfo;
    quint16 r16;
    quint32 r32;
    quint64 r64;

    in >> chInfo.all;
    st.txInfo = feLinkChannelInfoFromReg(chInfo);
    st.communication_evn = static_cast<fe_link_com_env_code>(chInfo.bit.envCode);

    in >> r16; st.feLinkVer = FeLinkProtocolVersion(r16);
    in >> r16;
    st.baseStatus = FeLinkBaseStatus(reg_fe_link_base_status{r16});
    st.sfpStatus = FeLinkSfpStatus(reg_fe_link_sfp_status{r16});
    st.phase_adjust_enabled = (r16>>15) & 0x1;
    in >> st.RX_Bitslide;
    in >> r16; st.remoteVer = FeLinkProtocolVersion(r16);
    in >> r64; st.remoteIndex = DeviceIndex(static_cast<DeviceId>((r64>>48)&0xFF),
                                 r64&0xffffFFFFffff);

    in >> chInfo.all;
    st.rxInfo = feLinkChannelInfoFromReg(chInfo);

    in >> st.timeCode.sec; // warn: 64-bit to 32-bit seconds
    in >> r32; st.timeCode.valid = r32 & 0x2; st.timeCode.ns = r32>>2;
    in >> st.Current_link_state_time;
    in >> st.gtx_status.all;

    in >> st.REG_FELINK_SYNC_RESEND_TIME;
    in >> st.REG_FELINK_TIME_FRAME_TIMEOUT;

    in >> st.counter.RX_code_error;
    in >> st.counter.RX_CRC_error;

    in >> st.counter.sync_fr_sent;
    in >> st.counter.sync_fr_received;
    in >> st.counter.time_fr_sent;
    in >> st.counter.time_fr_received;
    in >> st.counter.trigger_fr_sent;
    in >> st.counter.trigger_fr_received;
    in >> st.counter.xoff_fr_sent;
    in >> st.counter.xoff_fr_received;
    in >> st.counter.data_fr_sent;
    in >> st.counter.data_fr_received;

    in >> st.counter.unknown_received;
    in >> st.counter.phy_reset;

    if (!in.isValid())
        return {};
    return st;
}

RegOpVector getFeLinkBuildParamsROV()
{
    RegOpVector r;

    r.RegRead32(REG32_GLOB_ID);
    r.RegRead16(REG_GLOB_CH_NUM);
    r.RegRead16(REG_GLOB_START_ADDR);
    r.RegRead16(REG_GLOB_ADDR_SIZE_PER_CH);

    return r;
}

std::optional<FeLinkBuildParams> getFeLinkBuildParams(const mlink::RegOpVector &regRes)
{
    FeLinkBuildParams bp;
    quint16 addrSize;

    RegOpStream in(regRes);
    in >> bp.id;
    in >> bp.chNum;
    in >> bp.startAddr;
    in >> addrSize;
    bp.addStep = 1 << addrSize;

    if(!bp.isValid())
        return {};

    return bp;
}

mlink::RegOpVector getWriteChConfigROV_v2(quint16 baseAddr, const FeLinkChConfig &cfg)
{
    RegOpVector r;

    reg_channel_cfg reg_cfg{0x7};
    reg_cfg.bit.rcvTrigEn = cfg.recvTrig;
    reg_cfg.bit.resetSoftPll = cfg.resetSoftPll;
    r.RegWrite16(baseAddr+REG_CH_CTRL, reg_cfg.all);

    reg_phy_level_ctrl phyCtrl{0x2};
    phyCtrl.bit.manualPhyReset = cfg.reset;
    r.RegWrite16(baseAddr+REG_CH_PHY_LEV_CTRL, phyCtrl.all);

    reg_gtx_ctrl gtxCtrl{0};
    gtxCtrl.bit.remoteGtxReset = cfg.remoteReset;
    r.RegWrite16(baseAddr+REG32_CH_GTX_CTRL, gtxCtrl.all);
    return r;
}

mlink::RegOpVector getPhyLevelCtrlRegsROV(const FeLinkBuildParams &bp)
{
    RegOpVector r;
    for(int ch=0; ch < bp.chNum; ++ch) {
        quint16 baseAddr = bp.startAddr+ch*bp.addStep;
        r.RegRead16(baseAddr+REG_CH_PHY_LEV_CTRL);
    }

    return r;
}

QVector<quint16> getPhyLevelCtrlRegs(const mlink::RegOpVector &regRes)
{
    QVector<quint16> ret;
    quint16 v;
    RegOpStream in(regRes);
    while(!in.eof()) {
        in >> v;
        ret.push_back(v);
    }

    return ret;
}

mlink::RegOpVector selfReleasePhyResetUtil(const QVector<quint16> &regs,
                                           const FeLinkBuildParams &bp,
                                           int resetCh)
{
    RegOpVector r;
    for(int ch=0; ch < bp.chNum; ++ch) {
        if(resetCh!=-1 && resetCh!=ch)
            continue;
        quint16 baseAddr = bp.startAddr+resetCh*bp.addStep;
        reg_phy_level_ctrl phyCtrl{regs[ch]};
        phyCtrl.bit.selfReleasePhyReset = 0;
        r.RegWrite16(baseAddr+REG_CH_PHY_LEV_CTRL, phyCtrl.all);
        phyCtrl.bit.selfReleasePhyReset = 1;
        r.RegWrite16(baseAddr+REG_CH_PHY_LEV_CTRL, phyCtrl.all);
    }

    return r;
}
