#ifndef FELINKSTATUS_H
#define FELINKSTATUS_H

#include <QMetaType>

#include "base/DeviceIndex.h"
#include "base/TaiTime.h"

class QDebug;

typedef union {
    quint16 all;
    struct {
        unsigned int PHY_level_sync_accure: 1;  // [0]
        unsigned int rx_sync: 1;                // [1] // remote device PHY level sync
        unsigned int remote_rx_sync: 1;         // [2] // remote device verified our PHY level sync
        unsigned int softpll_locked: 1;         // [3] // only for time receiving side
        unsigned int link_ok: 1;                // [4]
        unsigned int remote_link_ok: 1;         // [5]
        unsigned int unused1: 10;               // [6-15]
    } bit;
} reg_fe_link_base_status;

typedef union {
    quint16 all;
    struct {
        unsigned int unused1: 8;                // [0-7]
        unsigned int sfp_inserted: 1;           // [8] - SFP inserted (inverted SFP_MOD_DETECT)
        unsigned int sfp_sig_detected: 1;       // [9] - SFP Signal Detect (inverted SFP_LOS)
        unsigned int sfp_tx_fault: 1;           // [10] - SFP_TX_FAULT
        unsigned int sfp_tx_disable: 1;         // [11] - SFP_TX_DISABLE
        unsigned int unused2: 4;                // [15:12] - reserved
    } bit;
} reg_fe_link_sfp_status;

typedef union {
    quint32 all;
    struct {
        unsigned int cpll_reset_state: 3;   // [2:0] - encoded CPLL reset state
        unsigned int cpll_lock: 1;          // [3] - CPLLLOCK
        unsigned int cpll_ref_clk_lost: 1;  // [4] - CPLLREFCLKLOST
        unsigned int cpll_pd: 1;            // [5] - CPLLPD
        unsigned int cpll_reset: 1;         // [6] - CPLLRESET
        unsigned int cpll_ready: 1;         // [7] - CPLL ready
        unsigned int tx_reset_state: 3;     // [10:8] - encoded TX reset state
        unsigned int tx_ready: 1;           // [11] - TX ready
        unsigned int tx_reset_done: 1;      // [12] - TX reset done
        unsigned int gt_tx_reset: 1;        // [13] - GTTXRESET
        unsigned int rx_reset_state: 3;     // [16:14] - encoded RX reset state
        unsigned int rx_ready: 1;           // [17] - RX ready
        unsigned int rx_reset_done: 1;      // [18] - RX reset done
        unsigned int gt_rx_reset: 1;        // [19] - GTRXRESET
        unsigned int rx_cdr_lock: 1;        // [20] - RX CDR lock
        unsigned int rx_comma_detect: 1;    // [21] - RX comma detect
        unsigned int rx_byte_alligned: 1;   // [22] - RX byte alligned
        unsigned int rx_synced: 1;          // [23] - RX synced
        unsigned int gtx_ready: 1;          // [24] - GTX ready
        unsigned int unused1: 7;            // [31:25] - reserved
    } bit;
} reg_fe_link_gtx_status;

inline bool operator==(const reg_fe_link_gtx_status &l, const reg_fe_link_gtx_status &r)
{
    return l.all == r.all;
}

enum fe_link_com_env_code {
    FE_LINK_COM_ENV_QSFP    = 0,
    FE_LINK_COM_ENV_SFP     = 1,
    FE_LINK_COM_ENV_VXS     = 2,
    FE_LINK_COM_ENV_OTHER   = 3,
};

struct FeLinkBaseStatus
{
    FeLinkBaseStatus() = default;
    FeLinkBaseStatus(reg_fe_link_base_status v)
        : PHY_level_sync_accure(v.bit.PHY_level_sync_accure)
        , rx_sync(v.bit.rx_sync)
        , remote_rx_sync(v.bit.remote_rx_sync)
        , softpll_locked(v.bit.softpll_locked)
        , link_ok(v.bit.link_ok)
        , remote_link_ok(v.bit.remote_link_ok)
    {}
    friend bool operator==(const FeLinkBaseStatus &, const FeLinkBaseStatus &) = default;
    bool isEmpty() const { return !PHY_level_sync_accure; }
    bool allOk() const;

    bool PHY_level_sync_accure = false;
    bool rx_sync = false;
    bool remote_rx_sync = false;
    bool softpll_locked = false;
    bool link_ok = false;
    bool remote_link_ok = false;
};
QDebug operator<<(QDebug dbg, const FeLinkBaseStatus &v);

struct FeLinkSfpStatus
{
    FeLinkSfpStatus() = default;
    FeLinkSfpStatus(reg_fe_link_sfp_status v)
        : inserted(v.bit.sfp_inserted)
        , sig_detected(v.bit.sfp_sig_detected)
        , tx_fault(v.bit.sfp_tx_fault)
        , tx_disable(v.bit.sfp_tx_disable)
    {}
    friend bool operator==(const FeLinkSfpStatus &, const FeLinkSfpStatus &) = default;

    bool inserted = false;
    bool sig_detected = false;
    bool tx_fault = false;
    bool tx_disable = false;
};
QDebug operator<<(QDebug dbg, const FeLinkSfpStatus &v);

struct FeLinkProtocolVersion
{
    FeLinkProtocolVersion() = default;
    FeLinkProtocolVersion(quint16 v) :
        maj ((v>>4) & 0xF),
        min ( v     & 0xF)
    {}
    friend bool operator==(const FeLinkProtocolVersion &, const FeLinkProtocolVersion &) = default;

    QString getInfoStr() const;

    quint8 maj=0;
    quint8 min=0;
};
QDebug operator<<(QDebug dbg, const FeLinkProtocolVersion &c);

struct FeLinkCounter
{
    friend bool operator==(const FeLinkCounter &, const FeLinkCounter &) = default;

    quint32 RX_code_error;
    quint32 RX_CRC_error;

    quint32 time_fr_sent;
    quint32 time_fr_received;

    quint32 trigger_fr_sent;
    quint32 trigger_fr_received;

    quint32 sync_fr_sent = 0;
    quint32 sync_fr_received = 0;

    quint32 xoff_fr_sent = 0;
    quint32 xoff_fr_received = 0;

    quint32 data_fr_sent = 0;
    quint32 data_fr_received = 0;

    quint32 unknown_received;
    quint32 phy_reset = 0;
};

struct FeLinkChannelInfo
{
    friend bool operator==(const FeLinkChannelInfo &, const FeLinkChannelInfo &) = default;
    QPair<uint8_t,uint8_t> getSfpConnector() const {
        return {slotNum >> 4, slotNum & 0xF};
    }

    int ch = 0;
    int slotNum = 0; // VXS slot or SFP con. num.
    bool isMaster = false;
};
QDebug operator<<(QDebug dbg, const FeLinkChannelInfo &v);

struct FeLinkStatus
{
    bool isEmpty() const { return baseStatus.isEmpty(); }
    bool allOk() const { return baseStatus.allOk() &&
                (baseStatus.softpll_locked || !phase_adjust_enabled || txInfo.isMaster); }
    DeviceIndex getMasterIndex() const { return txInfo.isMaster? readerIndex : remoteIndex; }
    DeviceIndex getSlaveIndex() const { return !txInfo.isMaster? readerIndex : remoteIndex; }
    friend bool operator==(const FeLinkStatus &, const FeLinkStatus &) = default;
    QString getInfoStr(bool shortType) const;
    QString getSfpStatusStr(bool full) const;
    QString getCommunicationEnvStr(bool remote) const;

    DeviceIndex readerIndex;
    FeLinkChannelInfo txInfo; // mine info
    FeLinkChannelInfo rxInfo; // remote info
    fe_link_com_env_code communication_evn = FE_LINK_COM_ENV_OTHER; // for tx only

    FeLinkBaseStatus baseStatus;
    FeLinkSfpStatus sfpStatus;
    bool phase_adjust_enabled = false;
    quint16 RX_Bitslide = 0;

    FeLinkProtocolVersion feLinkVer;
    FeLinkProtocolVersion remoteVer;
    DeviceIndex remoteIndex;
    quint16 REG_FELINK_SYNC_RESEND_TIME; // R/W
    quint16 REG_FELINK_TIME_FRAME_TIMEOUT; // R/W

    reg_fe_link_gtx_status gtx_status;

    TaiTime timeCode;

    quint32 Current_link_state_time; // miliseconds

    FeLinkCounter counter;
};
Q_DECLARE_METATYPE(FeLinkStatus);
QDebug operator<<(QDebug dbg, const FeLinkStatus &c);

#endif // FELINKSTATUS_H
