#include "RunLogicModule.h"

#include "DeviceModuleCommand.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

#include "DeviceCommandType.h"
#include "RunLogicConfig.h"
#include "RunLogicStatus.h"

#include <cerrno>

namespace {
enum {
    REG32_ID             = 0x00,
    REG_CTRL             = 0x02,
    REG_LOCK             = 0x03,
    REG_MODE             = 0x04,
    REG_STATE            = 0x05,
    REG32_TIME_LIMIT     = 0x06,
    REG32_EVENT_LIMIT    = 0x08,
    REG32_EVENT_COUNT    = 0x0A,
    REG64_START_TIME     = 0x10,
    REG64_STOP_TIME      = 0x14,
    REG64_CURRENT_TIME   = 0x18,
    REG32_RUN_TIME       = 0x1C,
};

typedef union reg_mode_t {
    quint16 all;
    struct {
        quint32 time_limit: 1; // [0]
        quint32 event_limit: 1; // [1]
        quint32 reserved: 14; // [15:2]
    } bit;
} reg_mode_t;

const quint16 CTRL_BIT_RESET = 0x0001;
const quint16 CTRL_BIT_RUN   = 0x8000;
const quint16 LOCK_BIT = 0x0001;

} // namespace

QList<AFI_SDB_ID> RunLogicModule::getSupportedSdbIds()
{
    return {AFI_SDB_RUN_LOGIC};
}

std::optional<RunLogicStatus> RunLogicModule::readStatus()
{
    RunLogicStatus result;
    if (!readStatus(&result))
        return {};
    return result;
}

bool RunLogicModule::readStatus(RunLogicStatus *result)
{
    RunLogicStatus status;
    RegOpVector r;
    r.RegRead32(REG32_ID, &status.id);
    r.RegRead16(REG_STATE, &status.state);
    r.RegWrite16(REG_LOCK, LOCK_BIT);
    r.RegRead32(REG32_RUN_TIME, &status.currentRunTime);
    r.RegRead32(REG32_EVENT_COUNT, &status.currentEventCount);
    r.RegRead64(REG64_CURRENT_TIME, &status.currentTimeRaw);
    r.RegRead64(REG64_START_TIME, &status.startTimeRaw);
    r.RegRead64(REG64_STOP_TIME, &status.finishTimeRaw);
    r.RegWrite16(REG_LOCK, 0);
    if (!regOpExecRebased(r))
        return {};
    if (result)
        *result = status;
    return true;
}

bool RunLogicModule::writeConfig(const RunLogicConfig &config)
{
    reg_mode_t reg_mode = {};
    reg_mode.bit.event_limit = config.eventLimitEnable;
    reg_mode.bit.time_limit = config.timeLimitEnable;
    RegOpVector r;
    r.RegWrite16(REG_MODE, 0); // disable everything
    r.RegWrite32(REG32_TIME_LIMIT, config.timeLimit);
    r.RegWrite32(REG32_EVENT_LIMIT, config.eventLimit);
    r.RegWrite16(REG_MODE, reg_mode.all);
    return regOpExecRebased(r);
}

bool RunLogicModule::writeSoftRun(bool state)
{
    RegOpVector r;
    r.RegWrite16(REG_CTRL, state ? CTRL_BIT_RUN : 0);
    return regOpExecRebased(r);
}

bool RunLogicModule::writeReset()
{
    RegOpVector r;
    r.RegWrite16(REG_CTRL, CTRL_BIT_RESET);
    r.RegWrite16(REG_CTRL, 0);
    return regOpExecRebased(r);
}

DeviceModuleCommandResult RunLogicModule::runCommand(const DeviceModuleCommand &command)
{
    DeviceModuleCommandResult result;
    result.seq = command.seq;
    bool ok = true;
    switch(command.type) {
    case DeviceCommandType::Start: {
        ok &= writeSoftRun(true);
        break;
    }
    case DeviceCommandType::Pause: {
        ok &= writeSoftRun(false);
        break;
    }
    case DeviceCommandType::Restart: {
        ok &= writeReset();
        // ok &= writeSoftRun(true);
        break;
    }
    default:
        result.error = EINVAL;
        return result;
    }

    result.error = ok ? 0 : EIO;
    return result;
}
