#ifndef WRSTATUS_H
#define WRSTATUS_H

#include <QtCore>

#include "base/TaiTime.h"

typedef union {
    quint16 all;
    struct {
        quint32 sfpOk: 1;      // [0]
        quint32 linkUp: 1;     // [1]
        quint32 timeSync: 1;   // [2]
        quint32 unused: 13;    // [15:3]
    } bit;
} reg_wr_status_code;

inline bool operator==(const reg_wr_status_code &l, const reg_wr_status_code &r)
{
    return
           l.all == r.all &&
           l.bit.sfpOk == r.bit.sfpOk &&
           l.bit.linkUp == r.bit.linkUp &&
           l.bit.timeSync == r.bit.timeSync &&
           l.bit.unused == r.bit.unused;
}

struct WrStatus
{
    friend bool operator==(const WrStatus &l, const WrStatus &r) = default;
    TaiTime timeCode() const;
    TaiTime savedTimeCode(int n) const;
    bool timeNull() const { return timeCode().isNull(); }
    bool timeValid() const {return timeCode().isValid(); }

    quint64 mSecsSinceEpoch() const {
        quint64 ms_sec = (quint64)seconds() * 1e3;
        quint64 ms_ns = (quint64)nanoseconds() / 1e6;
        return  ms_sec + ms_ns;
    }
    quint32 seconds() const { return timeCode().sec; }
    quint32 nanoseconds() const { return timeCode().ns; }
    /// Check time is in sync with local time, ntpd should be running.
    int wrTimeCheck() const
    {
        auto wrTimeOffsetMs = mSecsSinceEpoch() - acquiredAtMs;
        if (acquireErrorMs < 5) {
            int wrTimeErrorMs = std::abs((int)(wrTimeOffsetMs + 500) % 1000 - 500); // milliseconds part
            // no check for seconds (offset changes every time leap second is added)
            if (wrTimeErrorMs > 50)
                return 2;
            if (wrTimeErrorMs > 10)
                return 1;
        }
        return 0;
    }
    WrStatus & operator-=(const WrStatus& st) {
        uptime     -= st.uptime;
        linkError  -= st.linkError;
        syncFail   -= st.syncFail;
        codeError  -= st.codeError;
        otherError -= st.otherError;
        return *this;
    }
    qint64 acquiredAtMs = 0;
    qint32 acquireErrorMs = 0;

    reg_wr_status_code statusCode = {0};    // Reg 0x00
    quint32 uptime = 0;                     // Reg 0x02
    quint64 timeCodeRaw;                    // Reg 0x04
    quint32 linkError = 0;                  // Reg 0x08
    quint32 syncFail = 0;                   // Reg 0x0A
    quint32 codeError = 0;                  // Reg 0x0C
    quint32 otherError = 0;                 // Reg 0x0E
    std::array<quint64, 4> savedTimeCodeRaw = {0,0,0,0}; // Reg 0x10+i*4
};

#endif // WRSTATUS_H
