#include "DeviceChModel.h"

#include <QDebug>
#include <QFont>
#include <QTimer>

#include "base/StatisticsDb.h"
#include "base/StatisticsDbUtil.h"
#include "TqdcAppConfigUtil.h"

static const double MINIMUM_RATE_UPDATE_INTERVAL = 0.99;

namespace {
QList<QMap<int,bool> TqdcCoreConfig::*> getConfigBoolMapList(TreeColumns col) {
    auto chTdcEn = &TqdcCoreConfig::chTdcEn;
    auto chAdcEn = &TqdcCoreConfig::chAdcEn;
    auto chInv = &TqdcCoreConfig::chInv;
    auto chTrigEn = &TqdcCoreConfig::chTrigEn;
    auto chDigTrigEn = &TqdcCoreConfig::chDigTrigEn;
    auto chDigTrigRising = &TqdcCoreConfig::chDigTrigRising;
    auto chZsDigEn = &TqdcCoreConfig::chDigZsEn;
    auto chZsDigRising = &TqdcCoreConfig::chDigZsRising;
    switch(col) {
    case TreeColumns::COL_NAME:     return {chTdcEn, chAdcEn};
    case TreeColumns::COL_TDC_EN:   return {chTdcEn};
    case TreeColumns::COL_ADC_EN:   return {chAdcEn};
    case TreeColumns::COL_INVERT:   return {chInv};
    case TreeColumns::COL_TRIG_EN:  return {chTrigEn};
    case TreeColumns::COL_DIG_TRIG_EN:  return {chDigTrigEn};
    case TreeColumns::COL_DIG_TRIG_EDGE:  return {chDigTrigRising};
    case TreeColumns::COL_ZS_DIG_EN:  return {chZsDigEn};
    case TreeColumns::COL_ZS_DIG_EDGE:  return {chZsDigRising};
    case TreeColumns::COL_TOTAL:    return {chTdcEn,  chAdcEn, chInv, chTrigEn,
        chDigTrigEn, chDigTrigRising, chZsDigEn, chZsDigRising};
    default:
        Q_UNREACHABLE();
    }
}

bool getDefConfigBoolVal(TreeColumns col) {
switch(col) {
case TreeColumns::COL_TDC_EN:   return TqdcCoreConfig::DEF_TDC_EN;
case TreeColumns::COL_ADC_EN:   return TqdcCoreConfig::DEF_ADC_EN;
case TreeColumns::COL_INVERT:   return TqdcCoreConfig::DEF_INV;
case TreeColumns::COL_TRIG_EN:  return TqdcCoreConfig::DEF_TRIG_EN;
case TreeColumns::COL_DIG_TRIG_EN:  return TqdcCoreConfig::DEF_DIG_TRIG_EN;
case TreeColumns::COL_DIG_TRIG_EDGE:  return TqdcCoreConfig::DEF_DIG_TRIG_RISING;
case TreeColumns::COL_ZS_DIG_EN:  return TqdcCoreConfig::DEF_ZS_EN;
case TreeColumns::COL_ZS_DIG_EDGE:  return TqdcCoreConfig::DEF_ZS_RISING;
default:
    Q_UNREACHABLE();
}
}

inline Qt::CheckState boolToCheck(bool v)
{ return v ? Qt::Checked : Qt::Unchecked; }

QString getRateStr(double rate)
{
    return QString("%L1").arg(rate, 12, 'f', 1, QChar(' '));
}
}

DeviceChModel::DeviceChModel(QObject *parent)
    :
      QAbstractItemModel(parent)
    , delayedSendConfigTimer(new QTimer(this))
{
    delayedSendConfigTimer->setInterval(100);
    delayedSendConfigTimer->setSingleShot(true);
    connect(delayedSendConfigTimer, &QTimer::timeout,
            this, &DeviceChModel::postCheckStateChange);
}

QVariant DeviceChModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::CheckStateRole:
        return checkStateData(index);
    case Qt::FontRole:
        return fontData(index);
    }
    return QVariant();
}

void DeviceChModel::enqueueConfigSend()
{
    delayedSendConfigTimer->start();
}

bool DeviceChModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!index.isValid())
        return false;

    switch (role) {
    case Qt::CheckStateRole:
    {
        const auto item = getItemByIndex(index);
        if(item->ch == -1) {
            auto rows = rowCount(index);
            bool ok = true;
            for(int r=0; r<rows; ++r){
                ok &= setData(this->index(r,index.column(),index), value, role);
            }
            return ok;
        } else {
            Q_ASSERT(!item->dev.isNull());
            auto col = static_cast<TreeColumns>(index.column());
            QList<QMap<int,bool> TqdcCoreConfig::*> ptrs = getConfigBoolMapList(col);
            if(ptrs.isEmpty())
                return false;

            Q_ASSERT(config.knownConfigs.contains(item->dev));
            if(!config.knownConfigs.contains(item->dev))
                return false;
            auto &ds = config.knownConfigs[item->dev].tqdcCore;
            Q_ASSERT(value!=Qt::PartiallyChecked);
            for(auto ptr : ptrs) {
                auto &map = (ds.*ptr);
                map[item->ch] = (value==Qt::Checked) ? true : false;
                enqueueConfigSend();
            }
        }
        return true;
    }
    case Qt::EditRole:
    {
        const auto item = getItemByIndex(index);
        auto col = static_cast<TreeColumns>(index.column());
        QMap<int,int> TqdcCoreConfig::* ptr = nullptr;
        switch(col) {
        case TreeColumns::COL_ANLG_THR:
            ptr = &TqdcCoreConfig::chThr; break;
        case TreeColumns::COL_DIG_TRIG_LVL:
            ptr = &TqdcCoreConfig::chDigThr; break;
        case TreeColumns::COL_ZS_DIG_THR:
            ptr = &TqdcCoreConfig::chDigZsThr; break;
        default:
            return false;
        }
        if(item->ch == -1) {
            if(index.parent() == QModelIndex()) {
                TqdcAppConfigUtil::setTqdcThreshold(&config, ptr, value.toInt());
            } else {
                config.knownConfigs[item->dev].tqdcCore.setField(ptr, value.toInt());
            }
        } else {
            auto &tqdcCfg = config.knownConfigs[item->dev].tqdcCore;
            (tqdcCfg.*ptr)[item->ch] = value.toInt();
        }
        enqueueConfigSend();
        static const QVector<int> roles{Qt::DisplayRole, Qt::EditRole};
        emit dataChanged(index, index, roles);
        recursiveDataChanged(index, roles);
        return true;
    }
    }

    return false;
}

void DeviceChModel::setEnabled(bool en)
{
    if(enable == en)
        return;

    enable = en;
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh);
}

Qt::ItemFlags DeviceChModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);
    auto col = static_cast<TreeColumns>(index.column());
    switch (col) {
    case TreeColumns::COL_NAME: // with COL_ENABLE,
    case TreeColumns::COL_TDC_EN:
    case TreeColumns::COL_ADC_EN:
    case TreeColumns::COL_INVERT:
    case TreeColumns::COL_TRIG_EN:
    case TreeColumns::COL_DIG_TRIG_EN:
    case TreeColumns::COL_DIG_TRIG_EDGE:
    case TreeColumns::COL_ZS_DIG_EN:
    case TreeColumns::COL_ZS_DIG_EDGE:
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
        break;
    case TreeColumns::COL_ANLG_THR:
    case TreeColumns::COL_DIG_TRIG_LVL:
    case TreeColumns::COL_ZS_DIG_THR:
        f |= Qt::ItemIsEditable;
        break;
    case TreeColumns::COL_RATE:
    case TreeColumns::COL_LEVEL:
    default:
        break;
    }
    if(!enable)
        f = f & (~Qt::ItemIsEnabled);
    return f;
}

QVariant DeviceChModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(orientation != Qt::Horizontal)
        return QVariant();
    if(role != Qt::DisplayRole)
        return QVariant();
    switch (static_cast<TreeColumns>(section)) {
    case TreeColumns::COL_NAME:
        return "Name";
    case TreeColumns::COL_TDC_EN:
        return "TdcEn";
    case TreeColumns::COL_ADC_EN:
        return "AdcEn";
    case TreeColumns::COL_INVERT:
        return "Inv";
    case TreeColumns::COL_TRIG_EN:
        return "TrigEn";
    case TreeColumns::COL_ANLG_THR:
        return "Anlg Thr";
    case TreeColumns::COL_DIG_TRIG_EN:
        return "DigTrigEn";
    case TreeColumns::COL_DIG_TRIG_EDGE:
        return "DigTrigEdge";
    case TreeColumns::COL_DIG_TRIG_LVL:
        return "DigTrigThr";
    case TreeColumns::COL_ZS_DIG_EN:
        return "ZsDigEn";
    case TreeColumns::COL_ZS_DIG_EDGE:
        return "ZsDigEdge";
    case TreeColumns::COL_ZS_DIG_THR:
        return "ZsDigThr";
    case TreeColumns::COL_LEVEL:
        return "Level";
    case TreeColumns::COL_RATE:
        return "Rate, Hz";
    default:
        return QVariant();
    }
}

QModelIndex DeviceChModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const DeviceChModelItem *parentItem = parent.isValid() ?
        static_cast<DeviceChModelItem*>(parent.internalPointer()) :
        &rootItem;
    DeviceChModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex DeviceChModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const DeviceChModelItem *childItem = static_cast<DeviceChModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    DeviceChModelItem *parentItem = childItem->parent;
    if(parentItem == &rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int DeviceChModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
        static_cast<DeviceChModelItem*>(parent.internalPointer()) :
        &rootItem;
    return parentItem->childCount();
}

static QMap<DeviceIndex, int> buildNchMap(const BaseDeviceAppConfig &config)
{
    QMap<DeviceIndex, int> nch;
    const auto &keys = config.knownConfigs.keys();
    for (auto index : keys) {
        nch.insert(index, TqdcCoreConfig::getNumberOfChannels(index.getDevId()));
    }
    return nch;
}

static bool needReset(const BaseDeviceAppConfig &config, const BaseDeviceAppConfig &newConfig)
{
    if (config.knownConfigs.keys() != newConfig.knownConfigs.keys())
        return true;
    const auto &oldNch = buildNchMap(config);
    const auto &newNch = buildNchMap(newConfig);
    if (oldNch != newNch)
        return true;
    for (const auto &ds : config.knownConfigs) {
        if (ds.enabled != newConfig.knownConfigs[ds.index].enabled)
            return true;
    }
    return false;
}

static bool hasChanges(const ModularDeviceConfig &c1, const ModularDeviceConfig &c2)
{
    if (c1.tqdcCore.chThr != c2.tqdcCore.chThr || c1.tqdcCore.chDigThr != c2.tqdcCore.chDigThr)
        return true;

    auto ptrs = getConfigBoolMapList(TreeColumns::COL_TOTAL);
    for(auto ptr : ptrs) {
        if(c1.tqdcCore.*ptr != c1.tqdcCore.*ptr)
            return true;
    }
    return false;
}

static bool hasChanges(const BaseDeviceAppConfig &c1, const BaseDeviceAppConfig &c2)
{
    if (hasChanges(c1.defaultSetup, c2.defaultSetup))
        return true;

    for (auto it1=c1.knownConfigs.cbegin(); it1!=c1.knownConfigs.cend(); ++it1) {
        auto it2 = c2.knownConfigs.constFind(it1.key());
        if (hasChanges(*it1, *it2))
            return true;
    }
    return false;
}

void DeviceChModel::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if (config == newConfig)
        return;
    const auto &newNch = buildNchMap(newConfig);
    bool performReset = needReset(config, newConfig);
    bool hasAnyChanges = performReset ? true : hasChanges(config, newConfig);

    config = newConfig;
    if (performReset) {
        beginResetModel();
        rootItem.children.clear();
        const auto &setups = config.knownConfigs;
        if (config.hasEnabledDevices()) {
            DeviceChModelItem *allItem = new DeviceChModelItem(&rootItem);
            rootItem.addChild(allItem);

            for (auto devIt = setups.cbegin(); devIt != setups.cend(); ++devIt) {
                if (!devIt->enabled)
                    continue;
                auto devItem = new DeviceChModelItem(allItem);
                devItem->dev = devIt.key();
                for (int ch = 0; ch < newNch[devItem->dev]; ch++) {
                    auto chItem = new DeviceChModelItem(devItem);
                    chItem->ch = ch;
                    devItem->addChild(chItem);
                }
                allItem->addChild(devItem);
            }
        }
        endResetModel();
    } else {
        if(hasAnyChanges)
            sendDataChanged();
    }
}

void DeviceChModel::deviceStatusUpdated(const ModularDeviceStatus &status)
{
    const auto &hitCountersOpt = status.hitCoreStatus;
    if (!hitCountersOpt)
        return;
    auto &prevStatus = dev_status[status.getDeviceIndex()];
    const auto &hitCounters = hitCountersOpt.value();
    auto &prevHitCountersOpt = prevStatus.hitCoreStatus;
    if (prevHitCountersOpt) {
        auto &prevHitCounters = prevHitCountersOpt.value();

        // difference is calculated on statistics, enforce minimum time interval between updates
        const auto timeDiff = hitCounters.taiTime() - prevHitCounters.taiTime();
        if (timeDiff.toNs() * 1e-9 < MINIMUM_RATE_UPDATE_INTERVAL)
            return;
    }
    prevStatus = status;

    auto &statDb = StatisticsDb::instance();
    // update statistics
    statDb.add(StatisticsDbUtil::key("input_hit_time", status.getDeviceIndex()), hitCounters.taiTime().toNs() * 1e-9);
    for (int i=0; i<hitCounters.cnt.size(); i++) {
        statDb.add(StatisticsDbUtil::key("input_hit_count", status.getDeviceIndex(), i), hitCounters.cnt[i]);
    }

    const auto &devIndex = index(getDevRow(status.getDeviceIndex()), 0, index(0,0));
    auto rows = rowCount(devIndex);
    if(rows) {
        const auto &top = index(0,static_cast<int>(TreeColumns::COL_RATE), devIndex);
        const auto &bottom = top.sibling(rows, top.column());
        emit dataChanged(top, bottom, {Qt::DisplayRole});
    }
}

void DeviceChModel::postCheckStateChange()
{
    sendDataChanged(QVector<int>{Qt::CheckStateRole});
    emit configChanged(this->config);
}

void DeviceChModel::sendDataChanged(QVector<int> roles)
{
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh, roles);
}

QVariant DeviceChModel::displayData(const QModelIndex &index) const
{
    const auto &statDb = StatisticsDb::instance();
    const auto item = getItemByIndex(index);
    switch (getColumnByIndex(index)) {
    case TreeColumns::COL_NAME:
        if(item->dev.isNull())
            return "All";
        if(item->ch == -1)
            return item->dev.getSerialStr();
        return QString::number(item->ch);
    case TreeColumns::COL_ANLG_THR:
    {
        const auto &val = editData(index);
        if (!val.isNull() && val.isValid())
            return QString("%1 mV").arg(val.toInt());
        break;
    }
    case TreeColumns::COL_DIG_TRIG_LVL:
    case TreeColumns::COL_ZS_DIG_THR:
    {
        const auto &val = editData(index);
        if (!val.isNull() && val.isValid())
            return QString("%1").arg(val.toInt());
        break;
    }
    case TreeColumns::COL_DIG_TRIG_EDGE:
    case TreeColumns::COL_ZS_DIG_EDGE:
    {
        auto chSt = checkStateData(index).value<Qt::CheckState>();
        switch(chSt){
        case Qt::Checked: return QChar(0x279a); // North East Arrow
        case Qt::Unchecked: return QChar(0x2798); // South East Arrow
        case Qt::PartiallyChecked: return QChar(0x2928); // North East and South East
        default: return {};
        }
        break;
    }

    case TreeColumns::COL_LEVEL:
        if(!item->dev.isNull() && item->ch != -1) {
            const auto &status = dev_status[item->dev];
            if (status.hitCoreStatus) {
                auto compState = status.hitCoreStatus->compState;
                quint64 ch = item->ch;
                bool comp = compState & (1ull << ch);
                return QString(comp ? 0x220E : 0x2022);
            }
        }
        return {};
    case TreeColumns::COL_RATE:
        if(!item->dev.isNull() && item->ch != -1) {
            const auto &key_time = StatisticsDbUtil::key("input_hit_time", item->dev);
            const auto &key_hcnt = StatisticsDbUtil::key("input_hit_count", item->dev, item->ch);
            if (statDb.count(key_hcnt) < 2)
                return {};
            auto hcnt = statDb.diff(key_hcnt);
            auto time = statDb.diff(key_time);
            auto v = (double)hcnt / (time ? time : 1);
            return getRateStr(v);
        }
        return {};
    default:
        return {};
    }
   return {};
}

QVariant DeviceChModel::editData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto &col = getColumnByIndex(index);
    QMap<int,int> TqdcCoreConfig::* ptr = nullptr;

    if (col == TreeColumns::COL_ANLG_THR)
        ptr = &TqdcCoreConfig::chThr;
    if(col == TreeColumns::COL_DIG_TRIG_LVL)
        ptr = &TqdcCoreConfig::chDigThr;
    if(col == TreeColumns::COL_ZS_DIG_THR)
        ptr = &TqdcCoreConfig::chDigZsThr;
    if(!ptr)
        return {};

    const auto &deviceIndex = item->dev;
    if (deviceIndex.isNull()) {
        return TqdcAppConfigUtil::getTqdcSimilarCfgField(config, ptr);
    }

    const auto deviceId = deviceIndex.getDevId();
    if (!TqdcCoreConfig::hasAdc(deviceId))
        return {};

    const auto &tqdcConfig = config.knownConfigs.value(deviceIndex).tqdcCore;
    const auto &ch = item->ch;
    if (ch == -1)
        return tqdcConfig.getSimilarField(ptr);

    auto map = tqdcConfig.*ptr;
    return map.value(ch, TqdcCoreConfig::getDefaultFieldVal(ptr));
}

static QVariant combineCheckState(const QVariant &st1, const QVariant &st2)
{
    if (st1 == st2)
        return st1;
    if (st2.isNull())
        return st1;
    if (st1.isNull())
        return st2;
    return Qt::PartiallyChecked;
}

bool DeviceChModel::enableColumn(const DeviceIndex &deviceIndex, TreeColumns col) const
{
    if (deviceIndex.isNull())
        return true;
    const auto deviceId = deviceIndex.getDevId();
    switch (col) {
    case TreeColumns::COL_TDC_EN:
        return TqdcCoreConfig::hasTdc(deviceId);
    case TreeColumns::COL_DIG_TRIG_EN:
    case TreeColumns::COL_DIG_TRIG_EDGE:
    case TreeColumns::COL_DIG_TRIG_LVL:
        return hasDigTrig(deviceIndex);
    case TreeColumns::COL_ZS_DIG_EN:
    case TreeColumns::COL_ZS_DIG_EDGE:
    case TreeColumns::COL_ZS_DIG_THR:
        return hasZsDig(deviceIndex);
    case TreeColumns::COL_ADC_EN:
        return TqdcCoreConfig::hasAdc(deviceId);
    case TreeColumns::COL_INVERT:
        return TqdcCoreConfig::hasInvert(deviceId);
    case TreeColumns::COL_TRIG_EN:
        return TqdcCoreConfig::hasTrig(deviceId);
    default: return true;
    }
}
QVariant DeviceChModel::checkStateData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    if (!item)
        return {};
    const auto deviceIndex = item->dev;
    // config.knownConfigs.value(deviceIndex).tqdcCore.chTdcEn.value(item->ch);

    auto col = getColumnByIndex(index);
    switch (col) {
    case TreeColumns::COL_NAME:
    {
        auto st1 = data(index.siblingAtColumn(static_cast<int>(TreeColumns::COL_TDC_EN)),
                        Qt::CheckStateRole);
        auto st2 = data(index.siblingAtColumn(static_cast<int>(TreeColumns::COL_ADC_EN)),
                        Qt::CheckStateRole);
        return combineCheckState(st1, st2);
    }
    case TreeColumns::COL_TDC_EN:
    case TreeColumns::COL_ADC_EN:
    case TreeColumns::COL_INVERT:
    case TreeColumns::COL_TRIG_EN:
    case TreeColumns::COL_DIG_TRIG_EN:
    case TreeColumns::COL_DIG_TRIG_EDGE:
    case TreeColumns::COL_ZS_DIG_EN:
    case TreeColumns::COL_ZS_DIG_EDGE:
    {
        if (!enableColumn(deviceIndex, col))
            return {};
        const auto mapPtrList = getConfigBoolMapList(getColumnByIndex(index));
        Q_ASSERT(mapPtrList.size()==1);
        QMap<int, bool> TqdcCoreConfig::* ptr = mapPtrList.first();
        return getRecursiveState(index, ptr, getDefConfigBoolVal(col));
    }
    default:
        return QVariant();
    }
}

QVariant DeviceChModel::fontData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    QFont font;
//    if(item->dev.isNull()) {
//        font.setItalic(true);
//        font.setBold(true);
//    } else
        if(item->ch == -1) {
        font.setItalic(true);
    } else {
//            fg.setItalic(true);
    }
    switch (getColumnByIndex(index)) {
    case TreeColumns::COL_RATE: font.setFamily("Monospace");
    default:;
    }
    return font;
}

TreeColumns DeviceChModel::getColumnByIndex(const QModelIndex &index) const
{
    return static_cast<TreeColumns>(index.column());
}

DeviceChModelItem *DeviceChModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<DeviceChModelItem*>(index.internalPointer());
}

QVariant DeviceChModel::getRecursiveState(const QModelIndex &mIndex, QMap<int, bool> TqdcCoreConfig::*ptr, bool defSt) const
{
    auto rows = rowCount(mIndex);
    if(rows) {
        QVariant allSt;
        for(int r=0; r<rows; ++r) {
            auto st = getRecursiveState(index(r,0,mIndex), ptr, defSt);
            if(allSt.isNull())
                allSt = st;
            else if(allSt != st)
                return Qt::PartiallyChecked;
        }
        return allSt.isNull() ? boolToCheck(defSt) : allSt;
    } else {
        auto item = getItemByIndex(mIndex);
        const auto &map = config.knownConfigs.value(item->dev).tqdcCore.*ptr;
//        if (!map.contains(item->ch))
//            return {};
        return boolToCheck(map.value(item->ch, defSt));
    }
}

bool DeviceChModel::hasDigTrig(const DeviceIndex i) const
{
    return true; //todo: build params
}

bool DeviceChModel::hasZsDig(const DeviceIndex i) const
{
    return true; //todo: build params
}

void DeviceChModel::recursiveDataChanged(const QModelIndex &i, const QVector<int> &roles)
{
    auto rows = rowCount(i);
    if(rows==0)
        return;

    QModelIndex first, last;
    for(auto r=0; r<rows; ++r) {
        const auto &rowIndex = index(r, i.column(), i);
        if(!first.isValid())
            first = rowIndex;
        last = rowIndex;
        recursiveDataChanged(rowIndex, roles);
    }

    emit dataChanged(first, last, roles);
}

int DeviceChModel::getDevRow(const DeviceIndex &di) const
{
    DeviceChModelItem * allItem = nullptr;
    if(!rootItem.children.isEmpty())
        allItem = rootItem.children.first();
    if(allItem) {
        for(int r=0; r<allItem->children.size(); ++r)
            if(allItem->children[r]->dev == di)
                return r;
    }
    return -1;
}

