#include "FileReaderDialog.h"
#include "ui_FileReaderDialog.h"

#include <QButtonGroup>
#include <QDebug>
#include "QFileDialog"
#include "QFileInfo"
#include <QSemaphore>
#include <QSettings>
#include <QStyle>
#include <QThread>
#include <QTimer>

#include "FileReader.h"
#include "base/threadQuit.h"

FileReaderDialog::FileReaderDialog(QSemaphore *freeRawDataSem, QWidget *parent)
    : QDialog(parent)
    , ui(new Ui::FileReaderDialog)
    , freeRawDataSem(freeRawDataSem)
    , reader(new FileReader(freeRawDataSem))
    , fileReaderThread(new QThread(this))
    , playbackTimer(new QTimer(this))
    , playbackForward(true)
    , block_write_settings(false)
{
    ui->setupUi(this);

    int cnt = 1;
    ui->comboBoxReadBy->blockSignals(true);
    do {
        ui->comboBoxReadBy->addItem(QString("%L1 event(s)").arg(cnt), cnt);
        cnt*=10;
    } while(cnt <= 1e5);
    ui->comboBoxReadBy->blockSignals(false);

    connect(playbackTimer, &QTimer::timeout, this, &FileReaderDialog::playback);
    playbackTimer->setInterval(ui->spinBoxInterval->value());
    playbackTimer->setSingleShot(false);

    QPixmap p;
    p = QApplication::style()->standardPixmap(QStyle::SP_MediaSeekBackward);
    ui->pushButtonBackward->setIcon(QIcon(p));
    p = QApplication::style()->standardPixmap(QStyle::SP_MediaStop);
    ui->pushButtonStop->setIcon(QIcon(p));
    p = QApplication::style()->standardPixmap(QStyle::SP_MediaSeekForward);
    ui->pushButtonForward->setIcon(QIcon(p));

    QButtonGroup *gr = new QButtonGroup(this);
    gr->addButton(ui->pushButtonBackward);
    gr->addButton(ui->pushButtonStop);
    gr->addButton(ui->pushButtonForward);

    QSettings settings;
    settings.beginGroup("FileReaderDialog");
    restoreGeometry(settings.value("geometry").toByteArray());
    if(settings.value("visible").toBool())
        QTimer::singleShot(0, this, &FileReaderDialog::show);
    settings.endGroup();

    connect(this, &FileReaderDialog::fileNameUpdated, reader, &FileReader::setFileName);
    connect(this, &FileReaderDialog::changeRecord, reader, &FileReader::readRecord);
    connect(this, &FileReaderDialog::packSizeLimitUpdated, reader, &FileReader::setPackSizeLimit);
    connect(reader, &FileReader::dataAccuired, this, &FileReaderDialog::dataAccuired);
    connect(reader, &FileReader::fileMapUpdated, this, &FileReaderDialog::fileMapUpdated);
    connect(reader, &FileReader::eventRead, this, &FileReaderDialog::eventRead);
    connect(reader, &QObject::destroyed, fileReaderThread, &QThread::quit);

    connect(this, &FileReaderDialog::closeAll, reader, &FileReader::closeAll);
    connect(this, &FileReaderDialog::closeFile, reader, &FileReader::closeFile);
    reader->moveToThread(fileReaderThread);
    fileReaderThread->start();
    fileReaderThread->setObjectName("File Reader");
}

FileReaderDialog::~FileReaderDialog()
{
    QSettings settings;
    settings.beginGroup("FileReaderDialog");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("visible", isVisible());
    settings.endGroup();

    emit closeAll();
    if(!threadQuiting(fileReaderThread, 2000)) {
        qWarning() << "Can't stop file reader thread";
    }

    delete ui;
}

void FileReaderDialog::configUpdated()
{
    block_write_settings = true;
    ui->spinBoxInterval->setValue(config.playbackInterval);
    int index = ui->comboBoxReadBy->findData(config.readBy);
    if(index != -1) {
        ui->comboBoxReadBy->setCurrentIndex(index);
    }
    block_write_settings = false;
}

void FileReaderDialog::setPackSizeLimit(qint64 limit)
{
    emit packSizeLimitUpdated(limit);
}

void FileReaderDialog::reopenLastFile()
{
    if(config.lastFile.isEmpty())
        return;

    emit fileNameUpdated(config.lastFile);
    ui->labelFileName->setText(config.lastFile);
    ui->spinBoxCurEvent->setEnabled(false);
    ui->spinBoxCurEvent->setMaximum(0);
}

void FileReaderDialog::setFileReaderConfig(const FileReaderConfig &cfg)
{
    if(config == cfg)
        return;
    config = cfg;
    configUpdated();
}

void FileReaderDialog::on_pushButtonSetFile_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this, "Set source file", config.lastFile);
    if(!fileName.isEmpty()) {
        config.lastFile = fileName;
        saveSettings();
        reopenLastFile();
    }
}

void FileReaderDialog::on_spinBoxCurEvent_valueChanged(int recordNum)
{
    if(recordNum) {
        emit changeRecord(recordNum);
        updateProgressBar(recordNum);
    }
}

void FileReaderDialog::on_pushButtonBackward_toggled(bool checked)
{
    if(checked) {
        playbackForward = false;
        playbackTimer->start();
    }
}

void FileReaderDialog::on_pushButtonStop_toggled(bool checked)
{
    if(checked) {
        playbackTimer->stop();
    }
}

void FileReaderDialog::on_pushButtonForward_toggled(bool checked)
{
    if(checked) {
        playbackForward = true;
        playbackTimer->start();
    }
}

void FileReaderDialog::on_spinBoxInterval_valueChanged(int intervalMs)
{
    playbackTimer->setInterval(intervalMs);
    config.playbackInterval = intervalMs;
    saveSettings();
}

void FileReaderDialog::on_comboBoxReadBy_currentIndexChanged(int)
{
    config.readBy = ui->comboBoxReadBy->currentData().toInt();
    saveSettings();
}

void FileReaderDialog::playback()
{
    if(freeRawDataSem->available()==0) return;

    int curRecord = ui->spinBoxCurEvent->value();
    int firstRecord, lastRecord, nextCurRecord;
    const int step = ui->comboBoxReadBy->currentData().toInt();
    if(playbackForward) {
        firstRecord = curRecord+1;
        lastRecord = curRecord+step;
        nextCurRecord = lastRecord;
    } else {
        lastRecord = curRecord-1;
        firstRecord = curRecord-step;
        nextCurRecord = firstRecord;
    }

    if((firstRecord < ui->spinBoxCurEvent->minimum()
        || firstRecord > ui->spinBoxCurEvent->maximum()) &&
            (lastRecord < ui->spinBoxCurEvent->minimum()
                        || lastRecord > ui->spinBoxCurEvent->maximum())){
        ui->pushButtonStop->setChecked(true);
    } else {
        ui->spinBoxCurEvent->blockSignals(true);
        ui->spinBoxCurEvent->setValue(nextCurRecord);
        ui->spinBoxCurEvent->blockSignals(false);
        firstRecord = qMax(firstRecord, ui->spinBoxCurEvent->minimum());
        lastRecord = qMin(lastRecord, ui->spinBoxCurEvent->maximum());
        emit changeRecord(firstRecord, lastRecord-firstRecord+1);
        updateProgressBar(nextCurRecord);
    }
}

void FileReaderDialog::fileMapUpdated(qint64 fileSize, QString lastModifyStr, int recordsSize, double decodedPart)
{
    ui->labelFileSize->setText(QString("%L1 bytes").arg(fileSize));
    ui->labelModDate->setText(lastModifyStr);

    ui->spinBoxCurEvent->setMaximum(recordsSize);
    ui->labelRecords->setText(QString::number(recordsSize));
    ui->progressBarDecoded->setValue(static_cast<int>(decodedPart*100));
    updateProgressBar(ui->spinBoxCurEvent->value());

    if(!ui->spinBoxCurEvent->isEnabled() && recordsSize) {
        ui->spinBoxCurEvent->setEnabled(true);
        ui->spinBoxCurEvent->blockSignals(true);
        ui->spinBoxCurEvent->setMinimum(1);
        ui->spinBoxCurEvent->blockSignals(false);
        if(ui->spinBoxCurEvent->value()==1)
            emit changeRecord(1);
        else
            ui->spinBoxCurEvent->setValue(1);
    }
}

void FileReaderDialog::eventRead(quint32 ev_first, quint32 ev_last, qint64 evSize)
{
    Q_UNUSED(evSize)
    QString txt = QString::number(ev_first);
    if(ev_first!=ev_last)
        txt.append(QString("-%2").arg(ev_last));
    txt.append(QString(" (%L1 B)").arg(evSize));
    ui->labelEvNum->setText(txt);
}

void FileReaderDialog::onCloseFile()
{
    ui->labelFileName->setText("empty");
    if(!ui->pushButtonStop->isChecked()) {
        qInfo() << "ui->pushButtonStop->setChecked(true);";
        ui->pushButtonStop->setChecked(true);
    }
    emit closeFile();
}

void FileReaderDialog::saveSettings()
{
    if(block_write_settings) return;
    emit fileReaderConfigChanged(config);
}

void FileReaderDialog::updateProgressBar(int recordNum) const
{
    int recMax = ui->spinBoxCurEvent->maximum();
    if(recMax) {
        const int val = static_cast<int>(100.*recordNum/recMax);
        if(ui->progressBarRecord->value() != val) {
            ui->progressBarRecord->setValue(val);
            ui->progressBarRecord->update();
        }
    }
}
