#ifndef HISTMNGRCFG_H
#define HISTMNGRCFG_H

#include <QHash>
#include <QMap>
#include <QMetaType>
#include <QVector>

#include "base/DaqModule.h"
#include "base/DeviceIndex.h"

const int HIST_MATCHWIN_MAX = 10000;
const int HIST_MATCHWIN_MIN = 0;

enum class HistId {
    unknown = 0,
    tdcTime = 1,
    tdcDoublePulseTime = 2,
    tdcCh = 3,
    tdcHitEv = 4,
    errors = 5,
    tdcTaiTime  = 6,
    tdcWidth = 7,
    adcTime = 8,
    tdcAdcDt = 9,
    adcCh = 10,
    adcHitEv = 11,
    chSamples = 12,
    chAmpl = 13,
    inl = 14,
    inlCheck = 15,
    mscPlot  = 16,
};
Q_DECLARE_METATYPE(HistId)

uint qHash(const HistId &key, uint seed = 0);
QString getStrKeyForHistId(HistId id);
HistId getHistIdForStrKey(QString key);

struct HistKey {
    HistKey():
        HistKey(HistId::unknown, DeviceIndex(), -1)
    {}
    HistKey(HistId id, DeviceIndex index = DeviceIndex() , int ch = -1):
        hist_id(id),
        index(index),
        ch(ch)
    {}
    void print_hist_key() const;
    bool operator <(const HistKey &key) const{
        if (hist_id < key.hist_id)
            return true;
        if (hist_id > key.hist_id)
            return false;
        if(index < key.index)
            return true;
        if(key.index < index)
            return false;
        if(ch < key.ch)
            return true;
        if(ch < key.ch)
            return false;
        return false;
    }
    bool operator ==(const HistKey &key) const{
        return ((hist_id == key.hist_id) && (index == key.index) && (ch == key.ch));
    }
    friend uint qHash(const HistKey &key, uint seed = 0) {
        return qHash(key.hist_id, seed) ^ qHash(key.index, seed)
                ^ qHash(key.ch, seed);
    }
    bool isEmpty() const { return hist_id == HistId::unknown; }

    HistId hist_id;
    DeviceIndex index;
    int ch;
};
Q_DECLARE_METATYPE(HistKey)

struct HistGroupInfo
{
    HistGroupInfo() = default;

    HistGroupInfo(HistId group_id, QString key, QString title, int n_bins, double min, double max,
                  QString toolTip)
        :
          strKey(key),
          title(title),
          tool_tip(toolTip),
          histId(group_id),
          n_bins(n_bins),
          min(min),
          max(max)
    {
    }

    QString strKey = "unknown";
    QString title;
    QString tool_tip;
    HistId histId = HistId::unknown;
    int n_bins = 0;
    double min = 0;
    double max = 0;
};

struct HistGroupCfg
{
    HistGroupCfg() = default;

    HistGroupCfg(HistId group_id) : hist_id(group_id)
    {
    }
    static bool isBaseHist(HistId histGroupId);
    bool isBaseHist() const { return isBaseHist(hist_id); }
    HistGroupInfo getHistGroupInfo() const { return getHistGroupInfo(hist_id); }
    static HistGroupInfo getHistGroupInfo(HistId id);

    HistId hist_id = HistId::unknown;
    bool hist_group_en = true;
    bool show_fit = true;
};
Q_DECLARE_METATYPE(HistGroupCfg)

struct HistMngrCfg
{
    HistMngrCfg() = default;

    void init(const DaqModule &daq_module_ = DaqModule(),
              bool is_inl_enabled_ = false);
    HistKey get_cur_hist_key() const;
    HistGroupInfo getHistGroupInfo(HistId id) const;
    void makeConfigValid();
private:
    void addHistGroup(HistId id);

public:
    DaqModule daq_module;
    bool is_inl_enabled = false;
    HistId cur_hist_id = HistId::unknown;

    DeviceIndex cur_dev_index;
    int cur_ch = 0;
    bool curEdgeLead = true;
    bool curEdgeTrail = false;
    bool use_stop_ch = false; // worker
    DeviceIndex stop_dev_index; // worker
    quint16 stop_ch = 0; // worker
    bool stopEdgeLead = true;
    bool stopEdgeTrail = false;

    bool inl_hist_code_dens = true; // hist
    bool global_enabled = true; // worker
    int tai_ts_window = 3600; // in sec, worker
    static const int mscPlotWin = 60; // in sec, worker
    int mw_min = 0; // match window min in ns
    int mw_max = 1000; // match window max in ns
    bool hitTimeEdgeLead = true;
    bool hitTimeEdgeTrail = false;
    bool log_y_axis = false;
    bool hide_zero_bin = false;
    bool ampl_hist_with_max = true; // fill adcAmpl with max or min value
    QMap<HistId, HistGroupCfg> hist_groups; // by hist group id
    QString histExportPath;
};
Q_DECLARE_METATYPE(HistMngrCfg)

inline bool operator==(const HistGroupCfg &l, const HistGroupCfg &r)
{
    return l.hist_id == r.hist_id &&
            l.hist_group_en == r.hist_group_en &&
            l.show_fit == r.show_fit;
}

inline bool operator==(const HistMngrCfg &l, const HistMngrCfg &r)
{
    return
            l.daq_module == r.daq_module &&
            l.is_inl_enabled == r.is_inl_enabled &&
            l.cur_hist_id == r.cur_hist_id &&
            l.inl_hist_code_dens == r.inl_hist_code_dens &&
            l.global_enabled == r.global_enabled &&
            l.tai_ts_window == r.tai_ts_window &&
            l.mw_min == r.mw_min &&
            l.mw_max == r.mw_max &&
            l.hitTimeEdgeLead == r.hitTimeEdgeLead &&
            l.hitTimeEdgeTrail == r.hitTimeEdgeTrail &&
            l.log_y_axis == r.log_y_axis &&
            l.hide_zero_bin == r.hide_zero_bin &&
            l.cur_ch == r.cur_ch &&
            l.curEdgeLead == r.curEdgeLead &&
            l.curEdgeTrail == r.curEdgeTrail &&
            l.stop_ch == r.stop_ch &&
            l.use_stop_ch == r.use_stop_ch &&
            l.stopEdgeLead == r.stopEdgeLead &&
            l.stopEdgeTrail == r.stopEdgeTrail &&
            l.ampl_hist_with_max == r.ampl_hist_with_max &&
            l.cur_dev_index == r.cur_dev_index &&
            l.stop_dev_index == r.stop_dev_index &&
            l.hist_groups == r.hist_groups &&
            l.histExportPath == r.histExportPath &&
            true;
}

QDebug operator<<(QDebug dbg, const HistId &r);
QDebug operator<<(QDebug dbg, const HistKey &r);
QDebug operator<<(QDebug dbg, const HistGroupCfg &c);
QDebug operator<<(QDebug dbg, const HistMngrCfg &c);
QDataStream &operator<<(QDataStream &out, const HistGroupCfg &r);
QDataStream &operator>>(QDataStream &in, HistGroupCfg &r);
QDataStream &operator<<(QDataStream &out, const HistMngrCfg &r);
QDataStream &operator>>(QDataStream &in, HistMngrCfg &r);

#endif // HISTMNGRCFG_H

