//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "RcClientManagerWidget.h"
#include "ui_RcClientManagerWidget.h"

#include "add-pnp-dialog/AddPnpDialog.h"
#include "base/ClientIndexConverter.h"
#include "base/Globals.h"
#include "base/ThreadController.h"
#include "mvc/MergeColumnsTableModel.h"
#include "pnp-server/PNPMessage.h"
#include "pnp-server/PnpDiscovery.h"
#include "remote-control-server/RunControlCommands.h"

#include "../RcClientManager.h"
#include "../RcClientStatusModel.h"
#include "../RcClientTreeModel.h"

#include <QCheckBox>
#include <QElapsedTimer>
#include <QHeaderView>
#include <QHostAddress>
#include <QLabel>
#include <QMessageBox>
#include <QJsonObject>
#include <QJsonDocument>
#include <QTableWidgetItem>
#include <QTimer>

namespace  {
bool isSupportEvNumCheck(const DaqModule &type)
{
    static const QList<DaqModule> EV_NUM_SUPPORT_LIST = {
        DaqModule(DaqModule::Type::Adc64),
        DaqModule(DaqModule::Type::Adc64System),
        DaqModule(DaqModule::Type::HrbSystem),
        DaqModule(DaqModule::Type::GenericClient),
        DaqModule(DaqModule::Type::Land),
        DaqModule(DaqModule::Type::TdcControl),
        DaqModule(DaqModule::Type::TQDC),
        DaqModule(DaqModule::Type::Ttvxs),
        DaqModule(DaqModule::Type::RunControl)
    };
    return EV_NUM_SUPPORT_LIST.contains(type);
}
}

RcClientManagerWidget::RcClientManagerWidget(QWidget *parent)
    : QWidget(parent)
    , ui(new Ui::RcClientManagerWidget)
    , enabledModel(new ClientEnableTableModel(this))
    , statusModel(new RcClientStatusModel(this))
    , mergeModel(new MergeColumnsTableModel(this))
    , rcClientTreeModel(new RcClientTreeModel(this))
    , applyCacheTimer(new QTimer(this))
{
    AbstractConfigConverter<ClientIndex>::registerTypes();
    qRegisterMetaType<RC_CLIENT_PROTO_ID>();
    qRegisterMetaType<FsmState>();
    qRegisterMetaType<ClientInfo>();
    ui->setupUi(this);
    if(!Globals::getDebugMode())
        ui->treeViewClients->hide();

    PnpDiscovery::subscribeCacheUpdaed(this);
    connect(this, &RcClientManagerWidget::clientInfoUpdated,
            this, &RcClientManagerWidget::onClientInfoUpdated);
    connect(statusModel, &RcClientStatusModel::rcClientsConfigChanged,
            this, &RcClientManagerWidget::rcClientsConfigChanged);
    connect(applyCacheTimer, &QTimer::timeout,
            this, &RcClientManagerWidget::updatePendingProgramDescription);
    applyCacheTimer->setSingleShot(true);
    applyCacheTimer->setInterval(500);
    applyCacheTimer->start();
    setupTable();
}

RcClientManagerWidget::~RcClientManagerWidget()
{
    delete ui;
}

void RcClientManagerWidget::setRcClientsConfig(const RcClientConfigMap &client_config_)
{
    if(config == client_config_)
        return;
    config = client_config_;
    statusModel->setRcClientsConfig(config);
    enabledModel->setClientList(config.getClientIndexEnabledMap());

    updateVisibilityColumn();
}

void RcClientManagerWidget::setupTable()
{
    connect(enabledModel, &ClientEnableTableModel::clientListChanged,
            this, &RcClientManagerWidget::setClientIndexEnabledMap);
    connect(this, &RcClientManagerWidget::clientInfoUpdated,
      statusModel, &RcClientStatusModel::updateClientStatus);

    mergeModel->addSourceModel(enabledModel, Qt::UserRole);
    mergeModel->addSourceModel(statusModel, Qt::UserRole);

    ui->tableViewClients->setModel(mergeModel);
    ui->tableViewClients->horizontalHeader()
            ->setSectionResizeMode(QHeaderView::ResizeToContents);
    for(auto col : {ClientEnableTableModel::COL_TYPE, ClientEnableTableModel::COL_INDEX}) {
        int globalCol = mergeModel->getColumnFromSource(enabledModel, col);
        ui->tableViewClients->hideColumn(globalCol);
    }
    auto model = ui->tableViewClients->selectionModel();
    connect(model, &QItemSelectionModel::selectionChanged,
            this, &RcClientManagerWidget::selectionChanged);

    ui->treeViewClients->setModel(rcClientTreeModel);
    connect(rcClientTreeModel, &RcClientTreeModel::modelReset,
            this, &RcClientManagerWidget::treeModelReseted);
    connect(this, &RcClientManagerWidget::clientRcProgramStateUpdated,
            rcClientTreeModel, &RcClientTreeModel::clientRcProgramStateUpdated);
    connect(this, &RcClientManagerWidget::myClientIndexUpdated,
            rcClientTreeModel, &RcClientTreeModel::setMyClientIndex);
}

void RcClientManagerWidget::on_pushButtonAdd_clicked()
{
    if(!hasFreeProgs()) {
        ui->pushButtonAdd->setEnabled(false);
        QMessageBox::information(this, "Add client...", "There is no free program");
        return;
    }

    AddPnpDialog dialog(pnpCache, true, this);
    PnpDiscovery::subscribeCacheUpdaed(&dialog);
    dialog.set_function_filter(std::bind(&RcClientManagerWidget::isFreeProgs,
                                         this, std::placeholders::_1));
 //   dialog.set_selected_progs(selectedProgDescr);
    if(dialog.exec() == QDialog::Accepted) {
        RcClientConfigMap newCfg(config);

        for(auto &descr : dialog.getProgDescr()) {
            auto index = descr.getClientIndex();
            RcClientConfig cfg;
            cfg.evNumCheck = isSupportEvNumCheck(index.getProgramType());
            newCfg.insert(index, cfg);
        }
        updateVisibilityColumn();
        emit rcClientsConfigChanged(newCfg);
    }
}

void RcClientManagerWidget::on_pushButtonRemove_clicked()
{
    auto clientId = getSelectedClient();
    if(clientId.isEmpty()){
        ui->pushButtonRemove->setEnabled(false);
        return;
    }

    RcClientConfigMap newCfg(config);
    Q_ASSERT(!clientId.isEmpty());
    newCfg.remove(clientId);

    emit rcClientsConfigChanged(newCfg);
}

void RcClientManagerWidget::selectionChanged(const QItemSelection &selected)
{
    auto list = selected.indexes();
    bool enRemove = !list.isEmpty() && isCfgChangeAllowed();
    ui->pushButtonRemove->setEnabled(enRemove);
}

void RcClientManagerWidget::pnpCacheUpdated(const PnpInfoCache &cache)
{
    pnpCache = cache;
    hasNewPnpCache = true;
}

void RcClientManagerWidget::updatePendingProgramDescription()
{
    if(hasNewPnpCache) {
        bool enAdd = hasFreeProgs() && isCfgChangeAllowed();
        if(ui->pushButtonAdd->isEnabled() != enAdd)
            ui->pushButtonAdd->setEnabled(enAdd);
    }
    applyCacheTimer->start();
}

void RcClientManagerWidget::onFsmStateChange(const FsmState &state)
{
    fsmState = state;
    statusModel->setExpectedFsmState(getExpectedClientsState(fsmState));
    updateButtonsState();
}

void RcClientManagerWidget::updateButtonsState()
{
    const bool enCfg= isCfgChangeAllowed();
    bool enAdd = hasFreeProgs() && enCfg;
    if(ui->pushButtonAdd->isEnabled() != enAdd)
        ui->pushButtonAdd->setEnabled(enAdd);
    bool enRemove = !getSelectedClient().isEmpty() && enCfg;
    if(ui->pushButtonRemove->isEnabled() != enRemove)
        ui->pushButtonRemove->setEnabled(enRemove);
    enabledModel->setEnabled(enCfg);
}

void RcClientManagerWidget::onClientInfoUpdated(const ClientInfo &info)
{
    auto &oldInfo = clients[info.getIndex()];
    bool upVme=oldInfo.hasVme() != info.hasVme();
    // bool upProto=oldInfo.protoId != info.protoId;

     oldInfo = info;
     if(upVme)
         updateVisibilityColumn(RcClientStatusModel::COL_IS_VME);
     // if(upProto)
     //     updateVisibilityColumn(RcClientStatusModel::COL_PROTOCOL);
}

void RcClientManagerWidget::setClientIndexEnabledMap(const ClientIndexEnabledMap &map)
{
    auto newConfig = config;
    ClientIndexSet old = makeClientIndexSet(newConfig.keys());
    for(auto it=map.begin(); it!=map.end(); ++it) {
        old.remove(it.key());
        auto it2 = newConfig.find(it.key());
        if(it2 == newConfig.end()) {
            RcClientConfig c(*it);
            newConfig.insert(it.key(), c);
        } else {
            if(it2->en == *it)
                continue;
            it2->en = *it;
        }
    }

    for(auto i : old)
        newConfig.remove(i);
    if(newConfig != config)
        emit rcClientsConfigChanged(newConfig);
}

void RcClientManagerWidget::treeModelReseted()
{
    ui->treeViewClients->expandAll();
    if(auto m = ui->treeViewClients->model()) {
        for(auto c=0; c<m->columnCount(); ++c)
            ui->treeViewClients->resizeColumnToContents(c);
    }
}

int RcClientManagerWidget::getClientRow(const ClientIndex &index) const
{
    for(int r=0; r<mergeModel->rowCount(); ++r) {
        const auto i = getClientId(r);
        if(index == i)
            return r;
    }
    return -1;
}

ClientIndex RcClientManagerWidget::getClientId(int row) const
{
    auto v = mergeModel->headerData(row, Qt::Vertical, Qt::UserRole);
    return v.value<ClientIndex>();
}

bool RcClientManagerWidget::hasFreeProgs() const
{
    for(const ProgramDescription &descr: pnpCache) {
        if(isFreeProgs(descr))
            return true;
    }
    return false;
}

bool RcClientManagerWidget::isFreeProgs(const ProgramDescription &descr) const
{
    bool ok;
    if(descr.isEmpty())
        return false;
    const ProgramInterface &interface = descr.getProgramInterface(ProgramInterfaceRemoteControl, &ok);
    if(!ok || !interface.enabled || !interface.isFree)
        return false;

    if(config.contains(descr.getClientIndex()))
        return false;

    return true;
}

ClientIndex RcClientManagerWidget::getSelectedClient() const
{
    QItemSelectionModel *select = ui->tableViewClients->selectionModel();
    if(!select->hasSelection())
        return {};

    auto indexes = select->selectedIndexes();
    if(indexes.isEmpty())
        return {};

    auto index = indexes.first();
    auto id = mergeModel->headerData(index.row(), Qt::Vertical, Qt::UserRole)
            .value<ClientIndex>();
    return id;
}

void RcClientManagerWidget::updateVisibilityColumn()
{
    for(auto col : {
                     // RcClientStatusModel::COL_PROTOCOL,
        RcClientStatusModel::COL_IS_VME})
        updateVisibilityColumn(col);
}

void RcClientManagerWidget::updateVisibilityColumn(int col)
{
    int globalCol = mergeModel->getColumnFromSource(statusModel, col);
    std::function<bool(const ClientInfo &)> isHidden = nullptr;
    bool hide = true;
    switch (col) {
    case RcClientStatusModel::COL_PROTOCOL:
        isHidden = [](const ClientInfo &ci) { return ci.protoId == RC_CLIENT_PROTO_V5; };
        break;
    case RcClientStatusModel::COL_IS_VME:
        isHidden = [](const ClientInfo &ci) { return !ci.hasVme(); };
        break;
    default:
        return;
    }

    for(auto it=config.cbegin(); it!=config.cend(); ++it) {
        if(!it->en)
            continue;
        hide = isHidden(clients[it.key()]);
        if(!hide) break;
    }
    ui->tableViewClients->setColumnHidden(globalCol, hide);
}

FsmState RcClientManagerWidget::getExpectedClientsState(const FsmState &state) const
{
    switch (state) {
    case FsmState::Calibrate:
    case FsmState::Starting:
    case FsmState::Run:
        return FsmState::Run;
    default:
        return FsmState::Ready;
    }
}

bool RcClientManagerWidget::isCfgChangeAllowed() const
{
    return fsmState == FsmState::Error || fsmState == FsmState::Ready;
}
