#include "TrigDeviceMgrFsm.h"

#include <QDebug>
#include <QLoggingCategory>

namespace {
Q_DECLARE_LOGGING_CATEGORY(fsm)
Q_LOGGING_CATEGORY(fsm, "afi.fsm.TrigDeviceMgrFsm")

struct Transaction
{
    FsmState from;
    FsmState to;
    QString reason;
    QString action;
};

// MM - megamaster flag
const QList<Transaction> TransactionMap = {
    {FsmState::Init, FsmState::Error, "", ""},
    {FsmState::Error, FsmState::Ready, "cfg written to device", ""},
    {FsmState::Ready, FsmState::Starting, "[!MM] on remote start request (external slot)", "open trigger in cfg and write to dev"},
    {FsmState::Ready, FsmState::Starting, "[MM] on local start request (external slot)", "close trigger in cfg and write to dev"},
    {FsmState::Starting, FsmState::Run, "[!MM] on cfg (with opened trig) written to device", ""},
    {FsmState::Starting, FsmState::CloseGlobTrig, "[MM] on cfg (with closed trig) written to device", "then wait while all client will be in run state"},
    {FsmState::CloseGlobTrig, FsmState::OpenGlobTrig, "[MM] all system is ready to accept triggers (external slot)", "open trigger in cfg and write to dev"},
    {FsmState::OpenGlobTrig, FsmState::Run, "[MM] on cfg (with opened trig) written to device", ""},
    {FsmState::Run, FsmState::Stopping, "[!MM] on remote stop request", "close trigger in cfg and write to dev"},
    {FsmState::Stopping, FsmState::Ready, "[!MM] on cfg (with closed trig) written to device", ""},
    {FsmState::Unknown, FsmState::Error, "devSet is empty", ""},
    {FsmState::Unknown, FsmState::Error, "", "close trig"},
};
}

TrigDeviceMgrFsm::TrigDeviceMgrFsm(QObject *p):
    DeviceMgrFsm(p)
{
    setObjectName("TrigDeviceMgrFsm");
    recreateInput(new TrigDeviceMgrFsmInput());
}

bool TrigDeviceMgrFsm::evaluate()
{
    if(getInput()->devSet.isEmpty()) {
        setState(FsmState::Error);
        return true;
    }

    if(DeviceMgrFsm::evaluate())
        return true;

    return false;
}

void TrigDeviceMgrFsm::printInputString(QDebug dbg) const
{
    dbg << *getInput();
}

void TrigDeviceMgrFsm::markStopEvNumber()
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName();
    getInput()->hasStopEvNumber = true;
    evaluate();
}

void TrigDeviceMgrFsm::clearStopEvNumber()
{
    qDebug(fsm) << __PRETTY_FUNCTION__ << objectName();
    getInput()->hasStopEvNumber = false;
    evaluate();
}

bool TrigDeviceMgrFsm::isTransitionAllowed(const FsmState &toState) const
{
    switch (getState()) {
    case FsmState::Starting:
        switch (toState) {
        // MM case
        case FsmState::CloseGlobTrig: return allDevicesWritten();
        default:;
        }
        break;
    case FsmState::CloseGlobTrig:
        switch (getRequestState()) {
        case FsmState::Stopping: return true;
        case FsmState::OpenGlobTrig: return allDevicesWritten();
        default:;
        }
        break;
    case FsmState::OpenGlobTrig:
        // MM case
        if(toState == FsmState::Run)
            return allDevicesWritten();
        break;
    case FsmState::Stopping:
        if(toState == FsmState::Ready)
            return getInput()->hasStopEvNumber && DeviceMgrFsm::isTransitionAllowed(toState);
        break;
    default:;
    }
    return DeviceMgrFsm::isTransitionAllowed(toState);
}

qint64 TrigDeviceMgrFsm::stateTimeoutMs(const FsmState &state) const
{
    switch (state) {
//    case FsmState::CloseGlobTrig: return 20e3;
    case FsmState::OpenGlobTrig: return DEVICE_CONFIG_WRITE_TIMEOUT;
    default:
        return DeviceMgrFsm::stateTimeoutMs(state);
    }
}

QDebug operator<<(QDebug dbg, const TrigDeviceMgrFsmInput &in)
{
    dbg << DeviceMgrFsmInput(in);

    dbg << "hasStopEvNumber:" << in.hasStopEvNumber << "; ";

    return dbg;
}
