/*
**    Copyright 2013-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "regop.h"

namespace mlink {

QDebug operator<<(QDebug debug, const mlink::RegOp &v)
{
    static const QMap<ByteCount, int> map = {
        {ByteCount::NONE, 0},
        {ByteCount::D16, 2},
        {ByteCount::D32, 4},
        {ByteCount::D64, 8}
    };
    int dataChars = map[v.dataWidth()] * 2;
    debug << QString("[%1] %2 %3")
                 .arg(v.addr, 4, 16, QChar('0'))
                 .arg(v.isWrite() ? "write" : (v.isRead() ? "read" : ""))
                 .arg(v.isWrite()
                          ? QString("%1").arg(v.wdata, dataChars, 16, QChar('0'))
                          : (v.isRead()
                                 ? (v.result.ok ? QString("%1").arg(v.rdata, dataChars, 16, QChar('0')) : "?")
                                 : QString("")));
    return debug;
}

ByteCount RegOp::dataWidth() const
{
    switch (mode) {
    case OpMode::NOP:
        return ByteCount::NONE;
    case OpMode::R16LE:
    case OpMode::R16BE:
    case OpMode::W16LE:
    case OpMode::W16BE:
        return ByteCount::D16;
    case OpMode::R32LE:
    case OpMode::R32BE:
    case OpMode::W32LE:
    case OpMode::W32BE:
        return ByteCount::D32;
    case OpMode::R64LE:
    case OpMode::R64BE:
    case OpMode::W64LE:
    case OpMode::W64BE:
        return ByteCount::D64;
    }
    return ByteCount::NONE;
}

bool RegOp::isWrite() const
{
    switch (mode) {
    case OpMode::NOP:
        return false;
    case OpMode::W16LE:
    case OpMode::W16BE:
    case OpMode::W32LE:
    case OpMode::W32BE:
    case OpMode::W64LE:
    case OpMode::W64BE:
        return true;
    case OpMode::R16LE:
    case OpMode::R16BE:
    case OpMode::R32LE:
    case OpMode::R32BE:
    case OpMode::R64LE:
    case OpMode::R64BE:
        return false;
    }
    return false;
}

bool RegOp::isRead() const
{
    switch (mode) {
    case OpMode::NOP:
        return false;
    case OpMode::W16LE:
    case OpMode::W16BE:
    case OpMode::W32LE:
    case OpMode::W32BE:
    case OpMode::W64LE:
    case OpMode::W64BE:
        return false;
    case OpMode::R16LE:
    case OpMode::R16BE:
    case OpMode::R32LE:
    case OpMode::R32BE:
    case OpMode::R64LE:
    case OpMode::R64BE:
        return true;
    }
    return false;
}

Endianness RegOp::endianness() const
{
    switch (mode) {
    case OpMode::NOP:
        return Endianness::LE;
    case OpMode::R16LE:
    case OpMode::W16LE:
    case OpMode::R32LE:
    case OpMode::W32LE:
    case OpMode::R64LE:
    case OpMode::W64LE:
        return Endianness::LE;
    case OpMode::R16BE:
    case OpMode::W16BE:
    case OpMode::R32BE:
    case OpMode::W32BE:
    case OpMode::R64BE:
    case OpMode::W64BE:
        return Endianness::BE;
    }
    return Endianness::LE;
}

void RegOp::decode()
{
    if (ok)
        *ok = result.ok;
    if (result.ok && rdptr && isRead()) {
        switch (dataWidth()) {
        case ByteCount::D16:
            *static_cast<quint16 *>(rdptr) = rdata;
            break;
        case ByteCount::D32:
            *static_cast<quint32 *>(rdptr) = rdata;
            break;
        case ByteCount::D64:
            *static_cast<quint64 *>(rdptr) = rdata;
            break;
        case ByteCount::NONE:
            break;
        }
    }
}

} // namespace mlink
